/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_mpi
 *  @brief   Declaration of the class LocalData.
 *
 *  This header file provides the declaration of the class LocalData.
 **/
/*-------------------------------------------------------------------------*/


#ifndef PEARL_LOCALDATA_H
#define PEARL_LOCALDATA_H


#include <cstdint>
#include <map>
#include <vector>

#include <pearl/Event.h>
#include <pearl/EventSet.h>


namespace pearl
{
// --- Forward declarations -------------------------------------------------

class Buffer;
class Communicator;
class MpiMessage;


/*-------------------------------------------------------------------------*/
/**
 *  @class   LocalData
 *  @ingroup PEARL_mpi
 *  @brief   Container class for a set of local events and memory buffers
 *           with associated roles.
 **/
/*-------------------------------------------------------------------------*/

class LocalData
{
    public:
        // --- Public methods ------------------------------

        /// @name Constructors & destructor
        /// @{

        ~LocalData();

        /// @}
        /// @name Adding/updating data
        /// @{

        void
        addEvent(const Event&  event,
                 std::uint32_t key);

        void
        updateEvent(const Event&  event,
                    std::uint32_t key);

        void
        addBuffer(Buffer*       buffer,
                  std::uint32_t key);

        /// @}
        /// @name Retrieving data
        /// @{

        Event
        getEvent(std::uint32_t key) const;

        Buffer*
        getBuffer(std::uint32_t key) const;

        /// @}
        /// @name Removing data
        /// @{

        void
        clear();

        /// @}
        /// @name Sending data
        /// @{

        MpiMessage*
        isend(const Communicator& comm,
              int                 dest,
              int                 tag) const;

        /// @}


    private:
        // --- Type definitions ----------------------------

        /// Container type for role |-@> index mapping
        typedef std::map< std::uint32_t, std::uint32_t > KeyMap;

        /// Container type for local buffers
        typedef std::vector< Buffer* > BufferList;


        // --- Data members --------------------------------

        /// Current set of events
        EventSet mEvents;

        /// Mapping table key (role ID) |-@> index for buffers
        KeyMap mBufferKeys;

        /// Current set of buffers
        BufferList mBuffers;


        // --- Private methods -----------------------------

        /// @name Packing messages
        /// @{

        void
        pack(MpiMessage& message) const;

        /// @}
};
}    // namespace pearl


#endif    // !PEARL_LOCALDATA_H
