/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2024                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef PEARL_NONBLOCKINGCOLLCOMPLETE_REP_H
#define PEARL_NONBLOCKINGCOLLCOMPLETE_REP_H


#include <cstdint>

#include <pearl/Event_rep.h>


/*-------------------------------------------------------------------------*/
/**
 *  @file    NonBlockingCollComplete_rep.h
 *  @ingroup PEARL_base
 *  @brief   Declaration of the class NonBlockingCollComplete_rep.
 *
 *  This header file provides the declaration of the class
 *  NonBlockingCollComplete_rep.
 **/
/*-------------------------------------------------------------------------*/


namespace pearl
{
/*-------------------------------------------------------------------------*/
/**
 *  @class   NonBlockingCollComplete_rep
 *  @ingroup PEARL_base
 *  @brief   %Event representation for NON_BLOCKING_COLLECTIVE_COMPLETE
 *           events.
 **/
/*-------------------------------------------------------------------------*/

class PEARL_NOPAD_ATTR NonBlockingCollComplete_rep
    : public Event_rep
{
    public:
        /// @name Constructors & destructor
        /// @{

        NonBlockingCollComplete_rep(timestamp_t         timestamp,
                                    CollectiveOpType    type,
                                    const Communicator& communicator,
                                    std::uint32_t       root,
                                    std::uint64_t       bytesSent,
                                    std::uint64_t       bytesReceived,
                                    std::uint64_t       requestId);

        NonBlockingCollComplete_rep(const GlobalDefs& defs,
                                    Buffer&           buffer);

        /// @}
        /// @name Event type information
        /// @{

        virtual event_t
        getType() const;

        virtual bool
        isOfType(event_t type) const;

        /// @}
        /// @name Access event data
        /// @{

        virtual const Communicator&
        getComm() const;

        virtual std::uint32_t
        getRoot() const;

        virtual std::uint64_t
        getBytesSent() const;

        virtual std::uint64_t
        getBytesReceived() const;

        virtual std::uint64_t
        getRequestId() const;

        CollectiveOpType
        getCollType() const;

        /// @}


    protected:
        /// @name Serialize event data
        /// @{

        virtual void
        pack(Buffer& buffer) const;

        /// @name Generate human-readable output of event data
        /// @{

        virtual std::ostream&
        output(std::ostream& stream) const;

        /// @}
        /// @name Find previous/next communication request entries
        /// @{

        virtual std::uint32_t
        getNextReqoffs() const;

        virtual std::uint32_t
        getPrevReqoffs() const;

        virtual void
        setPrevReqoffs(std::uint32_t offset);

        virtual void
        setNextReqoffs(std::uint32_t offset);

        /// @}


    private:
        /// Type of the nonblocking collective operation
        CollectiveOpType mType;

        /// Communicator used in collective operation
        const Communicator& mCommunicator;

        /// Local rank of root location w.r.t. communicator
        std::uint32_t mRoot;

        /// Number of bytes sent
        std::uint64_t mBytesSent;

        /// Number of bytes received
        std::uint64_t mBytesReceived;

        /// Non-blocking request ID
        std::uint64_t mRequestId;

        /// Offset to previous test or non-blocking collective request event
        std::uint32_t mPrevReqoffs;
};
}    // namespace pearl


#endif    // !PEARL_NONBLOCKINGCOLLECTIVECOMPLETE_REP_H
