/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2021                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief Backported functionality from post-C++11 `<memory>`.
 *
 *  This header file provides a compatibility layer for selected (useful)
 *  capabilities of the C++ standard library header `<memory>` that were
 *  introduced in language standard versions newer than C++11.
 *
 *  @note
 *      This file is not intended to be feature-complete, but only contains
 *      functionality used in the code base.
 **/
/*-------------------------------------------------------------------------*/


#ifndef CXXSLB_MEMORY_H
#define CXXSLB_MEMORY_H


#include <cstddef>
#include <memory>
#include <utility>

#include <cxxslb/type_traits.h>


namespace cxxslb
{
/// @name Smart pointer non-member operations
/// @{

#if __cplusplus >= 201402L
    using std::make_unique;
#else    // C++11
    namespace detail
    {
    template< class T >
    struct MakeUniqueIf
    {
        typedef std::unique_ptr< T > SingleObject;
    };


    template< class T >
    struct MakeUniqueIf< T[] >
    {
        typedef std::unique_ptr< T[] > UnknownBound;
    };


    template< class T,
              std::size_t N >
    struct MakeUniqueIf< T[N] >
    {
        typedef void KnownBound;
    };
    }    // namespace cxxslb::detail


    /// @brief Creates a unique pointer that manages a new object.
    ///
    /// Constructs an object of non-array type @p T and wraps it in an
    /// std::unique_ptr.  The arguments @p args are passed to the constructor
    /// of @p T.
    ///
    /// @tparam T
    ///     Non-array type
    /// @tparam Args
    ///     Constructor argument types
    /// @param args
    ///     List of arguments with which an instance of T will be constructed
    /// @returns
    ///     Newly allocated object wrapped inside an std::unique_ptr
    ///
    /// @isocpp{14,memory,N3656}
    ///
    template< class T,
              class ... Args >
    typename detail::MakeUniqueIf< T >::SingleObject
    make_unique(Args&& ... args)
    {
        return std::unique_ptr< T >(new T(std::forward< Args >(args) ...));
    }


    /// @brief Creates a unique pointer that manages a new array.
    ///
    /// Constructs an array of size @p n of unknown bound array type @p T
    /// and wraps it in an std::unique_ptr.
    ///
    /// @tparam T
    ///     Unknown bound array type
    /// @param n
    ///     Number of array elements
    /// @returns
    ///     Newly allocated array wrapped inside an std:unique_ptr
    ///
    /// @isocpp{14,memory,N3656}
    ///
    template< class T >
    typename detail::MakeUniqueIf< T >::UnknownBound
    make_unique(std::size_t n)
    {
        return std::unique_ptr< T >(new cxxslb::remove_extent_t< T >[n]());
    }


    /// @brief Construction of arrays of known bound array types is disallowed.
    template< class T,
              class ... Args >
    typename detail::MakeUniqueIf< T >::KnownBound
    make_unique(Args&& ...) = delete;
#endif    // C++11

/// @}
}    // namespace cxxslb


#endif    // !CXXSLB_MEMORY_H
