/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_mpi
 *  @brief   Implementation of the class MpiDefsFactory.
 *
 *  This file provides the implementation of the class MpiDefsFactory.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include "MpiDefsFactory.h"

#include <algorithm>
#include <string>

#include <cxxslb/memory.h>

#include <mpi.h>

#include <pearl/Error.h>
#include <pearl/MpiComm.h>
#include <pearl/MpiGroup.h>
#include <pearl/Paradigm.h>

#include "MutableGlobalDefs.h"

using namespace std;
using namespace pearl;
using namespace pearl::detail;


// --- Generic factory methods ----------------------------------------------

void
MpiDefsFactory::createCommunicator(MutableGlobalDefs&             defs,
                                   const uint32_t                 id,
                                   const uint32_t                 nameId,
                                   const Communicator::Properties properties,
                                   const uint32_t                 commSetId,
                                   const uint32_t                 parentId) const
{
    const GroupingSet& groupingSet = defs.getGroupingSet(commSetId);
    if (groupingSet.getType() != GroupingSet::COMM_SET)
    {
        throw FatalError("pearl::detail::MpiDefsFactory::createCommunicator -- "
                         "Grouping set "
                         + std::to_string(commSetId)
                         + " is not a communication set");
    }

    const CommSet& commSet  = static_cast< const CommSet& >(groupingSet);
    const Paradigm paradigm = commSet.getParadigm();
    if (paradigm != Paradigm::MPI)
    {
        DefsFactory::createCommunicator(defs,
                                        id,
                                        nameId,
                                        properties,
                                        commSetId,
                                        parentId);

        return;
    }

    auto& parent = defs.getCommunicator(parentId);
    if (  (parent != Communicator::UNDEFINED)
       && (parent.getParadigm() != paradigm))
    {
        throw FatalError("pearl::detail::MpiDefsFactory::createCommunicator -- "
                         "Paradigm mismatch: "
                         + toString(paradigm)
                         + " (communication set) vs. "
                         + toString(parent.getParadigm())
                         + " (parent)");
    }

    auto& group = static_cast< const MpiGroup& >(commSet);
    defs.addCommunicator(
        cxxslb::make_unique< MpiComm >(id,
                                       defs.getString(nameId),
                                       properties,
                                       group,
                                       (parent != Communicator::UNDEFINED)
                                       ? &parent
                                       : nullptr));
}


void
MpiDefsFactory::createGroupingSet(MutableGlobalDefs&             defs,
                                  const uint32_t                 id,
                                  const uint32_t                 nameId,
                                  const GroupingSet::SetType     type,
                                  const Paradigm                 paradigm,
                                  const GroupingSet::Properties& properties,
                                  const uint32_t                 numMembers,
                                  const uint64_t* const          members) const
{
    if (  (type == GroupingSet::COMM_SET)
       && (paradigm == Paradigm::MPI))
    {
        vector< uint32_t > ranks;

        // Special case 1: self-like groups
        if (properties.test(GroupingSet::PROPERTY_SELF))
        {
            int rank;
            MPI_Comm_rank(MPI_COMM_WORLD, &rank);

            ranks.push_back(rank);
        }

        // Special case 2: world-like groups => empty ranks container;
        // Handle the general case instead
        else if (!properties.test(GroupingSet::PROPERTY_WORLD))
        {
            ranks.reserve(numMembers);
            copy(members, members + numMembers, back_inserter(ranks));
        }

        auto group =
            cxxslb::make_unique< MpiGroup >(id,
                                            defs.getString(nameId),
                                            properties,
                                            defs.getCommLocationSet(paradigm),
                                            ranks);

        // Sanity check
        int world_size;
        MPI_Comm_size(MPI_COMM_WORLD, &world_size);
        if (static_cast< uint32_t >(world_size) < group->numRanks())
        {
            throw FatalError("pearl::detail::MpiDefsFactory::createGroupingSet() -- "
                             "Experiment requires at least "
                             + std::to_string(defs.numLocationGroups())
                             + " MPI ranks.");
        }

        defs.addGroupingSet(std::move(group));
    }
    else
    {
        DefsFactory::createGroupingSet(defs,
                                       id,
                                       nameId,
                                       type,
                                       paradigm,
                                       properties,
                                       numMembers,
                                       members);
    }
}
