/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_base
 *  @brief   Declaration of the class DefsFactory.
 *
 *  This header file provides the declaration of the class DefsFactory.
 **/
/*-------------------------------------------------------------------------*/


#ifndef PEARL_DEFSFACTORY_H
#define PEARL_DEFSFACTORY_H


#include <cstdint>
#include <string>
#include <vector>

#include <UTILS_Error.h>

#include <pearl/CartDimension.h>
#include <pearl/Communicator.h>
#include <pearl/GroupingSet.h>
#include <pearl/Location.h>
#include <pearl/LocationGroup.h>
#include <pearl/Region.h>
#include <pearl/pearl_types.h>


namespace pearl
{
namespace detail
{
// --- Forward declarations -------------------------------------------------

class MutableGlobalDefs;


/*-------------------------------------------------------------------------*/
/**
 *  @class   DefsFactory
 *  @ingroup PEARL_base
 *  @brief   Abstract factory class for global definition objects.
 **/
/*-------------------------------------------------------------------------*/

class DefsFactory
{
    public:
        /// Internal region specifiers
        enum InternalRegionType
        {
            USER_REGION,       ///< User-defined region
            PAUSING_REGION,    ///< %Region where measurement was turned off
            FLUSHING_REGION,   ///< %Buffer flush region
            TASKROOT_REGION,   ///< %Task root region
            THREADROOT_REGION, ///< %Thread root region

            NUMBER_OF_TYPES    ///< Not a valid type, for internal use only
        };


        /// @name Constructors & destructor
        /// @{

        virtual
        ~DefsFactory();

        /// @}
        /// @name Generic factory methods
        /// @{

        void
        createCallingContext(MutableGlobalDefs& defs,
                             std::uint32_t      id,
                             std::uint32_t      regionId,
                             std::uint32_t      sclId,
                             std::uint32_t      parentId) const;

        void
        createCallpath(MutableGlobalDefs& defs,
                       std::uint32_t      id,
                       std::uint32_t      regionId,
                       std::uint32_t      callsiteId,
                       std::uint32_t      parentId) const;

        void
        createCallsite(MutableGlobalDefs& defs,
                       std::uint32_t      id,
                       std::uint32_t      filenameId,
                       std::uint32_t      line,
                       std::uint32_t      regionId) const;

        void
        createCartDimension(MutableGlobalDefs&         defs,
                            std::uint32_t              id,
                            std::uint32_t              nameId,
                            std::uint32_t              size,
                            CartDimension::Periodicity periodicity) const;

        void
        createCartTopology(MutableGlobalDefs&   defs,
                           std::uint32_t        id,
                           std::uint32_t        nameId,
                           std::uint32_t        commId,
                           std::uint8_t         numDimensions,
                           const std::uint32_t* dimensionIds) const;

        virtual void
        createCommunicator(MutableGlobalDefs&       defs,
                           std::uint32_t            id,
                           std::uint32_t            nameId,
                           Communicator::Properties properties,
                           std::uint32_t            commSetId,
                           std::uint32_t            parentId) const;

        virtual void
        createGroupingSet(MutableGlobalDefs&             defs,
                          std::uint32_t                  id,
                          std::uint32_t                  nameId,
                          GroupingSet::SetType           type,
                          Paradigm                       paradigm,
                          const GroupingSet::Properties& properties,
                          std::uint32_t                  numMembers,
                          const std::uint64_t*           members) const;

        void
        createLocation(MutableGlobalDefs& defs,
                       std::uint64_t      id,
                       std::uint32_t      nameId,
                       Location::Type     type,
                       std::uint64_t      numEvents,
                       std::uint32_t      parentId) const;

        void
        createLocationGroup(MutableGlobalDefs&  defs,
                            std::uint32_t       id,
                            std::uint32_t       nameId,
                            LocationGroup::Type type,
                            std::uint32_t       parentId,
                            std::uint32_t       creatorId) const;

        void
        createRegion(MutableGlobalDefs& defs,
                     std::uint32_t      id,
                     std::uint32_t      canonicalNameId,
                     std::uint32_t      nameId,
                     std::uint32_t      descriptionId,
                     Region::Role       role,
                     Paradigm           paradigm,
                     std::uint32_t      filenameId,
                     std::uint32_t      startLine,
                     std::uint32_t      endLine,
                     InternalRegionType type) const;

        void
        createSourceLocation(MutableGlobalDefs& defs,
                             std::uint32_t      id,
                             std::uint32_t      filenameId,
                             std::uint32_t      line) const;

        void
        createString(MutableGlobalDefs& defs,
                     std::uint32_t      id,
                     const std::string& str) const;

        void
        createSystemNode(MutableGlobalDefs& defs,
                         std::uint32_t      id,
                         std::uint32_t      nameId,
                         std::uint32_t      classId,
                         std::uint32_t      parentId) const;

        void
        createMetric(MutableGlobalDefs& defs,
                     ident_t            id,
                     const std::string& name,
                     const std::string& descr,
                     std::uint8_t       type,
                     std::uint8_t       mode,
                     std::uint8_t       interval) const;

        void
        createRmaWindow(MutableGlobalDefs& defs,
                        std::uint32_t      id,
                        std::uint32_t      comm_id) const;

        /// @}
        /// @name Singleton interface
        /// @{

        static const DefsFactory*
        instance();

        static void
        registerFactory(const DefsFactory* factory);

        static void
        deregisterFactory();

        /// @}


    private:
        /// Single factory instance
        static const DefsFactory* m_instance;
};


// --- Inline methods -------------------------------------------------------

inline const DefsFactory*
DefsFactory::instance()
{
    UTILS_ASSERT(m_instance != nullptr);

    return m_instance;
}
}    // namespace pearl::detail
}    // namespace pearl


#endif    // !PEARL_DEFSFACTORY_H
