/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include <pearl/EventSet.h>

#include <algorithm>
#include <string>

#include <UTILS_Error.h>

#include <pearl/Buffer.h>
#include <pearl/Error.h>

using namespace std;
using namespace pearl;


// --- Adding/updating events -----------------------------------------------

void
EventSet::addEvent(const Event& event,
                   uint32_t     key)
{
    const auto index = storeEvent(event);

    auto result = mKeys.emplace(key, index);

    UTILS_ASSERT(result.second == true);
}


void
EventSet::updateEvent(const Event& event,
                      uint32_t     key)
{
    const auto index = storeEvent(event);

    mKeys[key] = index;
}


// --- Retrieving events ----------------------------------------------------

Event
EventSet::getEvent(uint32_t key) const
{
    KeyMap::const_iterator it = mKeys.find(key);
    if (it == mKeys.end())
    {
        throw RuntimeError("EventSet::getEvent(uint32_t) "
                           "-- Unknown key: " + std::to_string(key));
    }

    return mEvents[it->second];
}


// --- Removing events ------------------------------------------------------

void
EventSet::clear()
{
    mKeys.clear();
    mEvents.clear();
}


// --- Serializing data -----------------------------------------------------

void
EventSet::pack(Buffer& buffer) const
{
    const uint32_t keyCount = mKeys.size();
    buffer.put_uint32(keyCount);
    for (const auto& mapItem : mKeys)
    {
        buffer.put_uint32(mapItem.first);
        buffer.put_uint32(mapItem.second);
    }

    const uint32_t eventCount = mEvents.size();
    buffer.put_uint32(eventCount);
    for (const auto& event : mEvents)
    {
        buffer.put_event(event);
    }
}


// --- Private methods ------------------------------------------------------

uint32_t
EventSet::storeEvent(const Event& event)
{
    auto it = find(mEvents.begin(), mEvents.end(), event);

    const auto index = it - mEvents.begin();
    if (it == mEvents.end())
    {
        mEvents.push_back(event);
    }

    return index;
}
