/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2017-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_base
 *  @brief   Implementation of the class GroupingSet.
 *
 *  This file provides the implementation of the class GroupingSet and
 *  related functions.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include <pearl/GroupingSet.h>

#include <string>

#include <UTILS_Error.h>

#include <pearl/iomanip.h>

#include "iomanip_detail.h"

using namespace std;
using namespace pearl;
using namespace pearl::detail;


// --- Constructors & destructor --------------------------------------------

GroupingSet::GroupingSet(const IdType  id,
                         const String& name)
    : mName(name),
      mIdentifier(id)
{
}


GroupingSet::~GroupingSet()
{
}


// --- Query functions ------------------------------------------------------

GroupingSet::IdType
GroupingSet::getId() const
{
    return mIdentifier;
}


const String&
GroupingSet::getName() const
{
    return mName;
}


// --- Stream I/O functions -------------------------------------------------

ostream&
pearl::operator<<(ostream&           stream,
                  const GroupingSet& item)
{
    const long detail = getDetail(stream);

    // Special case: skipped output
    if (detail <= DETAIL_SKIP)
    {
        return stream;
    }

    // Print data
    return item.output(stream);
}


// --- Conversion functions -------------------------------------------------

string
pearl::toStdString(const GroupingSet::SetType type)
{
    switch (type)
    {
        case GroupingSet::LOCATION_SET:
            return "Location set";

        case GroupingSet::REGION_SET:
            return "Region set";

        case GroupingSet::METRIC_SET:
            return "Metric set";

        case GroupingSet::COMM_LOCATION_SET:
            return "Comm location set";

        case GroupingSet::COMM_SET:
            return "Comm set";

        default:
            break;
    }

    // We should never get here...
    UTILS_BUG("Unknown grouping set type (%d)", static_cast< int >(type));
}


string
pearl::toStdString(const GroupingSet::Properties& properties)
{
    string result = std::to_string(properties.getValue()).append(" (");
    if (properties.test(GroupingSet::PROPERTY_NONE))
    {
        result += "NONE";
    }
    else
    {
        if (properties.test(GroupingSet::PROPERTY_SELF))
        {
            result += "self";
        }
        if (properties.test(GroupingSet::PROPERTY_WORLD))
        {
            if (result.back() != '(')
            {
                result += ',';
            }
            result += "world";
        }
        if (properties.test(GroupingSet::PROPERTY_GLOBAL_RANKS))
        {
            if (result.back() != '(')
            {
                result += ',';
            }
            result += "global ranks";
        }
    }
    result += ')';

    return result;
}
