/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_base
 *  @brief   Implementation of the class MutableGlobalDefs.
 *
 *  This file provides the implementation of the class MutableGlobalDefs.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include "MutableGlobalDefs.h"

#include <cstddef>
#include <limits>
#include <string>
#include <utility>

#include <cxxslb/memory.h>

#include <UTILS_Error.h>

#include <pearl/CommLocationSet.h>
#include <pearl/Error.h>

#include "CallTree.h"
#include "DefinitionContainer.h"
#include "DefinitionTree.h"
#include "MutableCartTopology.h"

using namespace std;
using namespace pearl;
using namespace pearl::detail;


// --- Helper functions -----------------------------------------------------

namespace
{
/// @brief   Ensures unique IDs for global definitions.
/// @relates pearl::detail::MutableGlobalDefs
///
/// Checks whether the global definition objects stored in @p container have
/// unique global identifiers and throws an exception if not.  This function
/// implicitly also optimizes lookups by identifier.
///
/// @param container
///     Global definition container
/// @param defName
///     Definition name used in the exception message
///
/// @tparam DefContainerT
///     Global definition container type providing the required typedef's and
///     member functions (DefinitionContainer or DefinitionTree)
///
/// @throws RuntimeError
///     on identifying a duplicate identifier
///
template< typename DefContainerT >
void
ensureUniqueIds(DefContainerT& container,
                const string&  defName)
{
    // Special case: empty container
    if (container.isEmpty())
    {
        return;
    }

    // Step 1: ensure that definitions are sorted by ID
    container.optimizeLookups();

    // Step 2: ensure that consecutive definitions have different IDs
    typename DefContainerT::ElementType::IdType prevId =
        container.getItemByIndex(0)->getId();
    const size_t numItems = container.numItems();
    for (size_t i = 1; i < numItems; ++i)
    {
        typename DefContainerT::ElementType::IdType currId =
            container.getItemByIndex(i)->getId();
        if (currId == prevId)
        {
            throw RuntimeError("pearl::detail::MutableGlobalDefs::"
                               "validate() -- "
                               "duplicate " + defName + " ID: "
                               + std::to_string(currId));
        }
        prevId = currId;
    }
}
}    // unnamed namespace


// --- Constructors & destructor --------------------------------------------

MutableGlobalDefs::MutableGlobalDefs()
    : mGlobalOffset(std::numeric_limits< double >::max()),
      mLastDefType(UNDEFINED)
{
}


// --- Insertion functions for global definitions ---------------------------

void
MutableGlobalDefs::addCallingContext(std::unique_ptr< CallingContext > item)
{
    trackType(CALLINGCONTEXT);

    mContextTree->addItem(std::move(item));
}


void
MutableGlobalDefs::addCallpath(std::unique_ptr< Callpath > item)
{
    trackType(CALLPATH);

    mCallTree->addCallpath(std::move(item));
}


void
MutableGlobalDefs::addCallsite(std::unique_ptr< Callsite > item)
{
    trackType(CALLSITE);

    mCallsites->addItem(std::move(item));
}


void
MutableGlobalDefs::addCartDimension(std::unique_ptr< CartDimension > item)
{
    trackType(CARTDIMENSION);

    mCartDimensions->addItem(std::move(item));
}


void
MutableGlobalDefs::addCartTopology(std::unique_ptr< CartTopology > item)
{
    trackType(CARTTOPOLOGY);

    mCartTopologies->addItem(std::move(item));
}


void
MutableGlobalDefs::addCommunicator(std::unique_ptr< Communicator > item)
{
    trackType(COMMUNICATOR);

    mCommunicators->addItem(std::move(item));
}


void
MutableGlobalDefs::addGroupingSet(std::unique_ptr< GroupingSet > item)
{
    trackType(GROUPINGSET);

    // Track communication location sets separately, using
    // "commit-or-rollback" to fulfill strong/basic exception
    // safety guarantee
    const bool isCommLocSet = (item->getType() == GroupingSet::COMM_LOCATION_SET);
    if (isCommLocSet)
    {
        mCommLocationSets.push_back(static_cast< CommLocationSet* >(item.get()));
    }
    try
    {
        mGroupingSets->addItem(std::move(item));
    }
    catch (exception& ex)
    {
        if (isCommLocSet)
        {
            mCommLocationSets.pop_back();
        }
        throw;
    }
}


void
MutableGlobalDefs::addLocation(std::unique_ptr< Location > item)
{
    trackType(LOCATION);

    mLocations->addItem(std::move(item));
}


void
MutableGlobalDefs::addLocationGroup(std::unique_ptr< LocationGroup > item)
{
    trackType(LOCATIONGROUP);

    mLocationGroups->addItem(std::move(item));
}


void
MutableGlobalDefs::addMetric(std::unique_ptr< Metric > item)
{
    trackType(METRIC);

    mMetrics->addItem(std::move(item));
}


void
MutableGlobalDefs::addRegion(std::unique_ptr< Region > item)
{
    trackType(REGION);

    mRegions->addItem(std::move(item));
}


void
MutableGlobalDefs::addRmaWindow(std::unique_ptr< RmaWindow > item)
{
    trackType(RMAWINDOW);

    mRmaWindows->addItem(std::move(item));
}


void
MutableGlobalDefs::addSourceLocation(std::unique_ptr< SourceLocation > item)
{
    trackType(SOURCELOCATION);

    mSourceLocations->addItem(std::move(item));
}


void
MutableGlobalDefs::addString(std::unique_ptr< String > item)
{
    trackType(STRING);

    mStrings->addItem(std::move(item));
}


void
MutableGlobalDefs::addSystemNode(std::unique_ptr< SystemNode > item)
{
    trackType(SYSTEMNODE);

    mSystemTree->addItem(std::move(item));
}


// --- Definition amendment functions ---------------------------------------

void
MutableGlobalDefs::assignCartCoordinate(const CartTopology::IdType cartTopoId,
                                        const uint32_t             rank,
                                        const CartCoordinate&      coordinate)
{
    MutableCartTopology* topology;
    try
    {
        CartTopology* temp = mCartTopologies->getItemById(cartTopoId);

        topology = static_cast< MutableCartTopology* >(temp);
    }
    catch (invalid_argument& ex)
    {
        throw RuntimeError("pearl::detail::MutableGlobalDefs::"
                           "assignCartCoordinate(CartTopology::IdType, "
                           "uint32_t, const CartCoordinate&) -- "
                           "invalid topology ID: " + std::to_string(cartTopoId));
    }

    topology->addRankCoordinate(rank, coordinate);
}


void
MutableGlobalDefs::assignCartCommunicator(const CartTopology::IdType cartTopoId,
                                          const Communicator&        communicator)
{
    MutableCartTopology* topology;
    try
    {
        CartTopology* temp = mCartTopologies->getItemById(cartTopoId);

        topology = static_cast< MutableCartTopology* >(temp);
    }
    catch (invalid_argument& ex)
    {
        throw RuntimeError("pearl::detail::MutableGlobalDefs::"
                           "assignCartCommunicator(CartTopology::IdType, "
                           "const Communicator&) -- "
                           "invalid topology ID: " + std::to_string(cartTopoId));
    }

    topology->setCommunicator(communicator);
}


// --- Query functions for communication location set definitions -----------

const CommLocationSet&
MutableGlobalDefs::getCommLocationSet(const Paradigm paradigm) const
{
    CommLocationSetContainer::const_iterator it = mCommLocationSets.begin();
    while (  (it != mCommLocationSets.end())
          && ((*it)->getParadigm() != paradigm))
    {
        ++it;
    }

    if (it == mCommLocationSets.end())
    {
        throw RuntimeError("pearl::detail::MutableGlobalDefs::"
                           "getCommLocationSet(Paradigm) -- "
                           "no communication location set defined for: "
                           + toString(paradigm));
    }

    return *(*it);
}


// --- Internal methods -----------------------------------------------------

void
MutableGlobalDefs::validate()
{
    ensureUniqueIds(*mContextTree, "CallingContext");
    ensureUniqueIds(*mCallsites, "Callsite");
    ensureUniqueIds(*mCartDimensions, "CartDimension");
    ensureUniqueIds(*mCartTopologies, "CartTopology");
    ensureUniqueIds(*mCommunicators, "Communicator");
    ensureUniqueIds(*mGroupingSets, "GroupingSet");
    ensureUniqueIds(*mLocations, "Location");
    ensureUniqueIds(*mLocationGroups, "LocationGroup");
    ensureUniqueIds(*mMetrics, "Metric");
    ensureUniqueIds(*mRegions, "Region");
    ensureUniqueIds(*mRmaWindows, "RmaWindow");
    ensureUniqueIds(*mSourceLocations, "SourceLocation");
    ensureUniqueIds(*mStrings, "String");
    ensureUniqueIds(*mSystemTree, "SystemNode");

    // Verify that at most one communication location set exists per paradigm
    map< Paradigm, bool > isUsed;
    for (CommLocationSetContainer::iterator it = mCommLocationSets.begin();
         it != mCommLocationSets.end();
         ++it)
    {
        const Paradigm paradigm = (*it)->getParadigm();
        if (isUsed.find(paradigm) != isUsed.end())
        {
            throw RuntimeError("pearl::detail::MutableGlobalDefs::"
                               "validate() -- "
                               "duplicate communication location set for: "
                               + toString(paradigm));
        }
        isUsed[paradigm] = true;
    }
}


/// @todo Refactor, then remove optimizeLookups() calls already triggered by
///       validate()
void
MutableGlobalDefs::setup()
{
    mContextTree->optimizeLookups();
    mCallsites->optimizeLookups();
    mCartDimensions->optimizeLookups();
    mCartTopologies->optimizeLookups();
    mCommunicators->optimizeLookups();
    mGroupingSets->optimizeLookups();
    mLocations->optimizeLookups();
    mLocationGroups->optimizeLookups();
    mMetrics->optimizeLookups();
    mRegions->optimizeLookups();
    mRmaWindows->optimizeLookups();
    mSourceLocations->optimizeLookups();
    mStrings->optimizeLookups();
    mSystemTree->optimizeLookups();

    // Determine next String identifier
    String::IdType nextStringId = 0;
    if (!mStrings->isEmpty())
    {
        const size_t lastIndex = mStrings->numItems() - 1;

        nextStringId = mStrings->getItemByIndex(lastIndex)->getId() + 1;
    }

    // Determine next Region identifier
    Region::IdType nextRegionId = 0;
    if (!mRegions->isEmpty())
    {
        const size_t lastIndex = mRegions->numItems() - 1;

        nextRegionId = mRegions->getItemByIndex(lastIndex)->getId() + 1;
    }

    // Create special "MEASUREMENT OFF" region if necessary
    if (!mPausingRegion)
    {
        // Create "MEASUREMENT OFF" string
        auto pauseStr =
            cxxslb::make_unique< String >(nextStringId++,
                                          "MEASUREMENT OFF");

        // Create "MEASUREMENT OFF" region
        auto region =
            cxxslb::make_unique< Region >(nextRegionId++,
                                          *pauseStr,
                                          *pauseStr,
                                          String::UNDEFINED,
                                          Region::ROLE_ARTIFICIAL,
                                          Paradigm::USER,
                                          String::UNDEFINED,
                                          PEARL_NO_NUM,
                                          PEARL_NO_NUM);
        mPausingRegion = region.get();
        addString(std::move(pauseStr));
        addRegion(std::move(region));
    }

    // Create special "TRACE BUFFER FLUSH" region if necessary
    if (!mFlushingRegion)
    {
        // Create "TRACE BUFFER FLUSH" string
        auto flushStr =
            cxxslb::make_unique< String >(nextStringId++,
                                          "TRACE BUFFER FLUSH");

        // Create "TRACE BUFFER FLUSH" region
        auto region =
            cxxslb::make_unique< Region >(nextRegionId++,
                                          *flushStr,
                                          *flushStr,
                                          String::UNDEFINED,
                                          Region::ROLE_ARTIFICIAL,
                                          Paradigm::MEASUREMENT_SYSTEM,
                                          String::UNDEFINED,
                                          PEARL_NO_NUM,
                                          PEARL_NO_NUM);
        mFlushingRegion = region.get();
        addString(std::move(flushStr));
        addRegion(std::move(region));
    }

    // Create special "TASKS" region if necessary
    if (!mTaskRootRegion)
    {
        // Create "TASKS" string
        auto taskRootStr =
            cxxslb::make_unique< String >(nextStringId++, "TASKS");

        // Create "TASKS" region
        auto region =
            cxxslb::make_unique< Region >(nextRegionId++,
                                          *taskRootStr,
                                          *taskRootStr,
                                          String::UNDEFINED,
                                          Region::ROLE_ARTIFICIAL,
                                          Paradigm::OPENMP,
                                          String::UNDEFINED,
                                          PEARL_NO_NUM,
                                          PEARL_NO_NUM);
        mTaskRootRegion = region.get();
        addString(std::move(taskRootStr));
        addRegion(std::move(region));
    }

    // Create special "THREADS" region if necessary
    if (!mThreadRootRegion)
    {
        // Create "THREADS" string
        auto threadRootStr =
            cxxslb::make_unique< String >(nextStringId++,
                                          "THREADS");

        // Create "THREADS" region
        auto region =
            cxxslb::make_unique< Region >(nextRegionId++,
                                          *threadRootStr,
                                          *threadRootStr,
                                          String::UNDEFINED,
                                          Region::ROLE_ARTIFICIAL,
                                          Paradigm::PTHREAD,
                                          *threadRootStr,
                                          PEARL_NO_NUM,
                                          PEARL_NO_NUM);
        mThreadRootRegion = region.get();
        addString(std::move(threadRootStr));
        addRegion(std::move(region));
    }

    // Reset call tree modification flag
    mCallTree->setModified(false);
}


// --- Set internal regions -------------------------------------------------

void
MutableGlobalDefs::setPausingRegion(Region* const region)
{
    if (mPausingRegion)
    {
        throw FatalError("Duplicate definition of PAUSING region!");
    }

    mPausingRegion = region;
}


void
MutableGlobalDefs::setFlushingRegion(Region* const region)
{
    if (mFlushingRegion)
    {
        throw FatalError("Duplicate definition of FLUSHING region!");
    }

    mFlushingRegion = region;
}


void
MutableGlobalDefs::setTaskRootRegion(Region* const region)
{
    if (mTaskRootRegion)
    {
        throw FatalError("Duplicate definition of Task Root region!");
    }

    mTaskRootRegion = region;
}


void
MutableGlobalDefs::setThreadRootRegion(Region* const region)
{
    if (mThreadRootRegion)
    {
        throw FatalError("Duplicate definition of THREADS Root region!");
    }

    mThreadRootRegion = region;
}


// --- Call tree handling ---------------------------------------------------

CallTree*
MutableGlobalDefs::getCallTree() const
{
    return mCallTree.get();
}


void
MutableGlobalDefs::setCallTree(CallTree* const callTree)
{
    if (callTree == mCallTree.get())
    {
        return;
    }

    mCallTree.reset(callTree);
}


// --- Global time offset handling ------------------------------------------

timestamp_t
MutableGlobalDefs::getGlobalOffset() const
{
    return mGlobalOffset;
}


void
MutableGlobalDefs::setGlobalOffset(const timestamp_t offset)
{
    mGlobalOffset = offset;
}


void
MutableGlobalDefs::setRealtimeOffset(const uint64_t offset)
{
    mRealtimeOffset = offset;
}


// --- Private methods ------------------------------------------------------

/// @brief Track definition type and optimize lookups for previous type.
///
/// Tracks the provided type @p currDefType of the current definition.  If
/// it differs from the type of the definition seen previously, optimizes
/// item lookups by global identifier for objects of the previous type.
///
/// @param currDefType
///     Type of the current definition
///
/// @exception_nothrow
///
void
MutableGlobalDefs::trackType(const DefinitionType currDefType)
{
    if (currDefType != mLastDefType)
    {
        // Optimize lookups for definitions of the previous type
        switch (mLastDefType)
        {
            // This is the initial value, i.e., we get here only for the first
            // definition object added.  In this case, there is nothing to be
            // optimized yet.
            case UNDEFINED:
                break;

            case CALLINGCONTEXT:
                mContextTree->optimizeLookups();
                break;

            // Call trees don't use the DefinitionTree class yet...
            case CALLPATH:
                break;

            case CALLSITE:
                mCallsites->optimizeLookups();
                break;

            case CARTDIMENSION:
                mCartDimensions->optimizeLookups();
                break;

            case CARTTOPOLOGY:
                mCartTopologies->optimizeLookups();
                break;

            case COMMUNICATOR:
                mCommunicators->optimizeLookups();
                break;

            case GROUPINGSET:
                mGroupingSets->optimizeLookups();
                break;

            case LOCATION:
                mLocations->optimizeLookups();
                break;

            case LOCATIONGROUP:
                mLocationGroups->optimizeLookups();
                break;

            case METRIC:
                mMetrics->optimizeLookups();
                break;

            case REGION:
                mRegions->optimizeLookups();
                break;

            case RMAWINDOW:
                mRmaWindows->optimizeLookups();
                break;

            case SOURCELOCATION:
                mSourceLocations->optimizeLookups();
                break;

            case STRING:
                mStrings->optimizeLookups();
                break;

            case SYSTEMNODE:
                mSystemTree->optimizeLookups();
                break;

            default:
                UTILS_BUG("Unknown definition type (%d)",
                          static_cast< int >(mLastDefType));
        }

        // Track definition type
        mLastDefType = currDefType;
    }
}
