/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2024                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_base
 *  @brief   Implementation of the class NonBlockingCollComplete_rep.
 *
 *  This file provides the implementation of the class
 *  NonBlockingCollComplete_rep.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include <pearl/NonBlockingCollComplete_rep.h>

#include <ostream>

#include <cxxslb/utility.h>

#include <pearl/Buffer.h>
#include <pearl/CommSet.h>
#include <pearl/Communicator.h>
#include <pearl/GlobalDefs.h>

#include "iomanip_detail.h"

using namespace std;
using namespace pearl;
using namespace pearl::detail;


// --- Constructors & destructor --------------------------------------------

NonBlockingCollComplete_rep::NonBlockingCollComplete_rep(const timestamp_t      timestamp,
                                                         const CollectiveOpType type,
                                                         const Communicator&    communicator,
                                                         const uint32_t         root,
                                                         const uint64_t         bytesSent,
                                                         const uint64_t         bytesReceived,
                                                         const uint64_t         requestId)
    : Event_rep(timestamp),
      mType(type),
      mCommunicator(communicator),
      mRoot(root),
      mBytesSent(bytesSent),
      mBytesReceived(bytesReceived),
      mRequestId(requestId),
      mPrevReqoffs(0)
{
}


NonBlockingCollComplete_rep::NonBlockingCollComplete_rep(const GlobalDefs& defs,
                                                         Buffer&           buffer)
    : Event_rep(defs, buffer),
      mType(CollectiveOpType(buffer.get_uint8())),
      mCommunicator(defs.getCommunicator(buffer.get_uint32())),
      mRoot(buffer.get_uint32()),
      mBytesSent(buffer.get_uint64()),
      mBytesReceived(buffer.get_uint64()),
      mRequestId(PEARL_NO_REQUEST),
      mPrevReqoffs(0)
{
}


// --- Event type information -----------------------------------------------

event_t
NonBlockingCollComplete_rep::getType() const
{
    return NON_BLOCKING_COLLECTIVE_COMPLETE;
}


bool
NonBlockingCollComplete_rep::isOfType(event_t type) const
{
    return (  (type == NON_BLOCKING_COLLECTIVE_COMPLETE)
           || (type == GROUP_NONBLOCK)
           || (type == GROUP_END)
           || (type == GROUP_ALL));
}


// --- Access event data ----------------------------------------------------

const Communicator&
NonBlockingCollComplete_rep::getComm() const
{
    return mCommunicator;
}


uint32_t
NonBlockingCollComplete_rep::getRoot() const
{
    return mRoot;
}


uint64_t
NonBlockingCollComplete_rep::getBytesSent() const
{
    return mBytesSent;
}


uint64_t
NonBlockingCollComplete_rep::getBytesReceived() const
{
    return mBytesReceived;
}


CollectiveOpType
NonBlockingCollComplete_rep::getCollType() const
{
    return mType;
}


uint64_t
NonBlockingCollComplete_rep::getRequestId() const
{
    return mRequestId;
}


// --- Serialize event data (protected) -------------------------------------

void
NonBlockingCollComplete_rep::pack(Buffer& buffer) const
{
    Event_rep::pack(buffer);

    buffer.put_uint8(cxxslb::to_underlying(mType));
    buffer.put_uint32(mCommunicator.getId());
    buffer.put_uint32(mRoot);
    buffer.put_uint64(mBytesSent);
    buffer.put_uint64(mBytesReceived);
}


// --- Generate human-readable output of event data (protected) -------------

ostream&
NonBlockingCollComplete_rep::output(ostream& stream) const
{
    Event_rep::output(stream);

    // Print data
    stream << sep
           << "type: " << mType
           << sep
           << "comm: " << mCommunicator
           << sep
           << "root: ";
    if (mRoot == PEARL_NO_ID)
    {
        stream << "n/a";
    }
    else
    {
        stream << mRoot
               << " (g: " << mCommunicator.getCommSet().getGlobalRank(mRoot) << ')';
    }
    stream << sep
           << "sent: " << mBytesSent
           << sep
           << "rcvd: " << mBytesReceived
           << sep
           << "request: " << mRequestId;

    return stream;
}


// --- Find next/prev request entries (protected) ---------------------------

uint32_t
NonBlockingCollComplete_rep::getPrevReqoffs() const
{
    return mPrevReqoffs;
}


uint32_t
NonBlockingCollComplete_rep::getNextReqoffs() const
{
    return 0;
}


void
NonBlockingCollComplete_rep::setPrevReqoffs(uint32_t offs)
{
    mPrevReqoffs = offs;
}


void
NonBlockingCollComplete_rep::setNextReqoffs(uint32_t ptr)
{
}
