/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2024                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_base
 *  @brief   Implementation of the class Otf2Archive.
 *
 *  This file provides the implementation of the class Otf2Archive and
 *  related functions.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include "Otf2Archive.h"

#include <UTILS_Error.h>

#include <pearl/Error.h>

#include "LocalIdMaps.h"
#include "MutableGlobalDefs.h"
#include "Otf2CollCallbacks.h"
#include "Otf2DefCallbacks.h"
#include "Otf2EvtCallbacks.h"
#include "Otf2LockCallbacks.h"
#include "pearl_ipc.h"

using namespace std;
using namespace pearl;
using namespace pearl::detail;


// --- Function prototypes --------------------------------------------------

namespace
{
void
registerDefCallbacks(OTF2_GlobalDefReaderCallbacks* callbacks);

void
registerEvtCallbacks(OTF2_EvtReaderCallbacks* callbacks);
}    // unnamed namespace


// --- Constructors & destructor --------------------------------------------

Otf2Archive::Otf2Archive(const string& anchorName,
                         const string& archiveDir)
    : TraceArchive(anchorName, archiveDir),
      mArchiveReader(nullptr),
      mTimerResolution(0),
      mGlobalOffset(0),
      mIdentifiedSelf(Communicator::NO_ID),
      mIdentifiedWorld(Communicator::NO_ID)
{
}


Otf2Archive::~Otf2Archive()
{
    if (mArchiveReader)
    {
        // Intentionally ignoring error code here since destructors should
        // *never* throw exceptions.
        OTF2_Reader_Close(mArchiveReader);
    }
}


// --- Private member functions ---------------------------------------------

void
Otf2Archive::openArchive()
{
    mArchiveReader = OTF2_Reader_Open(getAnchorName().c_str());
    if (!mArchiveReader)
    {
        throw RuntimeError("Error opening OTF2 experiment archive \""
                           + getAnchorName() + "\"!");
    }

    // Set locking callbacks
    static const OTF2_LockingCallbacks lockCbs = {
        nullptr,                        // .otf2_release
        pearl::detail::otf2LockCreate,  // .otf2_create
        pearl::detail::otf2LockDestroy, // .otf2_destroy
        pearl::detail::otf2LockAcquire, // .otf2_lock
        pearl::detail::otf2LockRelease  // .otf2_unlock
    };

    OTF2_ErrorCode result;
    result = OTF2_Reader_SetLockingCallbacks(mArchiveReader, &lockCbs, nullptr);
    if (result != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 locking callbacks!");
    }

    // Check OTF2 format version
    uint8_t major;
    uint8_t minor;
    uint8_t bugfix;
    result = OTF2_Reader_GetVersion(mArchiveReader, &major, &minor, &bugfix);
    if (result != OTF2_SUCCESS)
    {
        throw RuntimeError("Error determining OTF2 format version!");
    }
    if ((major == 1) && (minor < 2))
    {
        throw RuntimeError("OTF2 format version <1.2 not supported!");
    }
}


void
Otf2Archive::readDefinitions(GlobalDefs* defs)
{
    // Create definition reader
    OTF2_GlobalDefReader* defReader;
    defReader = OTF2_Reader_GetGlobalDefReader(mArchiveReader);
    if (!defReader)
    {
        throw RuntimeError("Error creating OTF2 global definition reader!");
    }

    // Create callback struct & register callbacks
    OTF2_GlobalDefReaderCallbacks* callbacks;
    callbacks = OTF2_GlobalDefReaderCallbacks_New();
    if (!callbacks)
    {
        throw RuntimeError("Error allocating OTF2 definition callback struct!");
    }
    registerDefCallbacks(callbacks);

    // Install callbacks
    MutableGlobalDefs* mutableDefs = dynamic_cast< MutableGlobalDefs* >(defs);
    UTILS_ASSERT(mutableDefs != nullptr);
    Otf2DefCbData  cbData(*mutableDefs);
    OTF2_ErrorCode result;
    result = OTF2_GlobalDefReader_SetCallbacks(defReader, callbacks, &cbData);
    if (result != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 definition callbacks!");
    }
    OTF2_GlobalDefReaderCallbacks_Delete(callbacks);

    // Read definition data
    uint64_t recordsRead;
    result = OTF2_GlobalDefReader_ReadDefinitions(defReader, UINT64_MAX, &recordsRead);
    if (  (result != OTF2_SUCCESS)
       || !cbData.mErrorMessage.empty())
    {
        throw RuntimeError(cbData.mErrorMessage);
    }

    // Store timer data
    mTimerResolution = cbData.mTimerResolution;
    mGlobalOffset    = cbData.mGlobalOffset;

    // Store IDs of COMM_SELF and COMM_WORLD
    mIdentifiedSelf  = cbData.mIdentifiedSelf;
    mIdentifiedWorld = cbData.mIdentifiedWorld;

    // Close definition reader
    result = OTF2_Reader_CloseGlobalDefReader(mArchiveReader, defReader);
    if (result != OTF2_SUCCESS)
    {
        throw RuntimeError("Error closing OTF2 global definition reader!");
    }

    // For OTF2 traces, global time offset subtraction is implicitly
    // applied during trace reading
    mutableDefs->setGlobalOffset(0.0);
}


void
Otf2Archive::openContainer(const LocationGroup& locGroup)
{
    static const OTF2_CollectiveCallbacks collCbs = {
        pearl::detail::otf2CollCbRelease,         // .otf2_release
        pearl::detail::otf2CollCbGetSize,         // .otf2_get_size
        pearl::detail::otf2CollCbGetRank,         // .otf2_get_rank
        pearl::detail::otf2CollCbCreateLocalComm, // .otf2_create_local_comm
        pearl::detail::otf2CollCbFreeLocalComm,   // .otf2_free_local_comm
        pearl::detail::otf2CollCbBarrier,         // .otf2_barrier
        pearl::detail::otf2CollCbBroadcast,       // .otf2_bcast
        pearl::detail::otf2CollCbGather,          // .otf2_gather
        pearl::detail::otf2CollCbGatherv,         // .otf2_gatherv
        pearl::detail::otf2CollCbScatter,         // .otf2_scatter
        pearl::detail::otf2CollCbScatterv         // .otf2_scatterv
    };

    // Install collective callbacks
    OTF2_ErrorCode result;
    result = OTF2_Reader_SetCollectiveCallbacks(mArchiveReader, &collCbs, nullptr,
                                                otf2CreateCollectiveContext(ipcCommWorld),
                                                nullptr);
    if (result != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 collective callbacks!");
    }

    // Select locations
    uint32_t count = locGroup.numLocations();
    for (uint32_t index = 0; index < count; ++index)
    {
        const Location& location = locGroup.getLocation(index);

        result = OTF2_Reader_SelectLocation(mArchiveReader, location.getId());
        if (result != OTF2_SUCCESS)
        {
            throw RuntimeError("Error selecting location for reading!");
        }
    }

    // Open containers
    result = OTF2_Reader_OpenDefFiles(mArchiveReader);
    if (result != OTF2_SUCCESS)
    {
        throw RuntimeError("Error opening OTF2 local definition container!");
    }
    result = OTF2_Reader_OpenEvtFiles(mArchiveReader);
    if (result != OTF2_SUCCESS)
    {
        throw RuntimeError("Error opening OTF2 trace data container!");
    }
}


void
Otf2Archive::closeContainer()
{
    // Close containers
    OTF2_ErrorCode result;
    result = OTF2_Reader_CloseEvtFiles(mArchiveReader);
    if (result != OTF2_SUCCESS)
    {
        throw RuntimeError("Error closing trace data container!");
    }
    result = OTF2_Reader_CloseDefFiles(mArchiveReader);
    if (result != OTF2_SUCCESS)
    {
        throw RuntimeError("Error closing local definition container!");
    }
}


/// @todo Suppress error messages thrown by OTF2 if optional local definition
///       file could not be found
LocalIdMaps*
Otf2Archive::readIdMaps(const Location& location)
{
    // Create local definition file reader
    OTF2_DefReader* defReader = OTF2_Reader_GetDefReader(mArchiveReader, location.getId());

    // NOTE:
    // The creation of the local definition reader may fail since local
    // definitions are optional. We therefore do not raise an exception
    // in this case.

    // Read local definition data
    if (defReader)
    {
        uint64_t       recordsRead;
        OTF2_ErrorCode result;
        result = OTF2_DefReader_ReadDefinitions(defReader, UINT64_MAX,
                                                &recordsRead);
        OTF2_Reader_CloseDefReader(mArchiveReader, defReader);
        if (result != OTF2_SUCCESS)
        {
            throw RuntimeError("Error reading OTF2 local definition data!");
        }
    }

    // Return dummy mapping data object
    return new LocalIdMaps;
}


void
Otf2Archive::readTrace(const GlobalDefs& defs,
                       const Location&   location,
                       LocalIdMaps*      idMaps,
                       LocalTrace*       trace)
{
    // Create event trace reader
    OTF2_EvtReader* evtReader = OTF2_Reader_GetEvtReader(mArchiveReader, location.getId());
    if (!evtReader)
    {
        throw RuntimeError("Error creating OTF2 local trace reader!");
    }

    // Create callback struct & register callbacks
    OTF2_EvtReaderCallbacks* callbacks = OTF2_EvtReaderCallbacks_New();
    if (!callbacks)
    {
        throw RuntimeError("Error allocating OTF2 event callback struct!");
    }
    registerEvtCallbacks(callbacks);

    // Install callbacks
    Otf2EvtCbData cbData(defs, *trace, mTimerResolution, mGlobalOffset,
                         mIdentifiedSelf, mIdentifiedWorld);
    OTF2_ErrorCode result;
    result = OTF2_EvtReader_SetCallbacks(evtReader, callbacks, &cbData);
    if (result != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 event callbacks!");
    }
    OTF2_EvtReaderCallbacks_Delete(callbacks);

    // Read event trace data
    uint64_t recordsRead;
    result = OTF2_EvtReader_ReadEvents(evtReader, UINT64_MAX, &recordsRead);
    if (  (result != OTF2_SUCCESS)
       || !cbData.mErrorMessage.empty())
    {
        throw RuntimeError(cbData.mErrorMessage);
    }
    OTF2_Reader_CloseEvtReader(mArchiveReader, evtReader);
}


// --- Helper functions -----------------------------------------------------

namespace
{
void
registerDefCallbacks(OTF2_GlobalDefReaderCallbacks* const callbacks)
{
    if (OTF2_GlobalDefReaderCallbacks_SetAttributeCallback(
            callbacks, otf2DefCbAttribute) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Attribute definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetCallingContextCallback(
            callbacks, otf2DefCbCallingContext) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 CallingContext definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetCallingContextPropertyCallback(
            callbacks, otf2DefCbCallingContextProperty) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 CallingContextProperty definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetCallpathCallback(
            callbacks, otf2DefCbCallpath) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Callpath definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetCallsiteCallback(
            callbacks, otf2DefCbCallsite) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Callsite definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetCartCoordinateCallback(
            callbacks, otf2DefCbCartCoordinate) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 CartCoordinate definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetCartDimensionCallback(
            callbacks, otf2DefCbCartDimension) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 CartDimension definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetCartTopologyCallback(
            callbacks, otf2DefCbCartTopology) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 CartTopology definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetClockPropertiesCallback(
            callbacks, otf2DefCbClockProperties) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ClockProperties definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetCommCallback(
            callbacks, otf2DefCbComm) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Comm definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetGroupCallback(
            callbacks, otf2DefCbGroup) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Group definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetInterCommCallback(
            callbacks, otf2DefCbInterComm) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 InterComm definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetInterruptGeneratorCallback(
            callbacks, otf2DefCbInterruptGenerator) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 InterruptGenerator definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetLocationCallback(
            callbacks, otf2DefCbLocation) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Location definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetLocationGroupCallback(
            callbacks, otf2DefCbLocationGroup) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 LocationGroup definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetLocationGroupPropertyCallback(
            callbacks, otf2DefCbLocationGroupProperty) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 LocationGroupProperty definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetLocationPropertyCallback(
            callbacks, otf2DefCbLocationProperty) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 LocationProperty definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetMetricClassCallback(
            callbacks, otf2DefCbMetricClass) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MetricClass definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetMetricClassRecorderCallback(
            callbacks, otf2DefCbMetricClassRecorder) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MetricClassRecorder definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetMetricInstanceCallback(
            callbacks, otf2DefCbMetricInstance) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MetricInstance definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetMetricMemberCallback(
            callbacks, otf2DefCbMetricMember) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MetricMember definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetParadigmCallback(
            callbacks, otf2DefCbParadigm) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Paradigm definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetParadigmPropertyCallback(
            callbacks, otf2DefCbParadigmProperty) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ParadigmProperty definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetParameterCallback(
            callbacks, otf2DefCbParameter) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Parameter definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetRegionCallback(
            callbacks, otf2DefCbRegion) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Region definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetRmaWinCallback(
            callbacks, otf2DefCbRmaWin) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaWin definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetSourceCodeLocationCallback(
            callbacks, otf2DefCbSourceCodeLocation) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 SourceCodeLocation definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetStringCallback(
            callbacks, otf2DefCbString) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 String definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetSystemTreeNodeCallback(
            callbacks, otf2DefCbSystemTreeNode) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 SystemTreeNode definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetSystemTreeNodeDomainCallback(
            callbacks, otf2DefCbSystemTreeNodeDomain) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 SystemTreeNodeDomain definition callback!");
    }
    if (OTF2_GlobalDefReaderCallbacks_SetSystemTreeNodePropertyCallback(
            callbacks, otf2DefCbSystemTreeNodeProperty) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 SystemTreeNodeProperty definition callback!");
    }
}


void
registerEvtCallbacks(OTF2_EvtReaderCallbacks* const callbacks)
{
    if (OTF2_EvtReaderCallbacks_SetBufferFlushCallback(
            callbacks, otf2EvtCbBufferFlush) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 BufferFlush event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetCallingContextEnterCallback(
            callbacks, otf2EvtCbCallingContextEnter) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 CallingContextEnter event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetCallingContextLeaveCallback(
            callbacks, otf2EvtCbCallingContextLeave) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 CallingContextLeave event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetCallingContextSampleCallback(
            callbacks, otf2EvtCbCallingContextSample) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 CallingContextSample event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetCommCreateCallback(
            callbacks, otf2EvtCbCommCreate) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 CommCreate event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetCommDestroyCallback(
            callbacks, otf2EvtCbCommDestroy) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 CommDestroy event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetEnterCallback(
            callbacks, otf2EvtCbEnter) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Enter event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetLeaveCallback(
            callbacks, otf2EvtCbLeave) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Leave event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMeasurementOnOffCallback(
            callbacks, otf2EvtCbMeasurementOnOff) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MeasurementOnOff event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMetricCallback(
            callbacks, otf2EvtCbMetric) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 Metric event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMpiCollectiveBeginCallback(
            callbacks, otf2EvtCbMpiCollectiveBegin) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MpiCollectiveBegin event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMpiCollectiveEndCallback(
            callbacks, otf2EvtCbMpiCollectiveEnd) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MpiCollectiveEnd event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMpiIrecvCallback(
            callbacks, otf2EvtCbMpiIrecv) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MpiIrecv event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMpiIrecvRequestCallback(
            callbacks, otf2EvtCbMpiIrecvRequest) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MpiIrecvRequest event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMpiIsendCallback(
            callbacks, otf2EvtCbMpiIsend) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MpiIsend event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMpiIsendCompleteCallback(
            callbacks, otf2EvtCbMpiIsendComplete) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MpiIsendComplete event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMpiRecvCallback(
            callbacks, otf2EvtCbMpiRecv) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MpiRecv event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMpiRequestCancelledCallback(
            callbacks, otf2EvtCbMpiRequestCancelled) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MpiRequestCancelled event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMpiRequestTestCallback(
            callbacks, otf2EvtCbMpiRequestTest) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MpiRequestTest event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetMpiSendCallback(
            callbacks, otf2EvtCbMpiSend) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 MpiSend event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetNonBlockingCollectiveCompleteCallback(
            callbacks, otf2EvtCbNonBlockingCollectiveComplete) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 NonBlockingCollectiveComplete callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetNonBlockingCollectiveRequestCallback(
            callbacks, otf2EvtCbNonBlockingCollectiveRequest) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 NonBlockingCollectiveRequest callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetParameterIntCallback(
            callbacks, otf2EvtCbParameterInt) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ParameterInt event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetParameterStringCallback(
            callbacks, otf2EvtCbParameterString) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ParameterString event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetParameterUnsignedIntCallback(
            callbacks, otf2EvtCbParameterUnsignedInt) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ParameterUnsignedInt event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetProgramBeginCallback(
            callbacks, otf2EvtCbProgramBegin) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ProgramBegin event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetProgramEndCallback(
            callbacks, otf2EvtCbProgramEnd) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ProgramEnd event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaAcquireLockCallback(
            callbacks, otf2EvtCbRmaAcquireLock) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaAcquireLock event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaAtomicCallback(
            callbacks, otf2EvtCbRmaAtomic) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaAtomic event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaCollectiveBeginCallback(
            callbacks, otf2EvtCbRmaCollectiveBegin) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaCollectiveBegin event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaCollectiveEndCallback(
            callbacks, otf2EvtCbRmaCollectiveEnd) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaCollectiveEnd event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaGetCallback(
            callbacks, otf2EvtCbRmaGet) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaGet event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaGroupSyncCallback(
            callbacks, otf2EvtCbRmaGroupSync) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaGroupSync event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaOpCompleteBlockingCallback(
            callbacks, otf2EvtCbRmaOpCompleteBlocking) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaOpCompleteBlocking event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaOpCompleteNonBlockingCallback(
            callbacks, otf2EvtCbRmaOpCompleteNonBlocking) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaOpCompleteNonBlocking event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaOpCompleteRemoteCallback(
            callbacks, otf2EvtCbRmaOpCompleteRemote) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaOpCompleteRemote event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaOpTestCallback(
            callbacks, otf2EvtCbRmaOpTest) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaOpTest event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaPutCallback(
            callbacks, otf2EvtCbRmaPut) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaPut event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaReleaseLockCallback(
            callbacks, otf2EvtCbRmaReleaseLock) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaReleaseLock event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaRequestLockCallback(
            callbacks, otf2EvtCbRmaRequestLock) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaRequestLock event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaSyncCallback(
            callbacks, otf2EvtCbRmaSync) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaSync event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaTryLockCallback(
            callbacks, otf2EvtCbRmaTryLock) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaTryLock event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaWaitChangeCallback(
            callbacks, otf2EvtCbRmaWaitChange) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaWaitChange event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaWinCreateCallback(
            callbacks, otf2EvtCbRmaWinCreate) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaWinCreate event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetRmaWinDestroyCallback(
            callbacks, otf2EvtCbRmaWinDestroy) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 RmaWinDestroy event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadAcquireLockCallback(
            callbacks, otf2EvtCbThreadAcquireLock) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadAcquireLock event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadBeginCallback(
            callbacks, otf2EvtCbThreadBegin) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadBegin event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadCreateCallback(
            callbacks, otf2EvtCbThreadCreate) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadCreate event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadEndCallback(
            callbacks, otf2EvtCbThreadEnd) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadEnd event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadForkCallback(
            callbacks, otf2EvtCbThreadFork) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadFork event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadJoinCallback(
            callbacks, otf2EvtCbThreadJoin) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadJoin event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadReleaseLockCallback(
            callbacks, otf2EvtCbThreadReleaseLock) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadReleaseLock event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadTaskCompleteCallback(
            callbacks, otf2EvtCbThreadTaskComplete) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadTaskComplete event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadTaskCreateCallback(
            callbacks, otf2EvtCbThreadTaskCreate) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadTaskCreate event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadTaskSwitchCallback(
            callbacks, otf2EvtCbThreadTaskSwitch) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadTaskSwitch event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadTeamBeginCallback(
            callbacks, otf2EvtCbThreadTeamBegin) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadTeamBegin event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadTeamEndCallback(
            callbacks, otf2EvtCbThreadTeamEnd) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadTeamEnd event callback!");
    }
    if (OTF2_EvtReaderCallbacks_SetThreadWaitCallback(
            callbacks, otf2EvtCbThreadWait) != OTF2_SUCCESS)
    {
        throw RuntimeError("Error setting OTF2 ThreadWait event callback!");
    }
}
}    // unnamed namespace
