/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2023                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_base
 *  @brief   Implementation of PEARL-internal error handling functionality.
 *
 *  This file provides the implementation of PEARL-internal functions to
 *  initialize and finalize a custom error handling.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include "pearl_error.h"

#include <cinttypes>
#include <cstdarg>
#include <cstdint>
#include <cstdio>
#include <cstring>
#include <sstream>
#include <string>

#include <UTILS_Debug.h>
#include <UTILS_Error.h>

#include "pearl_debug.h"
#include "pearl_ipc.h"

using namespace std;
using namespace pearl::detail;


// --- Local helper functions ----------------------------------------------

namespace
{
extern "C" SCALASCA_ErrorCode
errorCallback(void* const              userData,
              const char* const        file,
              const uint64_t           line,
              const char* const        function,
              const SCALASCA_ErrorCode errorCode,
              const char* const        msgFormatString,
              va_list                  va)
{
    // Construct message
    ostringstream message;
    if (errorCode == SCALASCA_WARNING)
    {
        message << "Warning";
    }
    else if (errorCode == SCALASCA_DEPRECATED)
    {
        message << "Deprecated";
    }
    else if (errorCode == SCALASCA_ABORT)
    {
        message <<  "Fatal";
    }
    else
    {
        message << "Error: "
                << SCALASCA_Error_GetDescription(errorCode);
    }
    if (  (msgFormatString != nullptr)
       && (strlen(msgFormatString) > 0))
    {
        // Determine number of characters that would be written, excluding
        // the terminating NUL.  Requires copying the va list, as it can
        // not be reused.
        va_list tmp_va;
        va_copy(tmp_va, va);
        size_t len = vsnprintf(nullptr, 0, msgFormatString, tmp_va);
        va_end(tmp_va);

        // Before C++17, overwriting the past-the-end NUL is undefined
        // behavior.  Thus, fiddle with the buffer size...
        string buffer(len + 1, '\0');
        vsnprintf(&buffer[0], len + 1, msgFormatString, va);
        buffer.resize(len);

        message << ": " << buffer;
    }

    // Print message, also on debug stream if requested.
    fprintf(stderr, "\n[%s] %s:%" PRIu64 ": %s\n",
            PACKAGE_NAME,
            file, line,
            message.str().c_str());
    #if HAVE(SCALASCA_DEBUG)
        if (UTILS_Debug_IsEnabled(SCALASCA_DEBUG_VERBOSE_ERRORS))
        {
            UTILS_Debug_Printf(0, AFS_PACKAGE_SRCDIR,
                               file, line, function,
                               "%s", message.str().c_str());
        }
    #endif    // HAVE(SCALASCA_DEBUG)

    // On fatal errors, try to clean up and abort (via the PEARL IPC layer)
    if (errorCode == SCALASCA_ABORT)
    {
        fprintf(stderr, "[%s] Please report this to <%s>.  Thank you!\n"
                        "[%s] See \"Reporting bugs\" in the User Guide for detailed instructions.\n",
                PACKAGE_NAME, PACKAGE_BUGREPORT,
                PACKAGE_NAME);

        va_end(va);
        #if HAVE(SCALASCA_DEBUG)
            debugFinalize();
        #endif    // HAVE(SCALASCA_DEBUG)
        ipcAbort();
    }

    return errorCode;
}
}    // unnamed namespace


namespace pearl
{
namespace detail
{
// --- Error handling functions ---------------------------------------------

void
errorInit()
{
    SCALASCA_Error_RegisterCallback(errorCallback, nullptr);
}


void
errorFinalize()
{
    SCALASCA_Error_RegisterCallback(nullptr, nullptr);
}
}    // namespace pearl::detail
}    // namespace pearl
