/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2022                                                     **
**  RWTH Aachen University, IT Center                                      **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_thread
 *  @brief   Implementation of the class pearl::detail::SmallBlockAllocator.
 *
 *  This file provides the implementation of the internal helper class
 *  pearl::detail::SmallBlockAllocator.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include "SmallBlockAllocator.h"

#include <algorithm>
#include <memory>
#include <new>

#include <UTILS_Error.h>

#include "ChunkAllocator.h"

using namespace std;
using namespace pearl::detail;


// --- Helpers --------------------------------------------------------------

namespace
{
// --- Constants -------------------------------------------

/// Number of required `%ChunkAllocator` instances
constexpr size_t NUM_ALLOCATORS = SmallBlockAllocator::MAX_BLOCKSIZE
                                  / SmallBlockAllocator::MIN_ALIGNMENT;


// --- Helper functions ------------------------------------

inline size_t
getAllocatorIndex(const size_t blockSize)
{
    const size_t index = ((blockSize + SmallBlockAllocator::MIN_ALIGNMENT - 1)
                          / SmallBlockAllocator::MIN_ALIGNMENT) - 1;
    UTILS_ASSERT(index < NUM_ALLOCATORS);

    return index;
}
}    // unnamed namespace


// --- Constructors & destructor --------------------------------------------

SmallBlockAllocator::SmallBlockAllocator()
    : mAllocators(NUM_ALLOCATORS, nullptr)
{
}


SmallBlockAllocator::~SmallBlockAllocator()
{
    std::for_each(mAllocators.begin(), mAllocators.end(),
                  std::default_delete< ChunkAllocator >());
}


// --- Memory allocation functions ------------------------------------------

void*
SmallBlockAllocator::allocate(const size_t blockSize)
{
    // Let the default allocator handle "extreme" cases
    if (  (blockSize == 0)
       || (blockSize > MAX_BLOCKSIZE))
    {
        return ::operator new(blockSize);
    }

    // Create chunk allocator if necessary
    const size_t index = getAllocatorIndex(blockSize);
    if (mAllocators[index] == nullptr)
    {
        mAllocators[index] = new ChunkAllocator((index + 1) * MIN_ALIGNMENT);
    }

    // Allocate block from chunk allocator
    return mAllocators[index]->allocate();
}


void
SmallBlockAllocator::deallocate(void* const       ptr,
                                const std::size_t blockSize)
{
    // Let the default allocator handle "extreme" cases
    if (  (ptr == nullptr)
       || (blockSize == 0)
       || (blockSize > MAX_BLOCKSIZE))
    {
        ::operator delete(ptr);

        return;
    }

    // Deallocate block to chunk allocator for block size
    const size_t index = getAllocatorIndex(blockSize);
    UTILS_ASSERT(mAllocators[index] != nullptr);
    mAllocators[index]->deallocate(ptr);
}
