/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2024                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2014                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief Implementation of the class LongTermDelayPattern.
 *
 *  This file provides the implementation of the class LongTermDelayPattern.
 **/
/*
 *  NOTE:
 *  Line numbers referenced in comments below refer to algorithm 1 on
 *  page 10 of Boehme et al., "Identifying the root causes of wait
 *  states in large-scale parallel applications", TOPC 3(2), article 11,
 *  July 2016.
 */
/*-------------------------------------------------------------------------*/


#include <config.h>

#include "LongTermDelayPattern.h"

#include <utility>

#include <UTILS_Error.h>

#include <pearl/Buffer.h>
#include <pearl/CallbackManager.h>
#include <pearl/GlobalDefs.h>
#include <pearl/LocalData.h>
#include <pearl/RemoteData.h>

#include "CbData.h"
#include "Roles.h"
#include "user_events.h"

using namespace std;
using namespace pearl;
using namespace scout;


// --- Constructors & destructor --------------------------------------------

LongTermDelayPattern::LongTermDelayPattern(const uint32_t bufferKey,
                                           const uint32_t rcostKey,
                                           const int      delayEvent)
    : mBufferKey(bufferKey),
      mRcostKey(rcostKey),
      mDelayEvent(delayEvent)
{
}


// --- Callback registration ------------------------------------------------

void
LongTermDelayPattern::registerCallbacks(CallbackManagerMap& cbmanagers)
{
    TypedPattern< double >::registerCallbacks(cbmanagers);

    // Determine "bwc" callback manager
    CallbackManagerMap::iterator it = cbmanagers.find("bwc");
    UTILS_ASSERT(it != cbmanagers.end());
    CallbackManagerMap::mapped_type cbmanager = it->second;

    #if defined(_MPI)
        // Register "MPI Late Sender" delay callbacks
        cbmanager->register_callback(
            PRE_RECV,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbMpiPreDelayLateSender));
        cbmanager->register_callback(
            DELAY_LATE_SENDER,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbMpiDelayLateSender));

        // Register "MPI Late Receiver" delay callbacks
        cbmanager->register_callback(
            PRE_INV_SENDCMP,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbMpiPreDelayLateReceiver));
        cbmanager->register_callback(
            DELAY_LATE_RECEIVER,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbMpiDelayLateReceiver));

        // Register "MPI Wait at Barrier" delay callbacks
        cbmanager->register_callback(
            PRE_DELAY_MPI_BARRIER,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbPreDelayCollective));
        cbmanager->register_callback(
            DELAY_MPI_BARRIER,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbDelayCollective));

        // Register "MPI Late Broadcast" delay callbacks
        cbmanager->register_callback(
            PRE_DELAY_MPI_COLL_12N,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbPreDelayCollective));
        cbmanager->register_callback(
            DELAY_MPI_COLL_12N,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbDelayCollective));

        // Register "MPI Early Reduce" delay callbacks
        cbmanager->register_callback(
            PRE_DELAY_MPI_COLL_N21,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbPreDelayCollective));
        cbmanager->register_callback(
            DELAY_MPI_COLL_N21,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbDelayCollective));

        // Register "MPI Wait at NxN" delay callbacks
        cbmanager->register_callback(
            PRE_DELAY_MPI_COLL_N2N,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbPreDelayCollective));
        cbmanager->register_callback(
            DELAY_MPI_COLL_N2N,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbDelayCollective));

        // Register "MPI Wait at Finalize" delay callbacks
        cbmanager->register_callback(
            PRE_DELAY_MPI_FINALIZE,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbPreDelayCollective));
        cbmanager->register_callback(
            DELAY_MPI_FINALIZE,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbDelayCollective));

        // Register "MPI Wait at Communicator Create" delay callbacks
        cbmanager->register_callback(
            PRE_DELAY_MPI_COLL_COMM_CREATE,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbPreDelayCollective));
        cbmanager->register_callback(
            DELAY_MPI_COLL_COMM_CREATE,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbDelayCollective));

        // Register "MPI Wait at Communicator Free" delay callbacks
        cbmanager->register_callback(
            PRE_DELAY_MPI_COLL_COMM_FREE,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbPreDelayCollective));
        cbmanager->register_callback(
            DELAY_MPI_COLL_COMM_FREE,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbDelayCollective));
    #endif    // _MPI

    #if defined(_OPENMP)
        // Register "OpenMP Wait at Barrier" delay callbacks
        cbmanager->register_callback(
            PRE_DELAY_OMP_BARRIER,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbPreDelayCollective));
        cbmanager->register_callback(
            DELAY_OMP_BARRIER,
            PEARL_create_callback(
                this,
                &LongTermDelayPattern::cbDelayCollective));
    #endif    // _OPENMP
}


// --- Callback data functions (bwc) ----------------------------------------

#if defined(_MPI)
    void
    LongTermDelayPattern::cbMpiPreDelayLateSender(const CallbackManager& cbmanager,
                                                  const int              userEvent,
                                                  const Event&           event,
                                                  CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        preDelayMpiPt2pt(cbmanager, userEvent, event, data->mLocal);
    }


    void
    LongTermDelayPattern::cbMpiDelayLateSender(const CallbackManager& cbmanager,
                                               const int              userEvent,
                                               const Event&           event,
                                               CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        delayMpiPt2pt(cbmanager, userEvent, event, data, data->mRemote);
    }


    void
    LongTermDelayPattern::cbMpiPreDelayLateReceiver(const CallbackManager& cbmanager,
                                                    const int              userEvent,
                                                    const Event&           event,
                                                    CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        preDelayMpiPt2pt(cbmanager, userEvent, event, data->mInvLocal);
    }


    void
    LongTermDelayPattern::cbMpiDelayLateReceiver(const CallbackManager& cbmanager,
                                                 const int              userEvent,
                                                 const Event&           event,
                                                 CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        delayMpiPt2pt(cbmanager, userEvent, event, data, data->mInvRemote);
    }
#endif    // _MPI


void
LongTermDelayPattern::cbPreDelayCollective(const CallbackManager& cbmanager,
                                           const int              userEvent,
                                           const Event&           event,
                                           CallbackData* const    cdata)
{
    CbData* const data = static_cast< CbData* >(cdata);

    // Setup long-term cost transfer (line 13)
    data->mInputScales.insert(make_pair(mRcostKey, mCost.getCosts(event)));
}


void
LongTermDelayPattern::cbDelayCollective(const CallbackManager& cbmanager,
                                        const int              userEvent,
                                        const Event&           event,
                                        CallbackData* const    cdata)
{
    CbData* const data = static_cast< CbData* >(cdata);

    // Update delay costs (line 25)
    // mCollDelayInfo (calculated during the cost transfer) contains
    // gamma_long * s * delta_p[c]
    map< uint32_t, TimeMapSum >::const_iterator dCit = data->mCollDelayInfo.find(mRcostKey);
    if (  (dCit != data->mCollDelayInfo.end())
       && (dCit->second.mSum > 0.0))
    {
        for (timemap_t::const_iterator it = dCit->second.mMap.begin();
             it != dCit->second.mMap.end();
             ++it)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                             data->mDefs->get_cnode(it->first),
                             it->second);
        }
    }

    // Update long-term costs for propagating wait states in synchronization
    // interval (line 28); split into two terms
    // mSumScales (calculated during the cost transfer) contains
    // gamma_(long|short) * s, multiplied by waiting time in addCosts()
    //
    // Add term for gamma_long
    map< uint32_t, ScaleMap >::const_iterator sCit = data->mSumScales.find(mRcostKey);
    if (sCit != data->mSumScales.end())
    {
        mCost.addCosts(data->mPropSpMap, sCit->second);
    }

    // Add term for gamma_short only when processing a synchronization point
    // of the wait state type related to this long-term delay pattern, as
    // gamma_short is zero otherwise.
    if (userEvent == mDelayEvent)
    {
        map< uint32_t, ScaleMap >::const_iterator sWit = data->mSumScales.find(KEY_WAITTIME);
        if (sWit != data->mSumScales.end())
        {
            mCost.addCosts(data->mPropSpMap, sWit->second);
        }
    }
}


// --- Internal helper functions --------------------------------------------

#if defined(_MPI)
    /// @todo [C++11] Use `std::unique_ptr` for `buffer`
    void
    LongTermDelayPattern::preDelayMpiPt2pt(const CallbackManager& cbmanager,
                                           const int              userEvent,
                                           const Event&           event,
                                           LocalData* const       localData)
    {
        // Setup long-term cost transfer (line 13)
        Buffer* const buffer = new Buffer(16);
        buffer->put_double(mCost.getCosts(event));

        localData->addBuffer(buffer, mBufferKey);
    }


    void
    LongTermDelayPattern::delayMpiPt2pt(const CallbackManager& cbmanager,
                                        const int              userEvent,
                                        const Event&           event,
                                        CallbackData* const    cdata,
                                        RemoteData* const      remoteData)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        // Shortcuts for difference profile and waiting time;
        // both contain only positive contributions
        const TimeMapSum& delta_p(data->mDelayInfo);
        const TimeMapSum& wait_t(data->mWaitInfo);

        // Skip calculation if there is no excess workload
        if (!((delta_p.mSum + wait_t.mSum) > 0))
        {
            return;
        }

        // Finish long-term cost transfer (line 18)
        const double gamma_long = remoteData->getBuffer(mBufferKey)->get_double();

        // Update delay costs (line 25)
        double scale = gamma_long / (delta_p.mSum + wait_t.mSum);
        if (scale > 0)
        {
            for (timemap_t::const_iterator it = delta_p.mMap.begin();
                 it != delta_p.mMap.end();
                 ++it)
            {
                INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                 data->mDefs->get_cnode(it->first),
                                 scale * it->second);
            }
        }

        // Update long-term costs for propagating wait states in synchronization
        // interval (line 28)
        //
        // Add gamma_short term (gamma_short * s) to scale (gamma_long * s)
        // only when processing a synchronization point of the wait state
        // type related to this long-term delay pattern, as gamma_short is
        // zero otherwise.
        if (userEvent == mDelayEvent)
        {
            scale += std::min(data->mRwait / (delta_p.mSum + wait_t.mSum), 1.0);
        }
        mCost.addCosts(data->mPropSpMap, scale);
    }
#endif    // _MPI
