/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2021-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief Declaration of the class LongTermDelayPattern.
 *
 *  This header file provides the declaration of the class
 *  LongTermDelayPattern, which serves as an abstract base class for all
 *  long-term delay cost performance properties calculated by SCOUT.
 **/
/*-------------------------------------------------------------------------*/


#ifndef SCOUT_LONGTERMDELAYPATTERN_H
#define SCOUT_LONGTERMDELAYPATTERN_H


#include <cstdint>

#include "DelayCostMap.h"
#include "TypedPattern.h"


// --- Forward declarations -------------------------------------------------

namespace pearl
{
class CallbackData;
class CallbackManager;
class Event;
class LocalData;
class RemoteData;
}    // namespace pearl


namespace scout
{
/*-------------------------------------------------------------------------*/
/**
 *  @class LongTermDelayPattern
 *  @brief Abstract base class for all long-term delay cost properties
 *         calculated by SCOUT.
 *
 *  The LongTermDelayPattern class serves as an abstract base class for all
 *  long-term delay cost performance properties calculated by SCOUT.
 **/
/*-------------------------------------------------------------------------*/

class LongTermDelayPattern
    : public TypedPattern< double >
{
    public:
        // --- Public methods ------------------------------

        /// @name Constructors & destructor
        /// @{

        /// @brief Constructor.
        ///
        /// Creates a new instance using the given key to store delay costs.
        ///
        /// @param bufferKey
        ///     Key under which delay costs are stored in point-to-point
        ///     transfer buffers
        /// @param rcostKey
        ///     Key under which input delay and scale values are stored
        /// @param delayEvent
        ///     DELAY event corresponding to this long-term delay pattern
        ///
        LongTermDelayPattern(std::uint32_t bufferKey,
                             std::uint32_t rcostKey,
                             int           delayEvent);

        /// @}
        /// @name Callback registration
        /// @{

        /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
        ///
        virtual void
        registerCallbacks(CallbackManagerMap& cbmanagers);

        /// @}


    protected:
        // --- Data members --------------------------------

        /// Maps long-term delay costs onto syncpoints in the current
        /// synchronization interval
        DelayCostMap mCost;


    private:
        // --- Data members --------------------------------

        /// Key under which delay costs are stored in point-to-point transfer
        /// buffers
        std::uint32_t mBufferKey;

        /// Key under which input delay and scale values are stored
        std::uint32_t mRcostKey;

        /// DELAY event corresponding to this long-term delay pattern
        int mDelayEvent;


        // --- Private methods -----------------------------

        /// @name Callback functions (bwc)
        /// @{

        /// @brief Pre-delay callback for *MPI Late Sender* wait states.
        ///
        /// Pre-delay callback function for *MPI Late Sender* wait states
        /// triggered by PRE_RECV user events before the delay data exchange
        /// during the critical-path and delay analysis replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param userEvent
        ///     User event triggering this callback
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        cbMpiPreDelayLateSender(const pearl::CallbackManager& cbmanager,
                                int                           userEvent,
                                const pearl::Event&           event,
                                pearl::CallbackData*          cdata);

        /// @brief Delay callback for *MPI Late Sender* wait states.
        ///
        /// Delay callback function for *MPI Late Sender* wait states
        /// triggered by DELAY_LATE_SENDER user events after the delay data
        /// exchange during the critical-path and delay analysis replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param userEvent
        ///     User event triggering this callback
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        cbMpiDelayLateSender(const pearl::CallbackManager& cbmanager,
                             int                           userEvent,
                             const pearl::Event&           event,
                             pearl::CallbackData*          cdata);

        /// @brief Pre-delay callback for *MPI Late Receiver* wait states.
        ///
        /// Pre-delay callback function for *MPI Late Receiver* wait states
        /// triggered by PRE_INV_SENDCMP user events before the delay data
        /// exchange during the critical-path and delay analysis replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param userEvent
        ///     User event triggering this callback
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        cbMpiPreDelayLateReceiver(const pearl::CallbackManager& cbmanager,
                                  int                           userEvent,
                                  const pearl::Event&           event,
                                  pearl::CallbackData*          cdata);

        /// @brief Delay callback for *MPI Late Recaiver* wait states.
        ///
        /// Delay callback function for *MPI Late Receiver* wait states
        /// triggered by DELAY_LATE_RECEIVER user events after the delay data
        /// exchange during the critical-path and delay analysis replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param userEvent
        ///     User event triggering this callback
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        cbMpiDelayLateReceiver(const pearl::CallbackManager& cbmanager,
                               int                           userEvent,
                               const pearl::Event&           event,
                               pearl::CallbackData*          cdata);

        /// @brief Pre-delay callback for MPI/OpenMP collective wait states.
        ///
        /// Pre-delay callback function for all MPI/OpenMP collective wait
        /// states triggered by PRE_DELAY_collective user events before the
        /// delay data exchange during the critical-path and delay analysis
        /// replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param userEvent
        ///     User event triggering this callback
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        cbPreDelayCollective(const pearl::CallbackManager& cbmanager,
                             int                           userEvent,
                             const pearl::Event&           event,
                             pearl::CallbackData*          cdata);

        /// @brief Delay callback for MPI/OpenMP collective wait states.
        ///
        /// Delay callback function for all MPI/OpenMP collective wait states
        /// triggered by DELAY_collective user events before the delay data
        /// exchange during the critical-path and delay analysis replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param userEvent
        ///     User event triggering this callback
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        cbDelayCollective(const pearl::CallbackManager& cbmanager,
                          int                           userEvent,
                          const pearl::Event&           event,
                          pearl::CallbackData*          cdata);

        /// @}
        /// @name Internal helper functions
        /// @{

        /// @brief Pre-delay callback helper for MPI point-to-point wait states.
        ///
        /// Parameterized pre-delay callback helper for all MPI point-to-point
        /// wait states triggered by the corresponding *Late Sender/Receiver*
        /// pre-delay callback.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param userEvent
        ///     Currently processed user event
        /// @param event
        ///     Currently processed event
        /// @param localData
        ///     Corresponding local pattern constituent container
        ///
        void
        preDelayMpiPt2pt(const pearl::CallbackManager& cbmanager,
                         int                           userEvent,
                         const pearl::Event&           event,
                         pearl::LocalData*             localData);

        /// @brief Delay callback helper for MPI point-to-point wait states.
        ///
        /// Parameterized delay callback helper for all MPI point-to-point
        /// wait states triggered by the corresponding *Late Sender/Receiver*
        /// delay callback.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param userEvent
        ///     Currently processed user event
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        /// @param remoteData
        ///     Corresponding remote pattern constituent container
        ///
        void
        delayMpiPt2pt(const pearl::CallbackManager& cbmanager,
                      int                           userEvent,
                      const pearl::Event&           event,
                      pearl::CallbackData*          cdata,
                      pearl::RemoteData*            remoteData);

        /// @}
};
}    // namespace scout


#endif    // !SCOUT_LONGTERMDELAYPATTERN_H
