/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2024                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2014                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  Copyright (c) 2019-2023                                                **
**  RWTH Aachen University, IT Center                                      **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "MpiCommunicationHandler.h"

#include <algorithm>
#include <cstddef>
#include <cstdint>
#include <cstring>
#include <limits>
#include <map>
#include <vector>

#include <cxxslb/memory.h>
#include <cxxslb/utility.h>

#define SCALASCA_DEBUG_MODULE_NAME    SCOUT
#include <UTILS_Debug.h>
#include <UTILS_Error.h>
#include <mpi.h>

#include <pearl/AmRuntime.h>
#include <pearl/CallbackManager.h>
#include <pearl/CommSet.h>
#include <pearl/Communicator.h>
#include <pearl/Event.h>
#include <pearl/EventSet.h>
#include <pearl/GlobalDefs.h>
#include <pearl/LocalData.h>
#include <pearl/MpiCollEnd_rep.h>
#include <pearl/MpiComm.h>
#include <pearl/MpiGroup.h>
#include <pearl/MpiHelpers.h>
#include <pearl/MpiMessage.h>
#include <pearl/MpiReceive_rep.h>
#include <pearl/MpiSend_rep.h>
#include <pearl/Region.h>
#include <pearl/RemoteData.h>
#include <pearl/RemoteEventSet.h>
#include <pearl/RmaWindow.h>

#include "Callstack.h"
#include "CbData.h"
#include "MpiNonBlockingBarrier.h"
#include "Roles.h"
#include "Settings.h"
#include "SynchpointHandler.h"
#include "scout_types.h"
#include "user_events.h"

using namespace scout;
using namespace pearl;
using namespace std;


namespace scout
{
namespace detail
{
static uint8_t
numBitsSet(uint8_t byte)
{
    static bool    is_initialized = false;
    static uint8_t bitSetTable[256];

    if (!is_initialized)
    {
        bitSetTable[0] = 0;
        for (int i = 0; i < 256; i++)
        {
            bitSetTable[i] = (i & 1) + bitSetTable[i / 2];
        }
        is_initialized = true;
    }

    return bitSetTable[byte];
}
}    // namespace scout::detail
}    // namespace scout


struct MpiCommunicationHandler::MpiCHImpl
{
    //
    // --- private data --------------------------------------------------
    //

    typedef map< Event, CollectiveInfo, EventKeyCompare > collinfo_map_t;
    typedef map< Communicator::IdType, MpiComm* >         communicator_map_t;

    collinfo_map_t mCollInfo;

    vector< MpiMessage* > mPendingMsgs;    ///< Pending message objects (main replay)
    vector< MPI_Request > mPendingReqs;    ///< Pending MPI requests (main replay)
    vector< MPI_Status >  mStatuses;
    vector< int >         mIndices;

    communicator_map_t mInvComms;

    // Offsets into RMA analysis window
    static const int LOCAL_POST     = 0;
    static const int LAST_POST      = 1 * sizeof(TimeRank);
    static const int LAST_COMPLETE  = 2 * sizeof(TimeRank);
    static const int LAST_RMA_OP    = 3 * sizeof(TimeRank);
    static const int BITFIELD_START = 4 * sizeof(TimeRank);

    static const TimeRank UNDEFINED_TIME_RANK;

    typedef map< uint32_t, pearl::timestamp_t > TimeMap;
    typedef map< uint32_t, TimeRank >           TimeRankMap;
    typedef struct
    {
        uint8_t*               mBuffer;
        uint8_t                mRankMask;
        MPI_Aint               mMaskOffset;
        size_t                 mBufferSize;
        size_t                 mBitfieldSize;
        pearl::timestamp_t     mCompleteEnterTime;
        MPI_Win                mWin;
        TimeRankMap            mLastOpTime;
        TimeMap                mPostTime;
        SynchpointInfo         mLastPostSyncPoint;
        MpiNonBlockingBarrier* mBarrier;
        set< uint32_t >        mAccessProcesses;
    } WindowInfo;
    typedef std::map< const RmaWindow*, WindowInfo > WindowInfoMap;

    WindowInfoMap mWindows;

    ~MpiCHImpl()
    {
        for (communicator_map_t::iterator it = mInvComms.begin();
             it != mInvComms.end();
             ++it)
        {
            delete it->second;
        }
    }

    //
    // --- helper functions ----------------------------------------------
    //

    void
    processPendingMsgs()
    {
        // No pending requests? ==> continue
        if (mPendingReqs.empty())
        {
            return;
        }

        // --- Check for completed messages

        int completed;
        int count = mPendingReqs.size();

        mIndices.resize(count);
        mStatuses.resize(count);

        MPI_Testsome(count, &mPendingReqs[0], &completed, &mIndices[0], &mStatuses[0]);


        // --- Update array of pending messages

        for (int i = 0; i < completed; ++i)
        {
            int index = mIndices[i];

            delete mPendingMsgs[index];
            mPendingMsgs[index] = nullptr;
        }

        if (completed > 0)
        {
            mPendingMsgs.erase(remove(mPendingMsgs.begin(), mPendingMsgs.end(),
                                      static_cast< MpiMessage* >(nullptr)),
                               mPendingMsgs.end());
            mPendingReqs.erase(remove(mPendingReqs.begin(), mPendingReqs.end(),
                                      static_cast< MPI_Request >(MPI_REQUEST_NULL)),
                               mPendingReqs.end());
        }
    }

    void
    processPendingActiveMsgs(const RmaWindow& window)
    {
        auto& winInfo = mWindows[&window];
        auto& runtime = AmRuntime::getInstance();
        auto& comm    = window.getCommunicator();

        unsigned int global_pending = 0;

        do
        {
            // complete all local requests
            runtime.fence(comm);

            // wait until all processes completed local requests
            winInfo.mBarrier->start();
            while (!winInfo.mBarrier->test())
            {
                runtime.advance();
            }

            // check if response requests were scheduled anywhere
            unsigned int local_pending = runtime.numPending(comm);
            MPI_Allreduce(&local_pending, &global_pending, 1, MPI_UNSIGNED, MPI_SUM, getMpiHandle(comm));
        }
        while (global_pending > 0);
    }

    /// @todo Avoid calling getLocalRank() on every invocation; it's \f$O(n)\f$
    void
    setupCollectiveData(const pearl::Event& event,
                        CbData* const       data,
                        const Communicator& comm) const
    {
        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_TRIGGER) << event;

        const Event enter = data->mCallstack->top();
        const Event begin = event.beginptr();
        const Event leave = event.leaveptr();

        data->mLocal->addEvent(enter, ROLE_ENTER_COLL);
        data->mLocal->addEvent(begin, ROLE_BEGIN_COLL);
        data->mLocal->addEvent(event, ROLE_END_COLL);
        data->mLocal->addEvent(leave, ROLE_LEAVE_COLL);

        CollectiveInfo& ci = data->mCollinfo;

        ci.my.mTime = begin->getTimestamp();
        ci.my.mRank = comm.getCommSet().getLocalRank(event.get_location().getRank());
    }

    void
    setRemoteAccessBit(const pearl::Event& event,
                       CbData*             data)
    {
        auto&       window  = event->getWindow();
        WindowInfo& winInfo = mWindows[&window];

        /* set bit corresponding to my rank on every remote location */
        for (set< uint32_t >::iterator it = winInfo.mAccessProcesses.begin();
             it != winInfo.mAccessProcesses.end();
             ++it)
        {
            MPI_Accumulate(&winInfo.mRankMask, 1, SCALASCA_MPI_UINT8_T,
                           static_cast< int >(*it), winInfo.mMaskOffset, 1,
                           SCALASCA_MPI_UINT8_T, MPI_BOR, winInfo.mWin);
        }
    }

    void
    setRemoteLastOp(const pearl::Event& event,
                    CbData*             data)
    {
        auto&       window  = event->getWindow();
        WindowInfo& winInfo = mWindows[&window];

        /* send last rma op times to remote processes */
        TimeRankMap::iterator it = winInfo.mLastOpTime.begin();
        while (it != winInfo.mLastOpTime.end())
        {
            MPI_Accumulate(&(it->second), 1, MPI_DOUBLE_INT,
                           it->first, LAST_RMA_OP,
                           1, MPI_DOUBLE_INT, MPI_MAXLOC, winInfo.mWin);
            ++it;
        }
    }

    //
    // --- main replay callbacks -----------------------------------------
    //

    void
    cb_finished(const CallbackManager& cbmanager,
                const int              user_event,
                const Event&           event,
                CallbackData* const    cdata)
    {
        MPI_Barrier(MPI_COMM_WORLD);

        processPendingMsgs();

        // Handle remaining messages
        if (!mPendingReqs.empty())
        {
            int count = mPendingReqs.size();

            // Print warning
            UTILS_WARNING("Encountered %d unreceived send operations!", count);

            // Cancel pending messages
            for (int i = 0; i < count; ++i)
            {
                MPI_Cancel(&mPendingReqs[i]);
                mPendingMsgs[i]->wait();
                delete mPendingMsgs[i];
            }
        }

        mPendingMsgs.clear();
        mPendingReqs.clear();
    }

    //
    // --- point-to-point communication
    //

    void
    cb_process_msgs(const CallbackManager& cbmanager,
                    const int              user_event,
                    const Event&           event,
                    CallbackData* const    cdata)
    {
        processPendingMsgs();
    }

    void
    cb_pre_send(const CallbackManager& cbmanager,
                const int              user_event,
                const Event&           event,
                CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        if (event.completion()->isOfType(MPI_CANCELLED))
        {
            return;
        }

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_TRIGGER) << event;

        cbmanager.notify(PRE_SEND, event, data);

        const MpiSend_rep& sendRep = event_cast< MpiSend_rep >(*event);
        auto* const        msg     = data->mLocal->isend(sendRep.getComm(),
                                                         sendRep.getDestination(),
                                                         sendRep.getTag());
        mPendingMsgs.push_back(msg);
        mPendingReqs.push_back(msg->get_request());

        cbmanager.notify(POST_SEND, event, data);
    }

    void
    cb_pre_recv(const CallbackManager& cbmanager,
                const int              user_event,
                const Event&           event,
                CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_TRIGGER) << event;

        cbmanager.notify(PRE_RECV, event, data);

        const MpiReceive_rep& recvRep = event_cast< MpiReceive_rep >(*event);
        data->mRemote->recv(*data->mDefs,
                            recvRep.getComm(),
                            recvRep.getSource(),
                            recvRep.getTag());

        cbmanager.notify(POST_RECV, event, data);
    }

    void
    cb_pre_send_setup(const CallbackManager& cbmanager,
                      const int              user_event,
                      const Event&           event,
                      CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        const Event enter = data->mCallstack->top();
        const Event leave = event.leaveptr();

        data->mLocal->addEvent(enter, ROLE_ENTER_SEND);
        data->mLocal->addEvent(event, ROLE_SEND);
        data->mLocal->addEvent(leave, ROLE_LEAVE_SEND);

        // Track communicators used in point-to-point communication.  The
        // placeholder will be updated when preparing the "bwc" replay.
        const MpiSend_rep& sendRep = event_cast< const MpiSend_rep >(*event);
        mInvComms.emplace(sendRep.getComm().getId(), nullptr);
    }

    void
    cb_pre_recv_setup(const CallbackManager& cbmanager,
                      const int              user_event,
                      const Event&           event,
                      CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        const Event enter = data->mCallstack->top();
        const Event leave = event.leaveptr();

        data->mLocal->addEvent(enter, ROLE_ENTER_RECV);
        data->mLocal->addEvent(event, ROLE_RECV);
        data->mLocal->addEvent(leave, ROLE_LEAVE_RECV);

        // Track communicators used in point-to-point communication.  The
        // placeholder will be updated when preparing the "bwc" replay.
        const MpiReceive_rep& recvRep = event_cast< const MpiReceive_rep >(*event);
        mInvComms.emplace(recvRep.getComm().getId(), nullptr);
    }

    //
    // --- mpi collective communication
    //

    // --- collective communication dispatch

    void
    cb_pre_mpi_collend(const CallbackManager& cbmanager,
                       const int              user_event,
                       const Event&           event,
                       CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        const MpiCollEnd_rep& endRep = event_cast< MpiCollEnd_rep >(*event);

        // Collectives on self-like communicators do not synchronize
        auto& comm = endRep.getComm();
        if (comm.getSize() == 1)
        {
            return;
        }

        // Dispatch
        const CollectiveOpType collType = endRep.getCollType();
        switch (collType)
        {
            // CASE 1: Barrier
            case CollectiveOpType::BARRIER:
                setupCollectiveData(event, data, comm);
                cbmanager.notify(PRE_SYNC_COLL, event, data);
                cbmanager.notify(SYNC_COLL, event, data);
                break;

            // CASE 2: 1-to-N communication
            case CollectiveOpType::BCAST:
            case CollectiveOpType::SCATTER:
            case CollectiveOpType::SCATTERV:
                {
                    uint64_t bytesSent = event->getBytesSent();
                    MPI_Bcast(&bytesSent, 1, SCALASCA_MPI_UINT64_T,
                              event->getRoot(), getMpiHandle(comm));

                    // No payload -> no syncpoint -> no waitstate
                    if (bytesSent == 0)
                    {
                        return;
                    }

                    setupCollectiveData(event, data, comm);
                    cbmanager.notify(PRE_COLL_12N, event, data);
                    cbmanager.notify(COLL_12N, event, data);
                }
                break;

            // CASE 3: N-to-N communication w/o MPI_Alltoallv & MPI_Alltoallw
            case CollectiveOpType::ALLGATHER:
            case CollectiveOpType::ALLGATHERV:
            case CollectiveOpType::ALLREDUCE:
            case CollectiveOpType::ALLTOALL:
            case CollectiveOpType::REDUCE_SCATTER:
            case CollectiveOpType::REDUCE_SCATTER_BLOCK:
                {
                    const uint64_t bytesSent     = event->getBytesSent();
                    const uint64_t bytesReceived = event->getBytesReceived();

                    // No payload -> no syncpoint -> no waitstate
                    if ((bytesSent == 0) && (bytesReceived == 0))
                    {
                        return;
                    }

                    setupCollectiveData(event, data, comm);
                    cbmanager.notify(PRE_COLL_N2N, event, data);
                    cbmanager.notify(COLL_N2N, event, data);
                }
                break;

            // CASE 4: N-to-1 communication
            case CollectiveOpType::GATHER:
            case CollectiveOpType::GATHERV:
            case CollectiveOpType::REDUCE:
                {
                    uint64_t bytesReceived = event->getBytesReceived();
                    MPI_Bcast(&bytesReceived, 1, SCALASCA_MPI_UINT64_T,
                              event->getRoot(), getMpiHandle(comm));

                    // No payload -> no syncpoint -> no waitstate
                    if (bytesReceived == 0)
                    {
                        return;
                    }

                    setupCollectiveData(event, data, comm);
                    cbmanager.notify(PRE_COLL_N21, event, data);
                    cbmanager.notify(COLL_N21, event, data);
                }
                break;

            // CASE 5: Prefix reduction
            case CollectiveOpType::EXSCAN:
            case CollectiveOpType::SCAN:
                {
                    const uint64_t bytesSent     = event->getBytesSent();
                    const uint64_t bytesReceived = event->getBytesReceived();

                    // No payload -> no syncpoint -> no waitstate
                    if (  (bytesSent == 0)
                       && (bytesReceived == 0))
                    {
                        return;
                    }

                    setupCollectiveData(event, data, comm);
                    cbmanager.notify(PRE_COLL_SCAN, event, data);
                    cbmanager.notify(COLL_SCAN, event, data);
                }
                break;

            // CASE 6: Communicator creation
            case CollectiveOpType::CREATE_HANDLE:
                {
                    const Event   enter  = data->mCallstack->top();
                    const Region& region = enter->getRegion();

                    setupCollectiveData(event, data, comm);

                    // Waiting time in MPI_Init calls is covered in the MPI
                    // Initialization Completion pattern
                    if (is_mpi_init(region))
                    {
                        cbmanager.notify(PRE_INIT, event, data);
                        cbmanager.notify(INIT, event, data);

                        break;
                    }
                    cbmanager.notify(PRE_COLL_COMM_CREATE, event, data);
                    cbmanager.notify(COLL_COMM_CREATE, event, data);
                }
                break;

            // CASE 7: Communicator destruction
            case CollectiveOpType::DESTROY_HANDLE:
                {
                    const Event   enter  = data->mCallstack->top();
                    const Region& region = enter->getRegion();

                    setupCollectiveData(event, data, comm);

                    // Waiting time in MPI_Finalize calls is covered in the MPI
                    // Wait at Finalize pattern
                    if (is_mpi_finalize(region))
                    {
                        cbmanager.notify(PRE_FINALIZE, event, data);
                        cbmanager.notify(FINALIZE, event, data);

                        break;
                    }

                    cbmanager.notify(PRE_COLL_COMM_FREE, event, data);
                    cbmanager.notify(COLL_COMM_FREE, event, data);
                }
                break;

            // MPI_Alltoallv and MPI_Alltoallw can be used to implement
            // arbitrary communication patterns (e.g., pair-wise message
            // exchange).  However, the trace data currently does not
            // provide enough information for a proper reconstruction,
            // thus, ignore those calls.
            //
            // Likewise, ignore events using collective operations not
            // handled above.
            default:
                return;
        }

        mCollInfo.insert(make_pair(event, data->mCollinfo));
    }

    // --- mpi collective operation handlers

    void
    cb_pre_coll_12n(const CallbackManager& cbmanager,
                    const int              user_event,
                    const Event&           event,
                    CallbackData* const    cdata)
    {
        auto* const data       = static_cast< CbData* >(cdata);
        auto&       comm       = event->getComm();
        MPI_Comm    commHandle = getMpiHandle(comm);

        CollectiveInfo& ci = data->mCollinfo;

        // Distribute root BEGIN
        ci.root = ci.my;
        MPI_Bcast(&ci.root, 1, MPI_DOUBLE_INT, event->getRoot(),
                  commHandle);

        // Earliest END of receivers waiting for root
        const uint64_t bytesReceived = event->getBytesReceived();
        const bool     waitsForRoot  = (  (bytesReceived > 0)
                                       && (ci.my.mTime < ci.root.mTime));

        TimeRank sendBuf(event->getTimestamp(), ci.my.mRank);
        if (!waitsForRoot)
        {
            sendBuf.mTime = numeric_limits< double >::max();
        }
        MPI_Allreduce(&sendBuf, &ci.earliest_end, 1, MPI_DOUBLE_INT, MPI_MINLOC,
                      commHandle);

        data->mIsSyncpoint = (  waitsForRoot
                             || (  (ci.root.mRank == ci.my.mRank)
                                && (ci.earliest_end.mTime < numeric_limits< double >::max())));

        if (data->mIsSyncpoint)
        {
            UTILS_DLOG_LEVEL(SCALASCA_DEBUG_WAITSTATE_METRICS)
                << "root BEGIN: " << globalize(ci.root, comm)
                << ", earliest sync END: " << globalize(ci.earliest_end, comm);
        }
    }

    void
    cb_pre_coll_n21(const CallbackManager& cbmanager,
                    const int              user_event,
                    const Event&           event,
                    CallbackData* const    cdata)
    {
        auto* const data       = static_cast< CbData* >(cdata);
        auto&       comm       = event->getComm();
        MPI_Comm    commHandle = getMpiHandle(comm);

        CollectiveInfo& ci = data->mCollinfo;

        // Distribute root BEGIN
        ci.root = ci.my;
        MPI_Bcast(&ci.root, 1, MPI_DOUBLE_INT, event->getRoot(),
                  commHandle);

        // Latest BEGIN of all senders
        const uint64_t bytesSent = event->getBytesSent();

        TimeRank sendBuf = ci.my;
        if (bytesSent == 0)
        {
            sendBuf.mTime = std::numeric_limits< double >::lowest();
        }
        MPI_Allreduce(&sendBuf, &ci.latest, 1, MPI_DOUBLE_INT, MPI_MAXLOC,
                      commHandle);

        // The latest sender and the root rank are considered synchronizing
        // if the root enters the operation earlier than the sender.
        data->mIsSyncpoint = (  (ci.root.mTime < ci.latest.mTime)
                             && (  (ci.my.mRank == ci.root.mRank)
                                || (ci.my.mRank == ci.latest.mRank)));

        // Earliest END of synchronizing ranks (i.e., root and latest sender)
        //
        // Although only two ranks are involved and interested in the result,
        // a collective operation over all involved ranks is used here.  A
        // point-to-point transfer (e.g., via MPI_Sendrecv) may interfere
        // with other non-blocking point-to-point message transfers currently
        // in flight unless a separate communicator or unique tag (which can
        // not be guaranteed) is used.
        sendBuf.mTime = numeric_limits< double >::max();
        if (data->mIsSyncpoint)
        {
            sendBuf.mTime = event->getTimestamp();
        }
        MPI_Allreduce(&sendBuf, &ci.earliest_end, 1, MPI_DOUBLE_INT, MPI_MINLOC,
                      commHandle);

        if (data->mIsSyncpoint)
        {
            UTILS_DLOG_LEVEL(SCALASCA_DEBUG_WAITSTATE_METRICS)
                << "root BEGIN: " << globalize(ci.root, comm)
                << ", latest sync BEGIN: " << globalize(ci.latest, comm)
                << ", earliest sync END: " << globalize(ci.earliest_end, comm);
        }
    }

    void
    cb_pre_coll_scan(const CallbackManager& cbmanager,
                     const int              user_event,
                     const Event&           event,
                     CallbackData* const    cdata)
    {
        auto* const data = static_cast< CbData* >(cdata);
        auto&       comm = event->getComm();

        CollectiveInfo& ci = data->mCollinfo;

        // Latest BEGIN of all senders < own rank
        const uint64_t bytesSent = event->getBytesSent();

        TimeRank sendBuf = ci.my;
        if (bytesSent == 0)
        {
            sendBuf.mTime = std::numeric_limits< double >::lowest();
        }
        MPI_Scan(&sendBuf, &ci.latest, 1, MPI_DOUBLE_INT, MPI_MAXLOC,
                 getMpiHandle(comm));

        // FIXME: Currently only covers the receiver side
        data->mIsSyncpoint = (ci.my.mTime < ci.latest.mTime);
    }

    void
    cb_pre_coll_n2n(const CallbackManager& cbmanager,
                    const int              user_event,
                    const Event&           event,
                    CallbackData* const    cdata)
    {
        auto* const data       = static_cast< CbData* >(cdata);
        auto&       comm       = event->getComm();
        MPI_Comm    commHandle = getMpiHandle(comm);

        CollectiveInfo& ci = data->mCollinfo;

        // --- Step 1: check for full overlap

        // Latest BEGIN of all ranks
        TimeRank sendBuf = ci.my;
        MPI_Allreduce(&sendBuf, &ci.latest, 1, MPI_DOUBLE_INT, MPI_MAXLOC,
                      commHandle);


        // Earliest END of all ranks
        sendBuf.mTime = event->getTimestamp();
        MPI_Allreduce(&sendBuf, &ci.earliest_end, 1, MPI_DOUBLE_INT, MPI_MINLOC,
                      commHandle);

        if (ci.earliest_end.mTime > ci.latest.mTime)
        {
            // With full overlap, all ranks are considered to be synchronizing
            data->mIsSyncpoint = true;

            UTILS_DLOG_LEVEL(SCALASCA_DEBUG_WAITSTATE_METRICS)
                << "latest BEGIN: " << globalize(ci.latest, comm)
                << ", earliest END: " << globalize(ci.earliest_end, comm)
                << " (full overlap)";

            return;
        }

        // --- Step 2: otherwise, handle partial overlap

        const uint64_t bytesSent     = event->getBytesSent();
        const uint64_t bytesReceived = event->getBytesReceived();

        // Latest BEGIN of all senders
        sendBuf.mTime = ci.my.mTime;
        if (bytesSent == 0)
        {
            sendBuf.mTime = std::numeric_limits< double >::lowest();
        }
        MPI_Allreduce(&sendBuf, &ci.latest, 1, MPI_DOUBLE_INT, MPI_MAXLOC,
                      commHandle);

        // With partial overlap, only receivers overlapping with the latest
        // sender and the latest sender itself are considered synchronizing
        data->mIsSyncpoint = (  (  (bytesReceived > 0)
                                && (ci.my.mTime < ci.latest.mTime))
                             || (  (bytesSent > 0)
                                && (ci.my.mRank == ci.latest.mRank)));

        // Earliest END of synchronizing ranks
        sendBuf.mTime = event->getTimestamp();
        if (!data->mIsSyncpoint)
        {
            sendBuf.mTime = numeric_limits< double >::max();
        }
        MPI_Allreduce(&sendBuf, &ci.earliest_end, 1, MPI_DOUBLE_INT, MPI_MINLOC,
                      commHandle);

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_WAITSTATE_METRICS)
            << "latest send BEGIN: " << globalize(ci.latest, comm)
            << ", earliest sync END: " << globalize(ci.earliest_end, comm)
            << " (partial overlap)";
    }

    void
    cb_pre_coll_comm_create(const CallbackManager& cbmanager,
                            const int              user_event,
                            const Event&           event,
                            CallbackData* const    cdata)
    {
        auto* const data       = static_cast< CbData* >(cdata);
        auto&       comm       = event->getComm();
        MPI_Comm    commHandle = getMpiHandle(comm);

        CollectiveInfo& ci = data->mCollinfo;

        // Latest BEGIN of all ranks in communicator
        TimeRank sendBuf = ci.my;
        MPI_Allreduce(&sendBuf, &ci.latest, 1, MPI_DOUBLE_INT, MPI_MAXLOC,
                      commHandle);

        // Earliest END of all ranks in communicator
        sendBuf.mTime = event->getTimestamp();
        MPI_Allreduce(&sendBuf, &ci.earliest_end, 1, MPI_DOUBLE_INT, MPI_MINLOC,
                      commHandle);

        // Communicator creation calls, e.g., MPI_COMM_SPLIT, synchronize
        // all processes in the communicator specified in the MPI_COLLECTIVE_END
        // event before the new communicators are created.
        data->mIsSyncpoint = true;

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_WAITSTATE_METRICS)
            << "latest BEGIN: " << globalize(ci.latest, comm)
            << ", earliest END: " << globalize(ci.earliest_end, comm)
            << " (full overlap)";
    }

    void
    cb_pre_coll_comm_free(const CallbackManager& cbmanager,
                          const int              user_event,
                          const Event&           event,
                          CallbackData* const    cdata)
    {
        auto* const data       = static_cast< CbData* >(cdata);
        auto&       comm       = event->getComm();
        MPI_Comm    commHandle = getMpiHandle(comm);

        CollectiveInfo& ci = data->mCollinfo;

        // Latest BEGIN of all ranks in communicator
        TimeRank sendBuf = ci.my;
        MPI_Allreduce(&sendBuf, &ci.latest, 1, MPI_DOUBLE_INT, MPI_MAXLOC,
                      commHandle);

        // Earliest END of all ranks in communicator
        sendBuf.mTime = event->getTimestamp();
        MPI_Allreduce(&sendBuf, &ci.earliest_end, 1, MPI_DOUBLE_INT, MPI_MINLOC,
                      commHandle);

        // Communicator destruction calls do not need to synchronize. As
        // a heuristic, only consider them synchronizing if there is full
        // overlap.
        if (ci.earliest_end.mTime <= ci.latest.mTime)
        {
            return;
        }

        data->mIsSyncpoint = true;

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_WAITSTATE_METRICS)
            << "latest BEGIN: " << globalize(ci.latest, comm)
            << ", earliest END: " << globalize(ci.earliest_end, comm)
            << " (full overlap)";
    }

    void
    cb_pre_sync_coll(const CallbackManager& cbmanager,
                     const int              user_event,
                     const Event&           event,
                     CallbackData* const    cdata)
    {
        auto* const data       = static_cast< CbData* >(cdata);
        auto&       comm       = event->getComm();
        MPI_Comm    commHandle = getMpiHandle(comm);

        CollectiveInfo& ci = data->mCollinfo;

        // All ranks are synchronizing at a Barrier
        data->mIsSyncpoint = true;

        // latest BEGIN
        TimeRank sendBuf = ci.my;
        MPI_Allreduce(&sendBuf, &ci.latest, 1, MPI_DOUBLE_INT, MPI_MAXLOC,
                      commHandle);

        // earliest END
        sendBuf.mTime = event->getTimestamp();
        MPI_Allreduce(&sendBuf, &ci.earliest_end, 1, MPI_DOUBLE_INT, MPI_MINLOC,
                      commHandle);

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_WAITSTATE_METRICS)
            << "latest BEGIN: " << globalize(ci.latest, comm)
            << ", earliest END: " << globalize(ci.earliest_end, comm)
            << " (full overlap)";
    }

    // --- rma collective operation handlers

    void
    cb_pre_rma_collend(const CallbackManager& cbmanager,
                       const int              user_event,
                       const Event&           event,
                       CallbackData* const    cdata)
    {
        auto* const data   = static_cast< CbData* >(cdata);
        auto&       window = event->getWindow();
        auto&       comm   = window.getCommunicator();

        // FIXME: This does not work due to the way callbacks are currently
        //        triggered, but should be reenabled after some refactoring.
        //
        // A single process will not wait for itself
        // if (comm.getSize() < 2)
        // {
        //     return;
        // }

        Event begin = event.beginptr();
        Event leave = event.leaveptr();
        Event enter = data->mCallstack->top();

        data->mLocal->addEvent(enter, ROLE_ENTER_COLL);
        data->mLocal->addEvent(begin, ROLE_BEGIN_COLL);
        data->mLocal->addEvent(event, ROLE_END_COLL);
        data->mLocal->addEvent(leave, ROLE_LEAVE_COLL);

        CollectiveInfo ci;

        ci.my.mTime = begin->getTimestamp();
        ci.my.mRank = comm.getCommSet().getLocalRank(event.get_location().getRank());

        data->mCollinfo = ci;

        // Process active messages
        const Region& region = enter->getRegion();
        if (  enableAsynchronous
           && !is_mpi_rma_create(region))
        {
            processPendingActiveMsgs(window);
        }

        // Compute potential waiting time
        cbmanager.notify(RMA_SYNC_COLLECTIVE, event, data);

        // Dispatch
        if (is_mpi_rma_fence(region))
        {
            cbmanager.notify(MPI_RMA_WIN_FENCE, event, data);
        }
        else if (is_mpi_rma_create(region))
        {
            cbmanager.notify(MPI_RMA_WIN_CREATE, event, data);
        }
        else if (is_mpi_rma_free(region))
        {
            cbmanager.notify(MPI_RMA_WIN_FREE, event, data);
        }

        mCollInfo.insert(make_pair(event, data->mCollinfo));
    }

    void
    cb_pre_rma_sync_coll(const CallbackManager& cbmanager,
                         const int              user_event,
                         const Event&           event,
                         CallbackData* const    cdata)
    {
        auto* const data       = static_cast< CbData* >(cdata);
        MPI_Comm    commHandle = getMpiHandle(event->getWindow().getCommunicator());

        CollectiveInfo& ci(data->mCollinfo);

        TimeRank trbuf;

        // latest BEGIN
        trbuf = ci.my;

        MPI_Allreduce(&trbuf, &ci.latest, 1, MPI_DOUBLE_INT, MPI_MAXLOC, commHandle);

        // earliest END
        trbuf.mTime = event->getTimestamp();
        trbuf.mRank = ci.my.mRank;

        MPI_Allreduce(&trbuf, &ci.earliest_end, 1, MPI_DOUBLE_INT, MPI_MINLOC, commHandle);
    }

    void
    cb_pre_rma_win_create(const CallbackManager& cbmanager,
                          const int              user_event,
                          const Event&           event,
                          CallbackData* const    cdata)
    {
        auto&    window     = event->getWindow();
        auto&    comm       = window.getCommunicator();
        MPI_Comm commHandle = getMpiHandle(comm);

        WindowInfo& winInfo = mWindows[&window];

        // allocate & initialize window buffer
        #if !defined(SCOUT_DISABLE_RMA_ACCESS_TRACKING)
            const int comm_size = comm.getSize();
            winInfo.mBitfieldSize = ((comm_size % 8) ? 1 : 0) + (comm_size / 8);
        #else
            winInfo.mBitfieldSize = 0;
        #endif
        winInfo.mBufferSize = BITFIELD_START + winInfo.mBitfieldSize;
        winInfo.mBuffer     = new uint8_t[winInfo.mBufferSize];

        // initialize TimeRank part of buffer
        TimeRank* timerank = reinterpret_cast< TimeRank* >(winInfo.mBuffer);

        timerank[LAST_COMPLETE / sizeof(TimeRank)] = UNDEFINED_TIME_RANK;
        timerank[LAST_RMA_OP   / sizeof(TimeRank)] = UNDEFINED_TIME_RANK;
        timerank[LAST_POST     / sizeof(TimeRank)] = UNDEFINED_TIME_RANK;
        timerank[LOCAL_POST    / sizeof(TimeRank)] = TimeRank(event.enterptr()->getTimestamp(), -1);

        // setup access mask & offset
        int myrank = comm.getCommSet().getLocalRank(event.get_location().getRank());
        winInfo.mRankMask   = 1 << (myrank % 8);
        winInfo.mMaskOffset = BITFIELD_START + (myrank / 8);

        MPI_Win_create(winInfo.mBuffer, winInfo.mBufferSize, sizeof(uint8_t),
                       MPI_INFO_NULL, commHandle, &winInfo.mWin);

        if (enableAsynchronous)
        {
            // initialize non-blocking barrier
            winInfo.mBarrier = new MpiNonBlockingBarrier(commHandle);
            winInfo.mBarrier->init();
            #if defined(DEBUG_AM)
                UTILS_DLOG << "Initialized non-blocking barrier on communicator "
                           << comm.getId();
            #endif

            // start listening on window communicator
            AmRuntime::getInstance().startListen(window.getCommunicator());
            #if defined(DEBUG_AM)
                UTILS_DLOG << "Listening for AM requests on communicator "
                           << window.getCommunicator().getId()
                           << " (" << window.getCommunicator() << ")"
                           << " of window " << window.getId();
            #endif
        }
    }

    void
    cb_pre_rma_win_fence(const CallbackManager& cbmanager,
                         const int              user_event,
                         const Event&           event,
                         CallbackData* const    cdata)
    {
        CbData*     data    = static_cast< CbData* >(cdata);
        auto&       window  = event->getWindow();
        WindowInfo& winInfo = mWindows[&window];

        // initialize TimeRank part of buffer
        TimeRank* timerank = reinterpret_cast< TimeRank* >(winInfo.mBuffer);
        UTILS_ASSERT(timerank != nullptr);

        timerank[LAST_COMPLETE / sizeof(TimeRank)] = UNDEFINED_TIME_RANK;
        timerank[LAST_RMA_OP   / sizeof(TimeRank)] = UNDEFINED_TIME_RANK;
        timerank[LAST_POST     / sizeof(TimeRank)] = UNDEFINED_TIME_RANK;
        timerank[LOCAL_POST    / sizeof(TimeRank)] = TimeRank(event.enterptr()->getTimestamp(), -1);

        #if !defined(SCOUT_DISABLE_RMA_ACCESS_TRACKING)
            memset(winInfo.mBuffer + BITFIELD_START, 0, winInfo.mBitfieldSize);
        #endif

        /* get latest rma op to this location */
        MPI_Win_fence(MPI_MODE_NOPRECEDE, winInfo.mWin);

        #if !defined(SCOUT_DISABLE_RMA_ACCESS_TRACKING)
            setRemoteAccessBit(event, data);
        #endif
        setRemoteLastOp(event, data);

        MPI_Win_fence(MPI_MODE_NOPUT | MPI_MODE_NOSUCCEED | MPI_MODE_NOSTORE, winInfo.mWin);

        // get last RMA Operation to local process
        data->mLastRmaOp.mTime = timerank[LAST_RMA_OP / sizeof(TimeRank)].mTime;

        data->mCount = 0;
        #if !defined(SCOUT_DISABLE_RMA_ACCESS_TRACKING)
            /* add own accesses to access tracking bitfield */
            set< uint32_t >::iterator target_rank = winInfo.mAccessProcesses.begin();
            while (target_rank != winInfo.mAccessProcesses.end())
            {
                winInfo.mBuffer[BITFIELD_START + (*target_rank / 8)] |= *target_rank % 8;
                ++target_rank;
            }
            winInfo.mAccessProcesses.clear();

            /* calculate number of accesses */
            for (size_t i = BITFIELD_START;
                 i < (BITFIELD_START + winInfo.mBitfieldSize);
                 ++i)
            {
                data->mCount += detail::numBitsSet(winInfo.mBuffer[i]);
            }
        #endif
    }

    void
    cb_pre_rma_win_free(const CallbackManager& cbmanager,
                        const int              user_event,
                        const Event&           event,
                        CallbackData* const    cdata)
    {
        auto&       window  = event->getWindow();
        WindowInfo& winInfo = mWindows[&window];

        // free window for active target exchange
        MPI_Win_free(&winInfo.mWin);
        delete[] winInfo.mBuffer;

        if (enableAsynchronous)
        {
            AmRuntime& runtime = AmRuntime::getInstance();

            // cleanup barrier
            winInfo.mBarrier->finalize();
            delete winInfo.mBarrier;
            #if defined(DEBUG_AM)
                auto& comm = window.getCommunicator();
                UTILS_DLOG << "Finalized non-blocking barrier on communicator "
                           << comm.getId();
            #endif

            runtime.stopListen(window.getCommunicator());
        }

        mWindows.erase(&window);
    }

    // --- rma group synchronization  handlers -----------------------

    void
    cb_pre_mpi_win_start(const CallbackManager& cbmanager,
                         const int              user_event,
                         const Event&           event,
                         CallbackData* const    cdata)
    {
        auto&       window  = event->getWindow();
        auto&       comm    = window.getCommunicator();
        auto&       group   = event->getGroup();
        WindowInfo& winInfo = mWindows[&window];

        /* start access epoch to exchange timestamps */
        MPI_Win_start(group.getHandle(), 0, winInfo.mWin);

        /* get post enter-time from each exposure epoch */
        for (uint32_t rank = 0; rank < group.numRanks(); ++rank)
        {
            /* create entry for location */
            timestamp_t& postTime = winInfo.mPostTime[ comm.getCommSet().getLocalRank(group.getGlobalRank(rank)) ];

            MPI_Get(&(postTime), 1, MPI_DOUBLE,
                    comm.getCommSet().getLocalRank(group.getGlobalRank(rank)), LOCAL_POST,
                    1, MPI_DOUBLE, winInfo.mWin);
        }
    }

    void
    cb_pre_mpi_win_complete(const CallbackManager& cbmanager,
                            const int              user_event,
                            const Event&           event,
                            CallbackData* const    cdata)
    {
        CbData* data = static_cast< CbData* >(cdata);

        auto&       window  = event->getWindow();
        auto&       comm    = window.getCommunicator();
        auto&       group   = event->getGroup();
        WindowInfo& winInfo = mWindows[&window];

        /* compute complete timestamp and rank*/
        timestamp_t completeEnterTime = event.enterptr()->getTimestamp();
        int         myrank            = comm.getCommSet().getLocalRank(event.get_location().getRank());
        TimeRank    completeEnter;
        completeEnter.mTime = completeEnterTime;
        completeEnter.mRank = myrank;

        #if !defined(SCOUT_DISABLE_RMA_ACCESS_TRACKING)
            setRemoteAccessBit(event, data);
        #endif
        setRemoteLastOp(event, data);

        /* send complete timestamp and rank to remote processes */
        for (uint32_t rank = 0; rank < group.numRanks(); ++rank)
        {
            MPI_Accumulate(&completeEnter, 1, MPI_DOUBLE_INT,
                           comm.getCommSet().getLocalRank(group.getGlobalRank(rank)), LAST_COMPLETE,
                           1, MPI_DOUBLE_INT, MPI_MAXLOC, winInfo.mWin);
        }

        /* complete access epoch on window */
        MPI_Win_complete(winInfo.mWin);

        // search maximum post timestamp
        data->mLastPost.mTime = std::numeric_limits< pearl::timestamp_t >::lowest();

        map< uint32_t, timestamp_t >::iterator it = winInfo.mPostTime.begin();
        while (it != winInfo.mPostTime.end())
        {
            if (data->mLastPost.mTime < it->second)
            {
                data->mLastPost.mTime = it->second;
                data->mLastPost.mRank = it->first;
            }
            ++it;
        }

        data->mCountTarget = winInfo.mAccessProcesses.size();

        // reset data structure
        winInfo.mLastOpTime.clear();
        winInfo.mPostTime.clear();
        winInfo.mAccessProcesses.clear();
    }

    void
    cb_pre_mpi_win_post(const CallbackManager& cbmanager,
                        const int              user_event,
                        const Event&           event,
                        CallbackData* const    cdata)
    {
        auto&       window   = event->getWindow();
        auto&       group    = event->getGroup();
        WindowInfo& winInfo  = mWindows[&window];
        TimeRank*   timerank = reinterpret_cast< TimeRank* >(winInfo.mBuffer);
        UTILS_ASSERT(timerank != nullptr);

        /* initialise window buffer */
        timerank[LAST_COMPLETE / sizeof(TimeRank)] = UNDEFINED_TIME_RANK;
        timerank[LAST_RMA_OP   / sizeof(TimeRank)] = UNDEFINED_TIME_RANK;
        timerank[LAST_POST     / sizeof(TimeRank)] = UNDEFINED_TIME_RANK;
        timerank[LOCAL_POST    / sizeof(TimeRank)] = TimeRank(event.enterptr()->getTimestamp(), -1);

        /* clear bitfield */
        memset(winInfo.mBuffer + BITFIELD_START, 0, winInfo.mBitfieldSize);

        /* start exposure epoch to exchange timestamps and ranks */
        MPI_Win_post(group.getHandle(), 0, winInfo.mWin);
    }

    void
    cb_pre_mpi_win_wait(const CallbackManager& cbmanager,
                        const int              user_event,
                        const Event&           event,
                        CallbackData* const    cdata)
    {
        CbData* data = static_cast< CbData* >(cdata);

        auto&       window   = event->getWindow();
        auto&       winInfo  = mWindows[&window];
        auto* const timerank = reinterpret_cast< TimeRank* >(winInfo.mBuffer);
        UTILS_ASSERT(timerank != nullptr);

        MPI_Win_wait(winInfo.mWin);

        data->mCountOrigin = 0;
        #if !defined(SCOUT_DISABLE_RMA_ACCESS_TRACKING)
            // Count number of origin proccesses that accessed during exposure epoch
            for (size_t i = BITFIELD_START;
                 i < (BITFIELD_START + winInfo.mBitfieldSize);
                 ++i)
            {
                data->mCountOrigin += detail::numBitsSet(winInfo.mBuffer[i]);
            }
        #endif

        /* Set TimeRank structures for time-based GATS patterns */
        data->mLastComplete = timerank[LAST_COMPLETE / sizeof(TimeRank)];
        data->mLastRmaOp    = timerank[LAST_RMA_OP   / sizeof(TimeRank)];
    }

    void
    cb_pre_rma_group_sync(const CallbackManager& cbmanager,
                          const int              user_event,
                          const Event&           event,
                          CallbackData* const    cdata)
    {
        // --- dispatch to specific handlers

        const Region& region = event.enterptr()->getRegion();

        if (is_mpi_rma_start(region))
        {
            cbmanager.notify(MPI_RMA_WIN_START, event, cdata);
        }
        else if (is_mpi_rma_complete(region))
        {
            cbmanager.notify(MPI_RMA_WIN_COMPLETE, event, cdata);
        }
        else if (is_mpi_rma_post(region))
        {
            cbmanager.notify(MPI_RMA_WIN_POST, event, cdata);
        }
        else if (is_mpi_rma_wait(region))
        {
            cbmanager.notify(MPI_RMA_WIN_WAIT, event, cdata);
        }
    }

    // --- rma operations handlers ---------------------------------------

    void
    cb_pre_rma_op_start(const CallbackManager& cbmanager,
                        const int              user_event,
                        const Event&           event,
                        CallbackData* const    cdata)
    {
        Event leave  = event.leaveptr();
        auto& window = event->getWindow();
        auto& comm   = window.getCommunicator();

        int myrank = comm.getCommSet().getLocalRank(event.get_location().getRank());

        WindowInfo& winInfo = mWindows[&window];

        TimeRank timerank(leave->getTimestamp(), myrank);
        winInfo.mLastOpTime[event->getRemote()] = timerank;

        winInfo.mAccessProcesses.insert(event->getRemote());
    }

    //
    // --- backward wait-state / synchpoint detection (bws) callbacks ----
    //

    void
    cb_progress_barrier(const CallbackManager& cbmanager,
                        const int              user_event,
                        const Event&           event,
                        CallbackData* const    cdata)
    {
        // In p2p-only replays (bws, fws, fwc), add artificial barriers
        // at collective events to prevent p2p overload
        MPI_Barrier(getMpiHandle(event->getComm()));
    }

    void
    cb_bw_pre_send(const CallbackManager& cbmanager,
                   const int              user_event,
                   const Event&           event,
                   CallbackData* const    cdata)
    {
        // backward communication: receive message from original sender

        CbData* const data = static_cast< CbData* >(cdata);

        if (event.completion()->isOfType(MPI_CANCELLED))
        {
            return;
        }

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_TRIGGER) << event;

        cbmanager.notify(PRE_SEND, event, data);

        const MpiSend_rep& sendRep = event_cast< MpiSend_rep >(*event);
        data->mRemote->recv(*data->mDefs,
                            sendRep.getComm(),
                            sendRep.getDestination(),
                            sendRep.getTag());

        cbmanager.notify(POST_SEND, event, data);
    }

    void
    cb_bw_pre_recv(const CallbackManager& cbmanager,
                   const int              user_event,
                   const Event&           event,
                   CallbackData* const    cdata)
    {
        // backward communication: send message to the original sender

        CbData* const data = static_cast< CbData* >(cdata);

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_TRIGGER) << event;

        cbmanager.notify(PRE_RECV, event, data);

        const MpiReceive_rep& recvRep = event_cast< MpiReceive_rep >(*event);
        auto* const           msg     = data->mLocal->isend(recvRep.getComm(),
                                                            recvRep.getSource(),
                                                            recvRep.getTag());
        mPendingMsgs.push_back(msg);
        mPendingReqs.push_back(msg->get_request());

        cbmanager.notify(POST_RECV, event, data);
    }

    void
    cb_bws_pre_send_setup(const CallbackManager& cbmanager,
                          const int              user_event,
                          const Event&           event,
                          CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        data->mLocal->addEvent(event, ROLE_SEND);

        const Event completion = event.completion();

        data->mLocal->addEvent(completion.enterptr(), ROLE_ENTER_SENDCMP);
        data->mLocal->addEvent(completion.leaveptr(), ROLE_LEAVE_SENDCMP);
    }

    void
    cb_bws_pre_recv_setup(const CallbackManager& cbmanager,
                          const int              user_event,
                          const Event&           event,
                          CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        data->mLocal->addEvent(event, ROLE_RECV);

        const Event request = event.request();

        data->mLocal->addEvent(request.enterptr(), ROLE_ENTER_RECVREQ);
        data->mLocal->addEvent(request.leaveptr(), ROLE_LEAVE_RECVREQ);
    }

    //
    // --- forward synchpoint detection (fws) callbacks ------------------
    //

    void
    cb_fw_pre_sendcmp(const CallbackManager& cbmanager,
                      const int              user_event,
                      const Event&           event,
                      CallbackData* const    cdata)
    {
        // Handle inverse communication:
        // In forward replay, this means msg from receive request to send
        // completion, i.e., receive message from receive request here.

        CbData* const data = static_cast< CbData* >(cdata);

        // Only handle Late Receiver syncpoints
        if (!data->mSynchpointHandler->isWaitstate(event))
        {
            return;
        }

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_TRIGGER) << event;

        const MpiSend_rep& sendRep = event_cast< MpiSend_rep >(*event.request());
        const auto         commIt  = mInvComms.find(sendRep.getComm().getId());
        UTILS_ASSERT((commIt != mInvComms.end()) && (commIt->second != nullptr));

        data->mInvRemote->recv(*data->mDefs,
                               *(commIt->second),
                               sendRep.getDestination(),
                               sendRep.getTag());

        cbmanager.notify(POST_INV_SENDCMP, event, data);
    }

    void
    cb_fw_pre_recvreq(const CallbackManager& cbmanager,
                      const int              user_event,
                      const Event&           event,
                      CallbackData* const    cdata)
    {
        // Handle inverse communication:
        // In forward replay, this means msg from receive request to send
        // completion, i.e., send message to send completion here.

        CbData* const data = static_cast< CbData* >(cdata);

        // Only handle Late Receiver syncpoints
        if (  !data->mSynchpointHandler->isSynchpoint(event)
           || data->mSynchpointHandler->isWaitstate(event))
        {
            return;
        }

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_TRIGGER) << event;

        cbmanager.notify(PRE_INV_RECVREQ, event, data);

        const MpiReceive_rep& recvRep = event_cast< MpiReceive_rep >(*event.completion());
        const auto            commIt  = mInvComms.find(recvRep.getComm().getId());
        UTILS_ASSERT((commIt != mInvComms.end()) && (commIt->second != nullptr));

        MpiMessage* msg = data->mInvLocal->isend(*(commIt->second),
                                                 recvRep.getSource(),
                                                 recvRep.getTag());
        mPendingMsgs.push_back(msg);
        mPendingReqs.push_back(msg->get_request());
    }

    //
    // --- backward delay / critical-path analysis (bwc) -----------------
    //

    void
    cb_bwc_prepare(const CallbackManager& cbmanager,
                   const int              user_event,
                   const Event&           event,
                   CallbackData* const    cdata)
    {
        const CbData* const data = static_cast< CbData* >(cdata);

        // --- duplicate communicators for inverse communication

        auto           commIt   = mInvComms.begin();
        const uint32_t numComms = data->mDefs->numCommunicators();
        for (uint32_t c = 0; c < numComms; ++c)
        {
            uint32_t localId = (commIt == mInvComms.end() ? Communicator::NO_ID : commIt->first);
            uint32_t globalId;

            MPI_Allreduce(&localId, &globalId, 1, SCALASCA_MPI_UINT32_T, MPI_MIN, MPI_COMM_WORLD);

            // FIXME: This assumes that Communicator::NO_ID is the max element!
            if (globalId == Communicator::NO_ID)
            {
                break;
            }

            // Create communicator duplicate
            const MpiComm& comm = static_cast< const MpiComm& >(data->mDefs->getCommunicator(globalId));
            if (comm.getHandle() != MPI_COMM_NULL)
            {
                mInvComms[globalId] = comm.duplicate();
            }

            if (  (commIt != mInvComms.end())
               && (commIt->first == globalId))
            {
                ++commIt;
            }
        }
    }

    // --- init/finalize

    void
    cb_bwc_pre_init_finalize(const CallbackManager& cbmanager,
                             const int              user_event,
                             const Event&           event,
                             CallbackData* const    cdata)
    {
        CbData* data = static_cast< CbData* >(cdata);

        collinfo_map_t::iterator it = mCollInfo.find(event);
        UTILS_ASSERT(it != mCollInfo.end());

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_TRIGGER) << event;

        data->mCollinfo = it->second;
    }

    // --- collective communication dispatch

    void
    cb_bw_pre_mpi_collend(const CallbackManager& cbmanager,
                          const int              user_event,
                          const Event&           event,
                          CallbackData* const    cdata)
    {
        CbData* const data = static_cast< CbData* >(cdata);

        // Skip collectives that have not been handled by the forward pass
        // (e.g., due to self-like communicators or MPI_Alltoallv/MPI_Alltoallw)
        collinfo_map_t::iterator it = mCollInfo.find(event);
        if (it == mCollInfo.end())
        {
            return;
        }

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_TRIGGER) << event;

        data->mCollinfo = it->second;

        // Dispatch
        const MpiCollEnd_rep&  endRep   = event_cast< MpiCollEnd_rep >(*event);
        const CollectiveOpType collType = endRep.getCollType();
        switch (collType)
        {
            // CASE 1: Barrier
            case CollectiveOpType::BARRIER:
                cbmanager.notify(SYNC_COLL, event, data);
                break;

            // CASE 2: 1-to-N communication
            case CollectiveOpType::BCAST:
            case CollectiveOpType::SCATTER:
            case CollectiveOpType::SCATTERV:
                cbmanager.notify(COLL_12N, event, data);
                break;

            // CASE 3: N-to-N communication w/o MPI_Alltoallv & MPI_Alltoallw
            case CollectiveOpType::ALLGATHER:
            case CollectiveOpType::ALLGATHERV:
            case CollectiveOpType::ALLREDUCE:
            case CollectiveOpType::ALLTOALL:
            case CollectiveOpType::REDUCE_SCATTER:
            case CollectiveOpType::REDUCE_SCATTER_BLOCK:
                cbmanager.notify(COLL_N2N, event, data);
                break;

            // CASE 4: N-to-1 communication
            case CollectiveOpType::GATHER:
            case CollectiveOpType::GATHERV:
            case CollectiveOpType::REDUCE:
                cbmanager.notify(COLL_N21, event, data);
                break;

            // CASE 5: Prefix reduction
            case CollectiveOpType::EXSCAN:
            case CollectiveOpType::SCAN:
                cbmanager.notify(COLL_SCAN, event, data);
                break;

            // CASE 6: Communicator creation
            case CollectiveOpType::CREATE_HANDLE:
                if (is_mpi_init(event.enterptr()->getRegion()))
                {
                    cbmanager.notify(PRE_INIT, event, data);
                    cbmanager.notify(INIT, event, data);
                    break;
                }
                cbmanager.notify(COLL_COMM_CREATE, event, data);
                break;

            // Case 7: Communicator destruction
            case CollectiveOpType::DESTROY_HANDLE:
                if (is_mpi_finalize(event.enterptr()->getRegion()))
                {
                    cbmanager.notify(PRE_FINALIZE, event, data);
                    cbmanager.notify(FINALIZE, event, data);
                    break;
                }
                cbmanager.notify(COLL_COMM_FREE, event, data);
                break;

            default:
                UTILS_BUG("Unknown MPI collective operation encountered (%d)",
                          static_cast< int >(cxxslb::to_underlying(collType)));
        }
    }

    void
    cb_bw_pre_sendcmp(const CallbackManager& cbmanager,
                      const int              user_event,
                      const Event&           event,
                      CallbackData* const    cdata)
    {
        // Handle inverse communication:
        // In backward replay, this means msg from send completion to receive
        // request, i.e., send message to receive request here.

        CbData* const data = static_cast< CbData* >(cdata);

        // Only handle Late Receiver syncpoints
        if (!data->mSynchpointHandler->isWaitstate(event))
        {
            return;
        }

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_TRIGGER) << event;

        cbmanager.notify(PRE_INV_SENDCMP, event, data);

        const MpiSend_rep& sendRep = event_cast< MpiSend_rep >(*event.request());
        const auto         commIt  = mInvComms.find(sendRep.getComm().getId());
        UTILS_ASSERT((commIt != mInvComms.end()) && (commIt->second != nullptr));

        MpiMessage* msg = data->mInvLocal->isend(*(commIt->second),
                                                 sendRep.getDestination(),
                                                 sendRep.getTag());
        mPendingMsgs.push_back(msg);
        mPendingReqs.push_back(msg->get_request());
    }

    void
    cb_bw_pre_recvreq(const CallbackManager& cbmanager,
                      const int              user_event,
                      const Event&           event,
                      CallbackData* const    cdata)
    {
        // Handle inverse communication:
        // In backward replay, this means msg from send completion to receive
        // request, i.e., receive message from send completion here.

        CbData* const data = static_cast< CbData* >(cdata);

        // Only handle Late Receiver syncpoints
        if (  !data->mSynchpointHandler->isSynchpoint(event)
           || data->mSynchpointHandler->isWaitstate(event))
        {
            return;
        }

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_TRIGGER) << event;

        const MpiReceive_rep& recvRep = event_cast< MpiReceive_rep >(*event.completion());
        const auto            commIt  = mInvComms.find(recvRep.getComm().getId());
        UTILS_ASSERT((commIt != mInvComms.end()) && (commIt->second != nullptr));

        data->mInvRemote->recv(*data->mDefs,
                               *(commIt->second),
                               recvRep.getSource(),
                               recvRep.getTag());

        cbmanager.notify(POST_INV_RECVREQ, event, data);
    }

    //
    // --- callback registration -----------------------------------------
    //

    void
    register_pre_pattern_callbacks(const CallbackManagerMap& cbm)
    {
        struct cb_evt_table_t
        {
            pearl::event_t evt;
            void           (MpiCHImpl::* cbp)(const pearl::CallbackManager&,
                                              int,
                                              const pearl::Event&,
                                              pearl::CallbackData*);
        };


        struct cb_uevt_table_t
        {
            int  evt;
            void (MpiCHImpl::* cbp)(const pearl::CallbackManager&,
                                    int,
                                    const pearl::Event&,
                                    pearl::CallbackData*);
        };


        // --- main forward replay callbacks

        // *INDENT-OFF*
        const struct cb_evt_table_t main_evt_tbl[] = {
            { MPI_COLLECTIVE_END,     &MpiCHImpl::cb_pre_mpi_collend    },
            { MPI_RMA_COLLECTIVE_END, &MpiCHImpl::cb_pre_rma_collend    },

            { MPI_RMA_GATS,           &MpiCHImpl::cb_pre_rma_group_sync },

            { GROUP_SEND,             &MpiCHImpl::cb_pre_send           },
            { GROUP_RECV,             &MpiCHImpl::cb_pre_recv           },

            { MPI_RMA_PUT_START,      &MpiCHImpl::cb_pre_rma_op_start   },
            { MPI_RMA_GET_START,      &MpiCHImpl::cb_pre_rma_op_start   },

            { GROUP_SEND,             &MpiCHImpl::cb_process_msgs       },
            { GROUP_RECV,             &MpiCHImpl::cb_process_msgs       },
            { GROUP_END,              &MpiCHImpl::cb_process_msgs       },

            { NUM_EVENT_TYPES,        nullptr                           }
        };
        // *INDENT-ON*


        // *INDENT-OFF*
        const struct cb_uevt_table_t main_uevt_tbl[] = {
            { PRE_SEND,              &MpiCHImpl::cb_pre_send_setup       },
            { PRE_RECV,              &MpiCHImpl::cb_pre_recv_setup       },
            { PRE_INIT,              &MpiCHImpl::cb_pre_sync_coll        },
            { PRE_FINALIZE,          &MpiCHImpl::cb_pre_sync_coll        },
            { PRE_COLL_12N,          &MpiCHImpl::cb_pre_coll_12n         },
            { PRE_COLL_N2N,          &MpiCHImpl::cb_pre_coll_n2n         },
            { PRE_COLL_N21,          &MpiCHImpl::cb_pre_coll_n21         },
            { PRE_COLL_SCAN,         &MpiCHImpl::cb_pre_coll_scan        },
            { PRE_COLL_COMM_CREATE,  &MpiCHImpl::cb_pre_coll_comm_create },
            { PRE_COLL_COMM_FREE,    &MpiCHImpl::cb_pre_coll_comm_free   },
            { PRE_SYNC_COLL,         &MpiCHImpl::cb_pre_sync_coll        },

            { RMA_SYNC_COLLECTIVE,   &MpiCHImpl::cb_pre_rma_sync_coll    },
            { MPI_RMA_WIN_CREATE,    &MpiCHImpl::cb_pre_rma_win_create   },
            { MPI_RMA_WIN_FENCE,     &MpiCHImpl::cb_pre_rma_win_fence    },
            { MPI_RMA_WIN_FREE,      &MpiCHImpl::cb_pre_rma_win_free     },
            { MPI_RMA_WIN_START,     &MpiCHImpl::cb_pre_mpi_win_start    },
            { MPI_RMA_WIN_COMPLETE,  &MpiCHImpl::cb_pre_mpi_win_complete },
            { MPI_RMA_WIN_POST,      &MpiCHImpl::cb_pre_mpi_win_post     },
            { MPI_RMA_WIN_WAIT,      &MpiCHImpl::cb_pre_mpi_win_wait     },

            { FINISHED,              &MpiCHImpl::cb_finished             },

            { 0,                     nullptr                             }
        };
        // *INDENT-ON*


        // --- backward wait-state / synchpoint detection callbacks

        // *INDENT-OFF*
        const struct cb_evt_table_t bws_evt_tbl[] = {
            { GROUP_SEND,         &MpiCHImpl::cb_bw_pre_send      },
            { GROUP_RECV,         &MpiCHImpl::cb_bw_pre_recv      },

            { GROUP_SEND,         &MpiCHImpl::cb_process_msgs     },
            { GROUP_RECV,         &MpiCHImpl::cb_process_msgs     },
            { GROUP_END,          &MpiCHImpl::cb_process_msgs     },

            { MPI_COLLECTIVE_END, &MpiCHImpl::cb_progress_barrier },

            { NUM_EVENT_TYPES,    nullptr                         }
        };
        // *INDENT-ON*


        // *INDENT-OFF*
        const struct cb_uevt_table_t bws_uevt_tbl[] = {
            { PRE_SEND, &MpiCHImpl::cb_bws_pre_send_setup },
            { PRE_RECV, &MpiCHImpl::cb_bws_pre_recv_setup },

            { FINISHED, &MpiCHImpl::cb_finished           },

            { 0,        nullptr                           }
        };
        // *INDENT-ON*


        // --- forward wait-state / synchpoint detection (fws) callbacks

        // *INDENT-OFF*
        const struct cb_evt_table_t fws_evt_tbl[] = {

            { GROUP_SEND,         &MpiCHImpl::cb_pre_send         },
            { GROUP_RECV,         &MpiCHImpl::cb_pre_recv         },

            { GROUP_SEND,         &MpiCHImpl::cb_process_msgs     },
            { GROUP_RECV,         &MpiCHImpl::cb_process_msgs     },
            { GROUP_END,          &MpiCHImpl::cb_process_msgs     },

            { MPI_COLLECTIVE_END, &MpiCHImpl::cb_progress_barrier },

            { NUM_EVENT_TYPES,    nullptr                         }
        };
        // *INDENT-ON*


        // *INDENT-OFF*
        const struct cb_uevt_table_t fws_uevt_tbl[] = {
            { FINISHED, &MpiCHImpl::cb_finished },

            { 0,        nullptr                 }
        };
        // *INDENT-ON*


        // --- delay / critical-path analysis backward replay (bwc) callbacks

        // *INDENT-OFF*
        const struct cb_evt_table_t bwc_evt_tbl[] = {
            { GROUP_SEND,         &MpiCHImpl::cb_bw_pre_send        },
            { GROUP_RECV,         &MpiCHImpl::cb_bw_pre_recv        },

            { MPI_SEND,           &MpiCHImpl::cb_bw_pre_sendcmp     },
            { MPI_SEND_COMPLETE,  &MpiCHImpl::cb_bw_pre_sendcmp     },
            { MPI_RECV,           &MpiCHImpl::cb_bw_pre_recvreq     },
            { MPI_RECV_REQUEST,   &MpiCHImpl::cb_bw_pre_recvreq     },

            { MPI_COLLECTIVE_END, &MpiCHImpl::cb_bw_pre_mpi_collend },

            { GROUP_SEND,         &MpiCHImpl::cb_process_msgs       },
            { GROUP_RECV,         &MpiCHImpl::cb_process_msgs       },
            { GROUP_END,          &MpiCHImpl::cb_process_msgs       },

            { NUM_EVENT_TYPES,    nullptr                           }
        };
        // *INDENT-ON*


        // *INDENT-OFF*
        const struct cb_uevt_table_t bwc_uevt_tbl[] = {
            { PREPARE,      &MpiCHImpl::cb_bwc_prepare           },
            { FINISHED,     &MpiCHImpl::cb_finished              },

            { PRE_INIT,     &MpiCHImpl::cb_bwc_pre_init_finalize },
            { PRE_FINALIZE, &MpiCHImpl::cb_bwc_pre_init_finalize },

            { 0,            nullptr                              }
        };
        // *INDENT-ON*


        // --- forward propagating/indirect wait-state analysis (fwc) callbacks

        // *INDENT-OFF*
        const struct cb_evt_table_t fwc_evt_tbl[] = {
            { GROUP_SEND,        &MpiCHImpl::cb_pre_send       },
            { GROUP_RECV,        &MpiCHImpl::cb_pre_recv       },

            { MPI_SEND,          &MpiCHImpl::cb_fw_pre_sendcmp },
            { MPI_SEND_COMPLETE, &MpiCHImpl::cb_fw_pre_sendcmp },
            { MPI_RECV,          &MpiCHImpl::cb_fw_pre_recvreq },
            { MPI_RECV_REQUEST,  &MpiCHImpl::cb_fw_pre_recvreq },

            { GROUP_SEND,        &MpiCHImpl::cb_process_msgs   },
            { GROUP_RECV,        &MpiCHImpl::cb_process_msgs   },
            { GROUP_END,         &MpiCHImpl::cb_process_msgs   },

            { NUM_EVENT_TYPES,   nullptr                       }
        };
        // *INDENT-ON*


        // *INDENT-OFF*
        const struct cb_uevt_table_t fwc_uevt_tbl[] = {
            { FINISHED, &MpiCHImpl::cb_finished },

            { 0,        nullptr                 }
        };
        // *INDENT-ON*


        // --- global callback tables table :)

        // *INDENT-OFF*
        const struct cb_tables_t
        {
            const char*                   stage;
            const struct cb_evt_table_t*  evts;
            const struct cb_uevt_table_t* uevts;
        }
        cb_tables[] = {
            { "",      main_evt_tbl, main_uevt_tbl },
            { "bws",   bws_evt_tbl,  bws_uevt_tbl  },
            { "fws",   fws_evt_tbl,  fws_uevt_tbl  },
            { "bwc",   bwc_evt_tbl,  bwc_uevt_tbl  },
            { "fwc",   fwc_evt_tbl,  fwc_uevt_tbl  },

            { nullptr, nullptr,      nullptr       }
        };
        // *INDENT-ON*


        // --- register callbacks

        for (const struct cb_tables_t* t = cb_tables; t->stage; ++t)
        {
            CallbackManagerMap::const_iterator it = cbm.find(t->stage);

            UTILS_ASSERT(it != cbm.end());

            for (const struct cb_evt_table_t* c = t->evts; c->cbp; ++c)
            {
                (it->second)->register_callback(c->evt, PEARL_create_callback(this, c->cbp));
            }
            for (const struct cb_uevt_table_t* c = t->uevts; c->cbp; ++c)
            {
                (it->second)->register_callback(c->evt, PEARL_create_callback(this, c->cbp));
            }
        }
    }
};


MpiCommunicationHandler::MpiCommunicationHandler()
    : mP(cxxslb::make_unique< MpiCHImpl >())
{
}


MpiCommunicationHandler::~MpiCommunicationHandler()
{
}


void
MpiCommunicationHandler::register_pre_pattern_callbacks(const CallbackManagerMap& cbm)
{
    mP->register_pre_pattern_callbacks(cbm);
}


const TimeRank MpiCommunicationHandler::MpiCHImpl::UNDEFINED_TIME_RANK;
