/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2012-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "MpiNonBlockingBarrier.h"

#include <UTILS_Error.h>

using namespace std;
using namespace scout;


// --- Constructors & destructor --------------------------------------------

MpiNonBlockingBarrier::MpiNonBlockingBarrier(MPI_Comm comm)
    : mStep(-2),
      mNumSteps(0),
      mSize(0),
      mRank(-1),
      mCommunicator(comm)
{
}


// --- Blocking intialization & finalization --------------------------------

void
MpiNonBlockingBarrier::init()
{
    // Identify duplicate call of init()
    UTILS_ASSERT(mStep == -2);

    MPI_Comm_size(mCommunicator, &mSize);
    MPI_Comm_rank(mCommunicator, &mRank);

    // Determine total number of steps
    mNumSteps = 0;
    for (int i = 1; i < mSize; i *= 2)
    {
        ++mNumSteps;
    }
    UTILS_ASSERT(mNumSteps > 0);

    // Create communicator duplicate to ensure independent p2p communication
    MPI_Comm tmpComm = mCommunicator;
    if (tmpComm != MPI_COMM_NULL)
    {
        MPI_Comm_dup(tmpComm, &mCommunicator);
    }

    // Set up receive requests
    mRequests.resize(mNumSteps);
    for (int i = 0; i < mNumSteps; ++i)
    {
        const int offset = 1 << i;
        const int source = (mRank + mSize - offset) % mSize;
        const int tag    = 0;

        MPI_Irecv(MPI_BOTTOM, 0, MPI_INT, source, tag, mCommunicator,
                  &mRequests[i]);
    }

    mStep = -1;
}


void
MpiNonBlockingBarrier::finalize()
{
    UTILS_ASSERT(mStep == -1);

    MPI_Comm_free(&mCommunicator);
    mRequests.clear();
}


// --- Non-blocking execution of a barrier ----------------------------------

void
MpiNonBlockingBarrier::start()
{
    UTILS_ASSERT(mStep == -1);

    performStep();
}


bool
MpiNonBlockingBarrier::test()
{
    UTILS_ASSERT(mStep >= 0);

    int flag = 0;
    do
    {
        MPI_Test(&mRequests[mStep], &flag, MPI_STATUS_IGNORE);
        if (flag)
        {
            performStep();
        }
    }
    while (mStep < mNumSteps && flag);

    if (mStep == mNumSteps)
    {
        mStep = -1;

        return true;
    }

    return false;
}


void
MpiNonBlockingBarrier::wait()
{
    while (!test())
    {
    }
}


void
MpiNonBlockingBarrier::performStep()
{
    // Increase step count
    mStep++;
    if (mStep >= mNumSteps)
    {
        return;
    }

    // Set communication parameters
    const int offset      = 1 << mStep;
    const int destination = (mRank + offset) % mSize;
    const int tag         = 0;

    // Send message (fire and forget)
    MPI_Request send_request;
    MPI_Isend(MPI_BOTTOM, 0, MPI_INT, destination, tag, mCommunicator,
              &send_request);
    MPI_Request_free(&send_request);
}
