/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief Implementation of the class template TypedPattern.
 *
 *  This file provides the implementation of the class template TypedPattern
 *  including a concrete instantiation for metric values of type `double`.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include "TypedPattern.h"

#include <cstdint>
#include <cstring>

#include <pearl/GlobalDefs.h>
#include <pearl/LocalTrace.h>
#include <pearl/LocationGroup.h>

#include "ReportData.h"

#if defined(_MPI)
    #include <mpi.h>
#endif    // _MPI
#if defined(_OPENMP)
    #include <omp.h>
#endif    // _OPENMP

using namespace std;
using namespace pearl;


namespace
{
// --- Function template prototypes -----------------------------------------

/// @brief Get Cube data type.
///
/// Returns the corresponding Cube data type string for the given data
/// type @p dataT of metric values.
///
/// @tparam dataT
///     Data type of metric values
/// @returns
///     Corresponding Cube data type string
///
template< typename dataT >
constexpr const char*
cubeDatatype();

#if defined(_MPI)
    /// @brief Get MPI data type.
    ///
    /// Returns the corresponding MPI data type for the given data type
    /// @p dataT of metric values.
    ///
    /// @tparam dataT
    ///     Data type of metric values
    /// @returns
    ///     Corresponding MPI data type
    ///
    template< typename dataT >
    MPI_Datatype
    mpiDatatype();
#endif    // _MPI


// --- Explicit specializations -----------------------------------------

/// @copydoc cubeDatatype()
///
template< >
constexpr const char*
cubeDatatype< double >()
{
    return "DOUBLE";
}


/// @copydoc cubeDatatype()
///
template< >
constexpr const char*
cubeDatatype< uint64_t >()
{
    return "UINT64";
}


#if defined(_MPI)
    /// @copydoc mpiDatatype()
    ///
    template< >
    MPI_Datatype
    mpiDatatype< double >()
    {
        return MPI_DOUBLE;
    }


    /// @copydoc mpiDatatype()
    ///
    template< >
    MPI_Datatype
    mpiDatatype< uint64_t >()
    {
        return SCALASCA_MPI_UINT64_T;
    }
#endif // _MPI
}      // unnamed namespace


namespace scout
{
// --- Pattern information access -------------------------------------------

template< typename dataT >
string
TypedPattern< dataT >::getDatatype() const
{
    return cubeDatatype< dataT >();
}


// --- Metric value collation and writing -----------------------------------

/// @todo Fix to deal with non-CPU locations & non-process location groups
///
template< typename dataT >
void
TypedPattern< dataT >::collateSeverities(ReportData&       data,
                                         int               rank,
                                         const LocalTrace& trace)
{
    const pearl::GlobalDefs& defs = trace.get_definitions();

    // --- Determine callpath usage vector and optimal order for writing ---
    const size_t numCallpaths = defs.numCallpaths();
    #pragma omp master
    {
        memset(data.mCallpathUsage, 0, data.mCallpathUsageSize);
    }
    #pragma omp barrier
    for (typename map< Callpath*, dataT >::const_iterator it = m_severity.begin();
         it != m_severity.end();
         ++it)
    {
        const Callpath::IdType id   = it->first->getId();
        const uint32_t         byte = id / 8;
        const uint8_t          bit  = 128 >> (id % 8);

        #pragma omp atomic
        data.mCallpathUsage[byte] |= bit;
    }
    #pragma omp barrier
    #pragma omp master
    {
        #if defined(_MPI)
            MPI_Allreduce(MPI_IN_PLACE, data.mCallpathUsage, data.mCallpathUsageSize,
                          SCALASCA_MPI_UINT8_T, MPI_BOR, MPI_COMM_WORLD);
        #endif    // _MPI

        if (rank == 0)
        {
            // Set callpath usage vector
            cube_set_known_cnodes_for_metric(data.cb, data.metrics[getId()],
                                             reinterpret_cast< char* >(data.mCallpathUsage));

            // Query optimal callpath order for writing
            carray* const callpaths = cube_get_cnodes_for_metric(data.cb,
                                                                 data.metrics[getId()]);
            for (size_t i = 0; i < numCallpaths; ++i)
            {
                data.mCallpathOrder[i] = static_cast< cube_cnode* >(callpaths->data[i])->id;
            }
        }
        #if defined(_MPI)
            MPI_Bcast(data.mCallpathOrder, numCallpaths,
                      SCALASCA_MPI_UINT32_T, 0, MPI_COMM_WORLD);
        #endif    // _MPI
    }

    // --- Set up local & global collation buffers ---
    // The local collation buffer is implicitly shared!
    static dataT* localBuffer  = nullptr;
    dataT*        globalBuffer = nullptr;
    const size_t  numLocations = defs.getLocationGroup(rank).numLocations();
    #pragma omp barrier
    #pragma omp master
    {
        localBuffer = new dataT[numLocations];
        if (rank == 0)
        {
            const uint32_t totalThreads = defs.numLocations();

            globalBuffer = new dataT[totalThreads];
        }
    }

    // --- Collate results ---
    const uint32_t threadId = trace.get_location().getThreadId();
    for (size_t i = 0; i < numCallpaths; i++)
    {
        const uint32_t id   = data.mCallpathOrder[i];
        const uint32_t byte = id / 8;
        const uint8_t  bit  = 128 >> (id % 8);

        if ((data.mCallpathUsage[byte] & bit) == 0)
        {
            continue;
        }

        Callpath* const callpath = defs.get_cnode(id);

        // Collate local severities
        #pragma omp barrier
        typename map< Callpath*, dataT >::const_iterator it = m_severity.find(callpath);
        localBuffer[threadId] = (it != m_severity.end())
                                ? it->second
                                : 0;
        #pragma omp barrier
        #pragma omp master
        {
            #if defined(_MPI)
                // Gather data from all processes
                MPI_Datatype datatype = mpiDatatype< dataT >();
                if (data.mEqualThreads)
                {
                    MPI_Gather(localBuffer, numLocations, datatype,
                               globalBuffer, numLocations, datatype,
                               0, MPI_COMM_WORLD);
                }
                else
                {
                    MPI_Gatherv(localBuffer, numLocations, datatype,
                                globalBuffer, data.mRecvCounts, data.mRecvDispls, datatype,
                                0, MPI_COMM_WORLD);
                }
                MPI_Barrier(MPI_COMM_WORLD);
            #else    // !_MPI
                memcpy(globalBuffer, localBuffer, numLocations * sizeof(dataT));
            #endif    // !_MPI

            if (rank == 0)
            {
                // Write data to file
                cube_write_sev_row(data.cb, data.metrics[getId()],
                                   data.cnodes[callpath->getId()],
                                   globalBuffer);
            }
        }
    }

    // --- Release local & global collation buffers ---
    #pragma omp barrier
    #pragma omp master
    {
        delete[] globalBuffer;
        delete[] localBuffer;
    }

    // Synchronize threads to avoid data races in reused data fields
    #pragma omp barrier
}


// --- Metric value modification --------------------------------------------

template< typename dataT >
dataT
TypedPattern< dataT >::incrementValue(Callpath* const callpath,
                                      const dataT     delta)
{
    auto it = m_severity.lower_bound(callpath);
    if (  (it == m_severity.end())
       || m_severity.key_comp()(callpath, it->first))
    {
        it = m_severity.emplace_hint(it, callpath, delta);
    }
    else
    {
        it->second += delta;
    }

    return it->second;
}


template< typename dataT >
dataT
TypedPattern< dataT >::incrementValue(Callpath* const     callpath,
                                      const dataT         delta,
                                      const std::uint64_t mask,
                                      const char* const   file,
                                      const std::uint64_t line,
                                      const char* const   function)
{
    auto totalValue = incrementValue(callpath, delta);

    #if HAVE(SCALASCA_DEBUG)
        // We can't use one of the DLOG macros here, as the provided
        // source code information has to be passed through.
        if (UTILS_Debug_IsEnabled(mask))
        {
            AFS_PACKAGE_name::utils::DebugStream(file, line, function)
                << getName() << " @ " << *callpath
                << ": " << delta
                << " [total: " << totalValue << ']';
        }
    #endif    // HAVE(SCALASCA_DEBUG)

    return totalValue;
}


template< typename dataT >
dataT
TypedPattern< dataT >::decrementValue(Callpath* const callpath,
                                      const dataT     delta)
{
    auto it = m_severity.lower_bound(callpath);
    if (  (it == m_severity.end())
       || m_severity.key_comp()(callpath, it->first))
    {
        it = m_severity.emplace_hint(it, callpath, -delta);
    }
    else
    {
        it->second -= delta;
    }

    return it->second;
}


template< typename dataT >
dataT
TypedPattern< dataT >::decrementValue(Callpath* const     callpath,
                                      const dataT         delta,
                                      const std::uint64_t mask,
                                      const char* const   file,
                                      const std::uint64_t line,
                                      const char* const   function)
{
    auto totalValue = decrementValue(callpath, delta);

    #if HAVE(SCALASCA_DEBUG)
        // We can't use one of the DLOG macros here, as the provided
        // source code information has to be passed through.
        if (UTILS_Debug_IsEnabled(mask))
        {
            AFS_PACKAGE_name::utils::DebugStream(file, line, function)
                << getName() << " @ " << *callpath
                << ": -" << delta
                << " [total: " << totalValue << ']';
        }
    #endif    // HAVE(SCALASCA_DEBUG)

    return totalValue;
}


// --- Explicit template instantiations -------------------------------------

template class TypedPattern< double >;
template class TypedPattern< uint64_t >;
}    // namespace scout
