/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2022                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief Declaration of helper classes to track waitstate candidate
 *         information.
 *
 *  This header file provides the declaration of various helper classes
 *  used to track information about potential waitstate instances of
 *  different types in case they cannot be handled immediately during the
 *  analysis.
 */
/*-------------------------------------------------------------------------*/


#ifndef SCOUT_WAITSTATECANDIDATES_H
#define SCOUT_WAITSTATECANDIDATES_H


#include <pearl/Event.h>
#include <pearl/RemoteEvent.h>
#include <pearl/pearl_types.h>


namespace pearl
{
// --- Forward declarations -------------------------------------------------

class CallbackManager;
}    // namespace pearl


namespace scout
{
// --- Forward declarations -------------------------------------------------

struct CbData;


/*-------------------------------------------------------------------------*/
/**
 *  @class WaitstateCandidate
 *  @brief Abstract base class for waitstate candidate information.
 *
 *  The class WaitstateCandidate is an abstract base class that defines the
 *  common interface of waitstate candidate information storage classes.
 *  It also stores basic properties of the potential waitstate as well as
 *  the associated local synchronization point that are independent of the
 *  waitstate type, such as the idle time (i.e., waitstate duration) and
 *  the completion time (i.e., the additional time spent in the region
 *  exhibiting the waitstate after the corresponding region has been left
 *  on synchronizing ranks).
 **/
/*-------------------------------------------------------------------------*/

class WaitstateCandidate
{
    public:
        // --- Public methods ------------------------------

        /// @name Constructors & destructor
        /// @{

        /// @brief Constructor.
        ///
        /// Initialize a new instance with the given @p idle time and
        /// @p completion time.
        ///
        /// @param idle
        ///     Idle time (i.e., waitstate duration)
        /// @param completion
        ///     Completion time
        ///
        WaitstateCandidate(pearl::timestamp_t idle,
                           pearl::timestamp_t completion);

        /// @brief Copy constructor.
        ///
        /// Initialize a new instance via member-wise copy-construction from
        /// @p other.
        ///
        /// @param other
        ///     Source operand
        ///
        WaitstateCandidate(const WaitstateCandidate& other) = default;

        /// @brief Move constructor.
        ///
        /// Initialize a new instance via member-wise move-construction from
        /// @p other.
        ///
        /// @param other
        ///     Source operand
        ///
        WaitstateCandidate(WaitstateCandidate&& other) = default;

        /// @brief Destructor.
        ///
        /// Destroy the instance.
        ///
        virtual
        ~WaitstateCandidate() = default;

        /// @}
        /// @name Assignment operators
        /// @{

        /// @brief Copy assignment.
        ///
        /// Assign @p rhs to the instance via member-wise copy.
        ///
        /// @param rhs
        ///     Right-hand operand
        ///
        WaitstateCandidate&
        operator=(const WaitstateCandidate& rhs) = default;

        /// @brief Move assignment.
        ///
        /// Assign @p rhs to the instance via member-wise move.
        ///
        /// @param rhs
        ///     Right-hand operand
        ///
        WaitstateCandidate&
        operator=(WaitstateCandidate&& rhs) = default;

        /// @}
        /// @name Notification functions
        /// @{

        /// @brief Send notification about a waitstate candidate.
        ///
        /// Trigger a waitstate candidate identification notification on the
        /// callback manager @p cbmanager via a corresponding user event,
        /// providing a waitstate-specific trigger event and the given callback
        /// @p data object.
        ///
        /// @note
        ///     Candidate notification is not supported by every waitstate
        ///     type.  If this is the case, calling this function is a no-op.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param event
        ///     Trace event for which the waitstate candidate is announced
        /// @param data
        ///     Callback data object
        ///
        void
        announce(const pearl::CallbackManager& cbmanager,
                 const pearl::Event&           event,
                 CbData*                       data) const;

        /// @brief Send notification about a confirmed waitstate.
        ///
        /// Trigger a waitstate identification notification on the callback
        /// manager @p cbmanager via a corresponding user event, providing a
        /// waitstate-specific trigger event and the given callback @p data
        /// object.  Beforehand, the idle and completion times currently
        /// tracked in @p data are replaced by the values stored for the
        /// candidate, and waitstate-specific information such as the events
        /// involved in the waitstate situation are added to /updated in
        /// @p data.  Afterwards, the modified idle and completion times are
        /// reset to their previous values.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param event
        ///     Trace event for which the waitstate is confirmed
        /// @param data
        ///     Callback data object
        ///
        void
        confirm(const pearl::CallbackManager& cbmanager,
                const pearl::Event&           event,
                CbData*                       data) const;

        /// @}


        // --- Data members --------------------------------

        /// Idle time
        pearl::timestamp_t mIdle;

        /// Completion time
        pearl::timestamp_t mCompletion;


    protected:
        // --- Protected methods ---------------------------

        /// @name Notification functions
        /// @{

        /// @brief Candidate notification hook.
        ///
        /// Trigger a waitstate candidate identification notification on the
        /// callback manager @p cbmanager via a corresponding user event if
        /// available, providing the corresponding trace @p event and the
        /// callback @p data object.  The default implementation is a no-op.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param event
        ///     Trace event for which the waitstate candidate is announced
        /// @param data
        ///     Callback data object
        ///
        virtual void
        notifyCandidate(const pearl::CallbackManager& cbmanager,
                        const pearl::Event&           event,
                        CbData*                       data) const;

        /// @brief Waitstate notification hook.
        ///
        /// Trigger a waitstate identification notification on the callback
        /// manager @p cbmanager via a corresponding user event, providing
        /// the corresponding trace @p event and the callback @p data object.
        /// Beforehand, add/update waitstate-specific information such as the
        /// events involved in the waitstate situation to @p data.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param event
        ///     Trace event for which the waitstate is confirmed
        /// @param data
        ///     Callback data object
        ///
        virtual void
        notifyInstance(const pearl::CallbackManager& cbmanager,
                       const pearl::Event&           event,
                       CbData*                       data) const = 0;

        /// @}
};


// --- Related functions ----------------------------------------------------

/// @name Comparison operators
/// @{

/// @brief   Less-than operator.
/// @relates WaitstateCandidate
///
/// Compare @p lhs and @p rhs based on their idle times.
///
/// @param lhs
///     Left-hand operand
/// @param rhs
///     Right-hand operand
/// @returns
///     `true` if the idle time of @p lhs is less than the idle time of
///     @p rhs, `false` otherwise
///
/// @exception_nothrow
///
bool
operator<(const WaitstateCandidate& lhs,
          const WaitstateCandidate& rhs) noexcept;

/// @}


/*-------------------------------------------------------------------------*/
/**
 *  @class LateSenderCandidate
 *  @brief Storage class for "Late Sender" waitstate candidate information.
 *
 *  The LateSenderCandidate class stores information related to "Late Sender"
 *  waitstate candidates.  Besides the idle and completion time, this also
 *  covers the corresponding send and receive events.
 **/
/*-------------------------------------------------------------------------*/

class LateSenderCandidate
    : public WaitstateCandidate
{
    public:
        // --- Public methods ------------------------------

        /// @name Constructors & destructor
        /// @{

        /// @brief Constructor.
        ///
        /// Initialize a new instance with the given "Late Sender"-related
        /// information.
        ///
        /// @param idle
        ///     Idle time (i.e., waitstate duration)
        /// @param completion
        ///     Completion time
        /// @param send
        ///     Send (request) event
        /// @param recv
        ///     Receive (completion) event
        ///
        LateSenderCandidate(pearl::timestamp_t        idle,
                            pearl::timestamp_t        completion,
                            const pearl::RemoteEvent& send,
                            const pearl::Event&       recv);

        /// @}


        // --- Data members --------------------------------

        /// Send (request) event
        pearl::RemoteEvent mSend;

        /// Receive (completion) event
        pearl::Event mRecv;


    protected:
        // --- Protected methods ---------------------------

        /// @name Notification functions
        /// @{

        /// @brief "Late Sender" candidate notification hook.
        ///
        /// Trigger a "Late Sender" candidate identification notification on
        /// the callback manager @p cbmanager via a `LATE_SENDER_CANDIDATE`
        /// user event, providing the corresponding trace @p event and the
        /// callback @p data object.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param event
        ///     Trace event for which the waitstate candidate is announced
        /// @param data
        ///     Callback data object
        ///
        void
        notifyCandidate(const pearl::CallbackManager& cbmanager,
                        const pearl::Event&           event,
                        CbData*                       data) const override;

        /// @brief Waitstate notification hook.
        ///
        /// Trigger a "Late Sender" waitstate identification notification on
        /// the callback manager @p cbmanager via the `LATE_SENDER` user event,
        /// providing the corresponding trace @p event and the callback @p data
        /// object.  Beforehand, add/update waitstate-specific information such
        /// as the events involved in the waitstate situation to @p data.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param event
        ///     Trace event for which the waitstate is confirmed
        /// @param data
        ///     Callback data object
        ///
        void
        notifyInstance(const pearl::CallbackManager& cbmanager,
                       const pearl::Event&           event,
                       CbData*                       data) const override;

        /// @}
};


/*-------------------------------------------------------------------------*/
/**
 *  @class LswoCandidate
 *  @brief Storage class for "Late Sender, wrong order" waitstate candidate
 *         information.
 *
 *  The LswoCandidate class stores information related to "Late Sender,
 *  wrong order" waitstate candidates.  Besides the idle and completion
 *  time, this also covers the corresponding send and receive events of
 *  the *Late Sender* situation.
 **/
/*-------------------------------------------------------------------------*/

class LswoCandidate
    : public LateSenderCandidate
{
    public:
        // --- Public methods ------------------------------

        /// @name Constructors & destructor
        /// @{

        /// @brief Constructor.
        ///
        /// Initialize a new instance with the given "Late Sender, wrong
        /// order"-related information.
        ///
        /// @param idle
        ///     Idle time (i.e., waitstate duration)
        /// @param completion
        ///     Completion time
        /// @param send
        ///     Send (request) event
        /// @param recv
        ///     Receive (completion) event
        ///
        LswoCandidate(pearl::timestamp_t        idle,
                      pearl::timestamp_t        completion,
                      const pearl::RemoteEvent& send,
                      const pearl::Event&       recv);

        /// @}


    protected:
        // --- Protected methods ---------------------------

        /// @name Notification functions
        /// @{

        /// @brief "Late Sender, wrong order" candidate notification hook.
        ///
        /// This is a no-op.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param event
        ///     Trace event for which the waitstate candidate is announced
        /// @param data
        ///     Callback data object
        ///
        void
        notifyCandidate(const pearl::CallbackManager& cbmanager,
                        const pearl::Event&           event,
                        CbData*                       data) const override;

        /// @brief Waitstate notification hook.
        ///
        /// Trigger a "Late Sender, wrong order" waitstate identification
        /// notification on the callback manager @p cbmanager via the
        /// `LATE_SENDER_WO` user event, providing the corresponding trace
        /// @p event and the callback @p data object.  Beforehand, add
        /// waitstate-specific information such as the events involved in
        /// the waitstate situation to @p data.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param event
        ///     Trace event for which the waitstate is confirmed
        /// @param data
        ///     Callback data object
        ///
        void
        notifyInstance(const pearl::CallbackManager& cbmanager,
                       const pearl::Event&           event,
                       CbData*                       data) const override;

        /// @}
};
}    // namespace scout


// --- Inline methods & related functions -----------------------------------

#include "WaitstateCandidates-inl.h"


#endif    // !SCOUT_WAITSTATECANDIDATES_H
