/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2019-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include <pearl/CartTopology.h>

#include <sstream>
#include <string>
#include <vector>

#include <cxxslb/memory.h>

#include <gtest/gtest.h>

#include <pearl/CartDimension.h>
#include <pearl/CommLocationSet.h>
#include <pearl/CommSet.h>
#include <pearl/Communicator.h>
#include <pearl/Error.h>
#include <pearl/Location.h>
#include <pearl/String.h>
#include <pearl/SystemNode.h>
#include <pearl/iomanip.h>

#include "DefinitionContainer.h"
#include "Process.h"
#include "iomanip_detail.h"

using namespace std;
using namespace testing;
using namespace pearl;
using namespace pearl::detail;


namespace
{
// --- Helper ---------------------------------------------------------------

// Test fixture for Cartesian topology tests
class CartTopologyT
    : public Test
{
    public:
        void
        createCommunicator(uint32_t size);


    protected:
        CartTopology::DimensionContainer   mDimensions;
        std::unique_ptr< Communicator >    mCommunicator;
        std::unique_ptr< CommSet >         mCommSet;
        std::unique_ptr< CommLocationSet > mCommLocationSet;
        DefinitionContainer< Location >    mLocations;
        std::unique_ptr< LocationGroup >   mLocationGroup;
        std::unique_ptr< SystemNode >      mSystemNode;
};


// --- CartTopology tests ---------------------------------------------------

TEST_F(CartTopologyT,
       getId_item_returnsIdentifier)
{
    CartTopology item0(0, String::UNDEFINED, Communicator::UNDEFINED,
                       mDimensions);
    EXPECT_EQ(0, item0.getId());

    CartTopology item1(1, String::UNDEFINED, Communicator::UNDEFINED,
                       mDimensions);
    EXPECT_EQ(1, item1.getId());

    CartTopology item42(42, String::UNDEFINED, Communicator::UNDEFINED,
                        mDimensions);
    EXPECT_EQ(42, item42.getId());
}


TEST_F(CartTopologyT,
       getName_item_returnsName)
{
    CartTopology item0(0, String::UNDEFINED, Communicator::UNDEFINED,
                       mDimensions);
    EXPECT_EQ(&String::UNDEFINED, &item0.getName());

    String       name1(0, "Foo");
    CartTopology item1(1, name1, Communicator::UNDEFINED, mDimensions);
    EXPECT_EQ(&name1, &item1.getName());
}


TEST_F(CartTopologyT,
       getCommunicator_item_returnsCommunicator)
{
    CartTopology item0(0, String::UNDEFINED, Communicator::UNDEFINED,
                       mDimensions);
    EXPECT_EQ(&Communicator::UNDEFINED, &item0.getCommunicator());

    createCommunicator(16);
    CartTopology item1(1, String::UNDEFINED, *mCommunicator, mDimensions);
    EXPECT_EQ(mCommunicator.get(), &item1.getCommunicator());
}


TEST_F(CartTopologyT,
       numDimensions_item_returnsDimensionCount)
{
    CartTopology item0(0, String::UNDEFINED, Communicator::UNDEFINED,
                       mDimensions);
    EXPECT_EQ(0, item0.numDimensions());

    CartDimension dim0(0, String::UNDEFINED, 0, CartDimension::PERIODIC);
    mDimensions.push_back(&dim0);

    CartTopology item1(1, String::UNDEFINED, Communicator::UNDEFINED,
                       mDimensions);
    EXPECT_EQ(1, item1.numDimensions());

    CartDimension dim1(1, String::UNDEFINED, 0, CartDimension::PERIODIC);
    mDimensions.push_back(&dim1);

    CartTopology item2(2, String::UNDEFINED, Communicator::UNDEFINED,
                       mDimensions);
    EXPECT_EQ(2, item2.numDimensions());
}


TEST_F(CartTopologyT,
       getDimension_itemAndValidIndex_returnsDimension)
{
    CartDimension dim0(0, String::UNDEFINED, 0, CartDimension::PERIODIC);
    mDimensions.push_back(&dim0);
    CartDimension dim1(1, String::UNDEFINED, 0, CartDimension::PERIODIC);
    mDimensions.push_back(&dim1);

    CartTopology item(0, String::UNDEFINED, Communicator::UNDEFINED,
                      mDimensions);
    EXPECT_EQ(&dim0, &item.getDimension(0));
    EXPECT_EQ(&dim1, &item.getDimension(1));
}


TEST_F(CartTopologyT,
       getDimension_itemAndInvalidIndex_throws)
{
    CartTopology item(0, String::UNDEFINED, Communicator::UNDEFINED,
                      mDimensions);
    EXPECT_THROW(item.getDimension(0), pearl::RuntimeError);
}


TEST_F(CartTopologyT,
       getRankCoordinates_item_returnsEmptyList)
{
    createCommunicator(8);
    CartTopology item(0, String::UNDEFINED, *mCommunicator, mDimensions);
    EXPECT_EQ(0, item.getRankCoordinates(0).size());
}


TEST_F(CartTopologyT,
       getRankCoordinates_itemAndInvalidRank_throws)
{
    CartTopology item(0, String::UNDEFINED, Communicator::UNDEFINED,
                      mDimensions);
    EXPECT_THROW(item.getRankCoordinates(0), pearl::RuntimeError);
}


TEST_F(CartTopologyT,
       operatorLeftShift_itemAtSkipDetail_printsNothing)
{
    String        name0(24, "Periodic");
    CartDimension dim0(42, name0, 2, CartDimension::PERIODIC);
    mDimensions.push_back(&dim0);

    String        name1(25, "Non-periodic");
    CartDimension dim1(43, name1, 3, CartDimension::NON_PERIODIC);
    mDimensions.push_back(&dim1);

    String       name2(4, "My topology");
    CartTopology item(17, name2, Communicator::UNDEFINED, mDimensions);

    string expected;

    ostringstream result;
    result << setDetail(DETAIL_SKIP)
           << item;
    EXPECT_EQ(expected, result.str());
}


TEST_F(CartTopologyT,
       operatorLeftShift_itemAtAbbrevDetail_printsIdentifier)
{
    String        name0(24, "Periodic");
    CartDimension dim0(42, name0, 2, CartDimension::PERIODIC);
    mDimensions.push_back(&dim0);

    String        name1(25, "Non-periodic");
    CartDimension dim1(43, name1, 3, CartDimension::NON_PERIODIC);
    mDimensions.push_back(&dim1);

    String       name2(4, "My topology");
    CartTopology item(17, name2, Communicator::UNDEFINED, mDimensions);

    string expected("17");

    ostringstream result;
    result << setDetail(DETAIL_ABBRV)
           << item;
    EXPECT_EQ(expected, result.str());
}


TEST_F(CartTopologyT,
       operatorLeftShift_itemAtCompactDetail_printsOneliner)
{
    String        name0(24, "Periodic");
    CartDimension dim0(42, name0, 2, CartDimension::PERIODIC);
    mDimensions.push_back(&dim0);

    String        name1(25, "Non-periodic");
    CartDimension dim1(43, name1, 3, CartDimension::NON_PERIODIC);
    mDimensions.push_back(&dim1);

    String       name2(4, "My topology");
    CartTopology item(17, name2, Communicator::UNDEFINED, mDimensions);

    string expected("CartTopology { id: 17, name: \"My topology\", comm: <undef>, #dims: 2, dims: [...], coords: [...] }");

    ostringstream result;
    result << item;
    EXPECT_EQ(expected, result.str());
}


TEST_F(CartTopologyT,
       operatorLeftShift_itemAtFullDetail_printsData)
{
    String        name0(24, "Periodic");
    CartDimension dim0(42, name0, 2, CartDimension::PERIODIC);
    mDimensions.push_back(&dim0);

    String        name1(25, "Non-periodic");
    CartDimension dim1(43, name1, 3, CartDimension::NON_PERIODIC);
    mDimensions.push_back(&dim1);

    String       name2(4, "My topology");
    CartTopology item(17, name2, Communicator::UNDEFINED, mDimensions);

    string expected("CartTopology {\n"
                    "    id: 17\n"
                    "    name: String { id: 4, text: \"My topology\" }\n"
                    "    comm: <undef>\n"
                    "    #dims: 2\n"
                    "    dims: [ 42, 43 ]\n"
                    "    coords: [  ]\n"
                    "}");

    ostringstream result;
    result << setDetail(DETAIL_FULL)
           << item;
    EXPECT_EQ(expected, result.str());
}


TEST_F(CartTopologyT,
       operatorLeftShift_itemAtIncreasedDetail_printsMembers)
{
    String        name0(24, "Periodic");
    CartDimension dim0(42, name0, 2, CartDimension::PERIODIC);
    mDimensions.push_back(&dim0);

    String        name1(25, "Non-periodic");
    CartDimension dim1(43, name1, 3, CartDimension::NON_PERIODIC);
    mDimensions.push_back(&dim1);

    String       name2(4, "My topology");
    CartTopology item(17, name2, Communicator::UNDEFINED, mDimensions);

    string expected("CartTopology {\n"
                    "    id: 17\n"
                    "    name: String {\n"
                    "        id: 4\n"
                    "        text: \"My topology\"\n"
                    "    }\n"
                    "    comm: <undef>\n"
                    "    #dims: 2\n"
                    "    dims: [\n"
                    "        CartDimension { id: 42, name: \"Periodic\", size: 2, periodic: true }\n"
                    "        CartDimension { id: 43, name: \"Non-periodic\", size: 3, periodic: false }\n"
                    "    ]\n"
                    "    coords: [\n"
                    "        \n"
                    "    ]\n"
                    "}");

    ostringstream result;
    result << setDetail(DETAIL_FULL + 1)
           << item;
    EXPECT_EQ(expected, result.str());
}


// --- Helper ---------------------------------------------------------------

void
CartTopologyT::createCommunicator(const uint32_t size)
{
    // Create dummy SystemNode (used as parent)
    mSystemNode =
        cxxslb::make_unique< SystemNode >(0,
                                          String::UNDEFINED,
                                          String::UNDEFINED,
                                          nullptr);

    // Create dummy LocationGroup (used as parent)
    mLocationGroup =
        cxxslb::make_unique< Process >(0,
                                       String::UNDEFINED,
                                       mSystemNode.get(),
                                       nullptr);

    // Create locations
    vector< uint32_t > ranks;
    ranks.reserve(size);
    vector< const Location* > setMembers;
    setMembers.reserve(size);
    for (uint32_t rank = 0; rank < size; ++rank)
    {
        ranks.push_back(rank);

        mLocations.addItem(
            cxxslb::make_unique< Location >(rank,
                                            String::UNDEFINED,
                                            Location::TYPE_CPU_THREAD,
                                            0,
                                            mLocationGroup.get()));
        setMembers.push_back(mLocations.getItemById(rank));
    }

    // Create CommLocationSet
    mCommLocationSet =
        cxxslb::make_unique< CommLocationSet >(0,
                                               String::UNDEFINED,
                                               Paradigm::USER,
                                               setMembers);

    // Create CommSet
    mCommSet =
        cxxslb::make_unique< CommSet >(0,
                                       String::UNDEFINED,
                                       GroupingSet::PROPERTY_NONE,
                                       *mCommLocationSet,
                                       ranks);

    // Create Communicator
    mCommunicator =
        cxxslb::make_unique< Communicator >(0,
                                            String::UNDEFINED,
                                            Communicator::PROPERTY_NONE,
                                            *mCommSet,
                                            nullptr);
}
}    // unnamed namespace
