/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2019-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include <pearl/Communicator.h>

#include <iostream>
#include <string>

#include <cxxslb/memory.h>

#include <gtest/gtest.h>

#include <pearl/CommLocationSet.h>
#include <pearl/CommSet.h>
#include <pearl/Location.h>
#include <pearl/String.h>
#include <pearl/SystemNode.h>
#include <pearl/iomanip.h>

#include "DefinitionContainer.h"
#include "Process.h"
#include "iomanip_detail.h"

using namespace std;
using namespace testing;
using namespace pearl;
using namespace pearl::detail;


namespace
{
// --- Helper ---------------------------------------------------------------

// Test fixture for Communicator tests
class CommunicatorT
    : public Test
{
    public:
        CommunicatorT();


    protected:
        std::unique_ptr< SystemNode >      mSystemNode;
        std::unique_ptr< LocationGroup >   mLocationGroup;
        DefinitionContainer< String >      mStrings;
        DefinitionContainer< Location >    mLocations;
        std::unique_ptr< CommLocationSet > mCommLocationSet;
        std::unique_ptr< CommSet >         mCommSetWorld;
        std::unique_ptr< CommSet >         mCommSetOdd;
        std::unique_ptr< Communicator >    mCommWorld;
        std::unique_ptr< Communicator >    mCommOdd;
};


// --- Communicator tests ---------------------------------------------------

TEST_F(CommunicatorT,
       operatorLeftShift_undefinedAtSkipDetail_printsNothing)
{
    string expected;

    ostringstream result;
    result << setDetail(DETAIL_SKIP)
           << Communicator::UNDEFINED;
    EXPECT_EQ(expected, result.str());
}


TEST_F(CommunicatorT,
       operatorLeftShift_itemAtSkipDetail_printsNothing)
{
    string expected;

    ostringstream result;
    result << setDetail(DETAIL_SKIP)
           << *mCommOdd;
    EXPECT_EQ(expected, result.str());
}


TEST_F(CommunicatorT,
       operatorLeftShift_undefinedAtAbbrvDetail_printsUndef)
{
    string expected("<undef>");

    ostringstream result;
    result << setDetail(DETAIL_ABBRV)
           << Communicator::UNDEFINED;
    EXPECT_EQ(expected, result.str());
}


TEST_F(CommunicatorT,
       operatorLeftShift_itemAtAbbrevDetail_printsIdentifier)
{
    string expected("1");

    ostringstream result;
    result << setDetail(DETAIL_ABBRV)
           << *mCommOdd;
    EXPECT_EQ(expected, result.str());
}


TEST_F(CommunicatorT,
       operatorLeftShift_undefinedAtCompactDetail_printsUndef)
{
    string expected("<undef>");

    ostringstream result;
    result << Communicator::UNDEFINED;
    EXPECT_EQ(expected, result.str());
}


TEST_F(CommunicatorT,
       operatorLeftShift_itemAtCompactDetail_printsOneliner)
{
    // Communicator w/o parent (world)
    {
        string expected("Communicator { id: 0, name: \"world_comm\", comm set: 0, parent: <undef>, flags: 0 (NONE) }");

        ostringstream result;
        result << *mCommWorld;
        EXPECT_EQ(expected, result.str());
    }

    // Communicator w/ parent
    {
        string expected("Communicator { id: 1, name: \"odd_comm\", comm set: 1, parent: 0, flags: 1 (create/destroy events) }");

        ostringstream result;
        result << *mCommOdd;
        EXPECT_EQ(expected, result.str());
    }
}


TEST_F(CommunicatorT,
       operatorLeftShift_undefinedAtFullDetail_printsUndef)
{
    string expected("<undef>");

    ostringstream result;
    result << setDetail(DETAIL_FULL)
           << Communicator::UNDEFINED;
    EXPECT_EQ(expected, result.str());
}


TEST_F(CommunicatorT,
       operatorLeftShift_itemAtFullDetail_printsData)
{
    // Communicator w/o parent (world)
    {
        string expected("Communicator {\n"
                        "    id: 0\n"
                        "    name: String { id: 18, text: \"world_comm\" }\n"
                        "    comm set: CommSet { id: 0, name: \"world_group\", props: 2 (world), paradigm: mpi, size: 16, members: [...] }\n"
                        "    parent: <undef>\n"
                        "    flags: 0 (NONE)\n"
                        "}");

        ostringstream result;
        result << setDetail(DETAIL_FULL)
               << *mCommWorld;
        EXPECT_EQ(expected, result.str());
    }

    // Communicator w/ parent
    {
        string expected("Communicator {\n"
                        "    id: 1\n"
                        "    name: String { id: 19, text: \"odd_comm\" }\n"
                        "    comm set: CommSet { id: 1, name: \"odd_group\", props: 0 (NONE), paradigm: mpi, size: 8, members: [...] }\n"
                        "    parent: Communicator { id: 0, name: \"world_comm\", comm set: 0, parent: <undef>, flags: 0 (NONE) }\n"
                        "    flags: 1 (create/destroy events)\n"
                        "}");

        ostringstream result;
        result << setDetail(DETAIL_FULL)
               << *mCommOdd;
        EXPECT_EQ(expected, result.str());
    }
}


// --- Helper ---------------------------------------------------------------

CommunicatorT::CommunicatorT()
{
    // Create dummy SystemNode (used as parent)
    mSystemNode =
        cxxslb::make_unique< SystemNode >(0,
                                          String::UNDEFINED,
                                          String::UNDEFINED,
                                          nullptr);

    // Create dummy LocationGroup (used as parent)
    mLocationGroup =
        cxxslb::make_unique< Process >(0,
                                       String::UNDEFINED,
                                       mSystemNode.get(),
                                       nullptr);

    // Create locations
    std::vector< const Location* > setMembers;
    setMembers.reserve(16);
    for (uint32_t id = 0; id < 16; ++id)
    {
        mStrings.addItem(
            cxxslb::make_unique< String >(id,
                                          "Rank " + std::to_string(id)));
        mLocations.addItem(
            cxxslb::make_unique< Location >(id,
                                            *mStrings.getItemById(id),
                                            Location::TYPE_CPU_THREAD,
                                            0,
                                            mLocationGroup.get()));
        setMembers.push_back(mLocations.getItemById(id));
    }

    // Create communication location set
    mCommLocationSet =
        cxxslb::make_unique< CommLocationSet >(0,
                                               String::UNDEFINED,
                                               Paradigm::MPI,
                                               setMembers);

    // Create "world" communication set
    vector< uint32_t > ranks;
    ranks.reserve(16);
    for (uint32_t rank = 0; rank < 16; ++rank)
    {
        ranks.push_back(rank);
    }
    mStrings.addItem(cxxslb::make_unique< String >(16, "world_group"));
    mCommSetWorld =
        cxxslb::make_unique< CommSet >(0,
                                       *mStrings.getItemById(16),
                                       GroupingSet::PROPERTY_WORLD,
                                       *mCommLocationSet,
                                       ranks);

    // Create "odd" communication set
    ranks.clear();
    for (uint32_t rank = 1; rank < 16; rank += 2)
    {
        ranks.push_back(rank);
    }
    mStrings.addItem(cxxslb::make_unique< String >(17, "odd_group"));
    mCommSetOdd =
        cxxslb::make_unique< CommSet >(1,
                                       *mStrings.getItemById(17),
                                       GroupingSet::PROPERTY_NONE,
                                       *mCommLocationSet,
                                       ranks);

    // Create "world" communicator
    mStrings.addItem(cxxslb::make_unique< String >(18, "world_comm"));
    mCommWorld =
        cxxslb::make_unique< Communicator >(0,
                                            *mStrings.getItemById(18),
                                            Communicator::PROPERTY_NONE,
                                            *mCommSetWorld,
                                            nullptr);

    // Create "odd" communicator
    mStrings.addItem(cxxslb::make_unique< String >(19, "odd_comm"));
    mCommOdd =
        cxxslb::make_unique< Communicator >(1,
                                            *mStrings.getItemById(19),
                                            Communicator::PROPERTY_CREATE_DESTROY_EVENTS,
                                            *mCommSetOdd,
                                            mCommWorld.get());
}
}    // unnamed namespace
