/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2016-2021                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "DefinitionContainer.h"

#include <utility>

#include <cxxslb/memory.h>

#include <gtest/gtest.h>

using namespace testing;
using namespace pearl::detail;


namespace
{
// --- Helper ---------------------------------------------------------------

// Mocked global definition class implementing the minimal requirements for
// the DefinitionContainer class template
class TestDef
{
    public:
        typedef unsigned int IdType;

        explicit
        TestDef(IdType id)
            : mId(id)
        {
        }

        IdType
        getId() const
        {
            return mId;
        }


    private:
        IdType mId;
};


// --- DefinitionContainer tests --------------------------------------------

TEST(DefinitionContainerT,
     isEmpty_defaultContainer_returnsTrue)
{
    DefinitionContainer< TestDef > container;

    EXPECT_TRUE(container.isEmpty());
}


TEST(DefinitionContainerT,
     isEmpty_nonEmptyContainer_returnsFalse)
{
    DefinitionContainer< TestDef > container;

    container.addItem(cxxslb::make_unique< TestDef >(0));
    EXPECT_FALSE(container.isEmpty());
}


TEST(DefinitionContainerT,
     numItems_defaultContainer_returnsZero)
{
    DefinitionContainer< TestDef > container;

    EXPECT_EQ(0, container.numItems());
}


TEST(DefinitionContainerT,
     getItemById_compactContainerAndValidId_returnsItem)
{
    DefinitionContainer< TestDef > compact;

    auto item0 = cxxslb::make_unique< TestDef >(0);
    auto item1 = cxxslb::make_unique< TestDef >(1);
    auto item2 = cxxslb::make_unique< TestDef >(2);

    auto* itemPtr0 = item0.get();
    auto* itemPtr1 = item1.get();
    auto* itemPtr2 = item2.get();

    compact.addItem(std::move(item0));
    compact.addItem(std::move(item1));
    compact.addItem(std::move(item2));

    EXPECT_EQ(itemPtr0, compact.getItemById(0));
    EXPECT_EQ(itemPtr1, compact.getItemById(1));
    EXPECT_EQ(itemPtr2, compact.getItemById(2));
}


TEST(DefinitionContainerT,
     getItemById_compactContainerAndInvalidId_throws)
{
    DefinitionContainer< TestDef > compact;

    compact.addItem(cxxslb::make_unique< TestDef >(0));
    compact.addItem(cxxslb::make_unique< TestDef >(1));
    compact.addItem(cxxslb::make_unique< TestDef >(2));

    EXPECT_THROW(compact.getItemById(3), std::invalid_argument);
}


TEST(DefinitionContainerT,
     getItemById_sortedContainerAndValidId_returnsItem)
{
    DefinitionContainer< TestDef > sorted;

    auto item0 = cxxslb::make_unique< TestDef >(0);
    auto item1 = cxxslb::make_unique< TestDef >(2);
    auto item2 = cxxslb::make_unique< TestDef >(4);

    auto* itemPtr0 = item0.get();
    auto* itemPtr1 = item1.get();
    auto* itemPtr2 = item2.get();

    sorted.addItem(std::move(item0));
    sorted.addItem(std::move(item1));
    sorted.addItem(std::move(item2));

    EXPECT_EQ(itemPtr0, sorted.getItemById(0));
    EXPECT_EQ(itemPtr1, sorted.getItemById(2));
    EXPECT_EQ(itemPtr2, sorted.getItemById(4));
}


TEST(DefinitionContainerT,
     getItemById_sortedContainerAndInvalidId_throws)
{
    DefinitionContainer< TestDef > sorted;

    sorted.addItem(cxxslb::make_unique< TestDef >(0));
    sorted.addItem(cxxslb::make_unique< TestDef >(2));
    sorted.addItem(cxxslb::make_unique< TestDef >(4));

    EXPECT_THROW(sorted.getItemById(3), std::invalid_argument);
}


TEST(DefinitionContainerT,
     getItemById_unsortedContainerAndValidId_returnsItem)
{
    DefinitionContainer< TestDef > unsorted;

    auto item0 = cxxslb::make_unique< TestDef >(3);
    auto item1 = cxxslb::make_unique< TestDef >(1);
    auto item2 = cxxslb::make_unique< TestDef >(2);

    auto* itemPtr0 = item0.get();
    auto* itemPtr1 = item1.get();
    auto* itemPtr2 = item2.get();

    unsorted.addItem(std::move(item0));
    unsorted.addItem(std::move(item1));
    unsorted.addItem(std::move(item2));

    EXPECT_EQ(itemPtr0, unsorted.getItemById(3));
    EXPECT_EQ(itemPtr1, unsorted.getItemById(1));
    EXPECT_EQ(itemPtr2, unsorted.getItemById(2));
}


TEST(DefinitionContainerT,
     getItemById_unsortedContainerAndInvalidId_throws)
{
    DefinitionContainer< TestDef > unsorted;

    unsorted.addItem(cxxslb::make_unique< TestDef >(3));
    unsorted.addItem(cxxslb::make_unique< TestDef >(1));
    unsorted.addItem(cxxslb::make_unique< TestDef >(2));

    EXPECT_THROW(unsorted.getItemById(0), std::invalid_argument);
}


TEST(DefinitionContainerT,
     getItemByIndex_validIndex_returnsItem)
{
    DefinitionContainer< TestDef > compact;

    auto item0 = cxxslb::make_unique< TestDef >(0);
    auto item1 = cxxslb::make_unique< TestDef >(1);
    auto item2 = cxxslb::make_unique< TestDef >(2);

    auto* itemPtr0 = item0.get();
    auto* itemPtr1 = item1.get();
    auto* itemPtr2 = item2.get();

    compact.addItem(std::move(item0));
    compact.addItem(std::move(item1));
    compact.addItem(std::move(item2));

    EXPECT_EQ(itemPtr0, compact.getItemByIndex(0));
    EXPECT_EQ(itemPtr1, compact.getItemByIndex(1));
    EXPECT_EQ(itemPtr2, compact.getItemByIndex(2));
}


TEST(DefinitionContainerT,
     getItemByIndex_invalidIndex_throws)
{
    DefinitionContainer< TestDef > sorted;

    sorted.addItem(cxxslb::make_unique< TestDef >(0));
    sorted.addItem(cxxslb::make_unique< TestDef >(2));
    sorted.addItem(cxxslb::make_unique< TestDef >(4));

    EXPECT_THROW(sorted.getItemByIndex(3), std::out_of_range);
}


TEST(DefinitionContainerT,
     addItem_itemPtr_incrementsNumItems)
{
    DefinitionContainer< TestDef > container;

    container.addItem(cxxslb::make_unique< TestDef >(0));
    EXPECT_EQ(1, container.numItems());

    container.addItem(cxxslb::make_unique< TestDef >(1));
    EXPECT_EQ(2, container.numItems());

    container.addItem(cxxslb::make_unique< TestDef >(2));
    EXPECT_EQ(3, container.numItems());
}


TEST(DefinitionContainerT,
     addItem_nullptr_throws)
{
    DefinitionContainer< TestDef > container;

    EXPECT_THROW(container.addItem(nullptr), std::invalid_argument);
}


TEST(DefinitionContainerT,
     optimizeLookups_unsortedContainer_sortsItems)
{
    DefinitionContainer< TestDef > unsorted;

    auto item0 = cxxslb::make_unique< TestDef >(3);
    auto item1 = cxxslb::make_unique< TestDef >(1);
    auto item2 = cxxslb::make_unique< TestDef >(2);

    auto* itemPtr0 = item0.get();
    auto* itemPtr1 = item1.get();
    auto* itemPtr2 = item2.get();

    unsorted.addItem(std::move(item0));
    unsorted.addItem(std::move(item1));
    unsorted.addItem(std::move(item2));
    unsorted.optimizeLookups();

    EXPECT_EQ(itemPtr1, unsorted.getItemByIndex(0));
    EXPECT_EQ(itemPtr2, unsorted.getItemByIndex(1));
    EXPECT_EQ(itemPtr0, unsorted.getItemByIndex(2));
    EXPECT_THROW(unsorted.getItemByIndex(3), std::out_of_range);
}
}    // unnamed namespace
