/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2017-2021                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "DefinitionTree.h"

#include <sstream>
#include <string>
#include <utility>

#include <cxxslb/memory.h>

#include <gtest/gtest.h>

using namespace std;
using namespace testing;
using namespace pearl::detail;


namespace
{
// --- Helper ---------------------------------------------------------------

// Mocked global definition class implementing the minimal requirements for
// the DefinitionTree class template
class TestDef
{
    public:
        typedef unsigned int IdType;


        TestDef(const IdType   id,
                TestDef* const parent)
            : mId(id),
              mParent(parent)
        {
        }

        IdType
        getId() const
        {
            return mId;
        }

        TestDef*
        getParent() const
        {
            return mParent;
        }


    private:
        IdType   mId;
        TestDef* mParent;
};


ostream&
print(ostream&       stream,
      const TestDef& item,
      const bool     isLastChild)
{
    return stream << (isLastChild ? "   " : " + ")
                  << "Definition "
                  << item.getId()
                  << endl;
}


// --- DefinitionTree tests -------------------------------------------------

//  NOTE:
//      Member functions that only delegate their work to the corresponding
//      members of the DefinitionContainer class template are not tested.
//      These should be sufficiently covered by the DefinitionContainer tests.


TEST(DefinitionTreeT,
     numRootItems_defaultTree_returnsZero)
{
    DefinitionTree< TestDef > tree;

    EXPECT_EQ(0, tree.numRootItems());
}


TEST(DefinitionTreeT,
     getRootItemByIndex_validIndex_returnsItem)
{
    DefinitionTree< TestDef > tree;

    auto item0 = cxxslb::make_unique< TestDef >(0, nullptr);
    auto item1 = cxxslb::make_unique< TestDef >(1, item0.get());
    auto item2 = cxxslb::make_unique< TestDef >(2, nullptr);

    auto* itemPtr0 = item0.get();
    auto* itemPtr2 = item2.get();

    tree.addItem(std::move(item0));
    tree.addItem(std::move(item1));
    tree.addItem(std::move(item2));

    EXPECT_EQ(itemPtr0, tree.getRootItemByIndex(0));
    EXPECT_EQ(itemPtr2, tree.getRootItemByIndex(1));
}


TEST(DefinitionTreeT,
     getRootItemByIndex_invalidIndex_throws)
{
    DefinitionTree< TestDef > tree;

    auto item0 = cxxslb::make_unique< TestDef >(0, nullptr);
    auto item1 = cxxslb::make_unique< TestDef >(1, item0.get());
    auto item2 = cxxslb::make_unique< TestDef >(2, nullptr);
    tree.addItem(std::move(item0));
    tree.addItem(std::move(item1));
    tree.addItem(std::move(item2));

    EXPECT_THROW(tree.getRootItemByIndex(2), std::out_of_range);
}


TEST(DefinitionTreeT,
     addItem_itemPtr_incrementsNumItems)
{
    DefinitionTree< TestDef > tree;

    auto  root    = cxxslb::make_unique< TestDef >(0, nullptr);
    auto* rootPtr = root.get();
    tree.addItem(std::move(root));
    EXPECT_EQ(1, tree.numItems());

    tree.addItem(cxxslb::make_unique< TestDef >(1, rootPtr));
    EXPECT_EQ(2, tree.numItems());

    tree.addItem(cxxslb::make_unique< TestDef >(2, rootPtr));
    EXPECT_EQ(3, tree.numItems());
}


TEST(DefinitionTreeT,
     addItem_itemPtr_incrementsNumRootItems)
{
    DefinitionTree< TestDef > tree;

    auto  root    = cxxslb::make_unique< TestDef >(0, nullptr);
    auto* rootPtr = root.get();
    tree.addItem(std::move(root));
    EXPECT_EQ(1, tree.numRootItems());

    tree.addItem(cxxslb::make_unique< TestDef >(1, rootPtr));
    EXPECT_EQ(1, tree.numRootItems());

    tree.addItem(cxxslb::make_unique< TestDef >(2, nullptr));
    EXPECT_EQ(2, tree.numRootItems());
}


TEST(DefinitionTreeT,
     addItem_nullptr_throws)
{
    DefinitionTree< TestDef > tree;

    EXPECT_THROW(tree.addItem(nullptr), std::invalid_argument);
}


TEST(DefinitionTreeT,
     optimizeLookups_treeWithUnsortedIds_keepsRootItems)
{
    DefinitionTree< TestDef > tree;

    auto root0 = cxxslb::make_unique< TestDef >(42, nullptr);
    auto root1 = cxxslb::make_unique< TestDef >(0, nullptr);
    auto root2 = cxxslb::make_unique< TestDef >(24, nullptr);

    auto* rootPtr0 = root0.get();
    auto* rootPtr1 = root1.get();
    auto* rootPtr2 = root2.get();

    tree.addItem(std::move(root0));
    tree.addItem(std::move(root1));
    tree.addItem(std::move(root2));

    EXPECT_EQ(rootPtr0, tree.getRootItemByIndex(0));
    EXPECT_EQ(rootPtr1, tree.getRootItemByIndex(1));
    EXPECT_EQ(rootPtr2, tree.getRootItemByIndex(2));
    EXPECT_THROW(tree.getRootItemByIndex(3), std::out_of_range);
}


TEST(DefinitionTreeT,
     operatorLeftShift_defaultTreeToStream_printsNothing)
{
    DefinitionTree< TestDef > tree;

    string expected;

    ostringstream result;
    result << tree;
    EXPECT_EQ(expected, result.str());
}


TEST(DefinitionTreeT,
     operatorLeftShift_nonEmptyTreeToStream_printsRootItems)
{
    DefinitionTree< TestDef > tree;

    auto item0 = cxxslb::make_unique< TestDef >(0, nullptr);
    auto item1 = cxxslb::make_unique< TestDef >(1, item0.get());
    auto item2 = cxxslb::make_unique< TestDef >(2, nullptr);
    auto item3 = cxxslb::make_unique< TestDef >(3, item2.get());
    auto item4 = cxxslb::make_unique< TestDef >(4, nullptr);
    auto item5 = cxxslb::make_unique< TestDef >(5, item4.get());

    tree.addItem(std::move(item0));
    tree.addItem(std::move(item1));
    tree.addItem(std::move(item2));
    tree.addItem(std::move(item3));
    tree.addItem(std::move(item4));
    tree.addItem(std::move(item5));

    string expected(" + Definition 0\n"
                    " + Definition 2\n"
                    "   Definition 4\n");

    ostringstream result;
    result << tree;
    EXPECT_EQ(expected, result.str());
}
}    // unnamed namespace
