/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2017-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include <pearl/LocationSet.h>

#include <sstream>
#include <string>
#include <vector>

#include <cxxslb/memory.h>

#include <gtest/gtest.h>

#include <pearl/Error.h>
#include <pearl/Location.h>
#include <pearl/String.h>
#include <pearl/SystemNode.h>
#include <pearl/iomanip.h>

#include "DefinitionContainer.h"
#include "Process.h"
#include "iomanip_detail.h"

using namespace std;
using namespace testing;
using namespace pearl;
using namespace pearl::detail;


namespace
{
// --- Helper ---------------------------------------------------------------

// Test fixture for location set tests
class LocationSetT
    : public Test
{
    public:
        void
        createSetMembers(Location::IdType maxId);


    protected:
        std::vector< const Location* >   mSetMembers;
        DefinitionContainer< Location >  mLocations;
        std::unique_ptr< LocationGroup > mLocationGroup;
        std::unique_ptr< SystemNode >    mSystemNode;
        DefinitionContainer< String >    mStrings;
};


// --- LocationSet tests ----------------------------------------------------

TEST_F(LocationSetT,
       getId_item_returnsIdentifier)
{
    LocationSet item0(0, String::UNDEFINED, mSetMembers);
    EXPECT_EQ(0, item0.getId());

    LocationSet item1(1, String::UNDEFINED, mSetMembers);
    EXPECT_EQ(1, item1.getId());

    LocationSet item42(42, String::UNDEFINED, mSetMembers);
    EXPECT_EQ(42, item42.getId());
}


TEST_F(LocationSetT,
       getName_item_returnsName)
{
    LocationSet item0(0, String::UNDEFINED, mSetMembers);
    EXPECT_EQ(&String::UNDEFINED, &item0.getName());

    String      name1(0, "Foo");
    LocationSet item1(1, name1, mSetMembers);
    EXPECT_EQ(&name1, &item1.getName());
}


TEST_F(LocationSetT,
       getType_item_returnsType)
{
    LocationSet item(0, String::UNDEFINED, mSetMembers);
    EXPECT_EQ(GroupingSet::LOCATION_SET, item.getType());
}


TEST_F(LocationSetT,
       numLocations_emptyItem_returnsZero)
{
    LocationSet item(0, String::UNDEFINED, mSetMembers);
    EXPECT_EQ(0, item.numLocations());
}


TEST_F(LocationSetT,
       numLocations_item_returnsMemberCount)
{
    createSetMembers(5);

    LocationSet item(0, String::UNDEFINED, mSetMembers);
    EXPECT_EQ(5, item.numLocations());
}


TEST_F(LocationSetT,
       getLocation_itemAndValidRank_returnsMember)
{
    createSetMembers(3);

    LocationSet item(0, String::UNDEFINED, mSetMembers);
    EXPECT_EQ(mSetMembers[0], &item.getLocation(0));
    EXPECT_EQ(mSetMembers[1], &item.getLocation(1));
    EXPECT_EQ(mSetMembers[2], &item.getLocation(2));
}


TEST_F(LocationSetT,
       getLocation_itemAndInvalidRank_throws)
{
    createSetMembers(3);

    LocationSet item(0, String::UNDEFINED, mSetMembers);
    EXPECT_THROW(item.getLocation(3), pearl::RuntimeError);
}


TEST_F(LocationSetT,
       operatorLeftShift_itemAtSkipDetail_printsNothing)
{
    createSetMembers(2);

    String      name(24, "Foo");
    LocationSet item(42, name, mSetMembers);

    string expected;

    ostringstream result;
    result << setDetail(DETAIL_SKIP)
           << item;
    EXPECT_EQ(expected, result.str());
}


TEST_F(LocationSetT,
       operatorLeftShift_itemAtAbbrevDetail_printsIdentifier)
{
    createSetMembers(2);

    String      name(24, "Foo");
    LocationSet item(42, name, mSetMembers);

    string expected("42");

    ostringstream result;
    result << setDetail(DETAIL_ABBRV)
           << item;
    EXPECT_EQ(expected, result.str());
}


TEST_F(LocationSetT,
       operatorLeftShift_itemAtCompactDetail_printsOneliner)
{
    createSetMembers(2);

    String      name(24, "Foo");
    LocationSet item(42, name, mSetMembers);

    string expected("LocationSet { id: 42, name: \"Foo\", size: 2, members: [...] }");

    ostringstream result;
    result << item;
    EXPECT_EQ(expected, result.str());
}


TEST_F(LocationSetT,
       operatorLeftShift_itemAtFullDetail_printsData)
{
    createSetMembers(2);

    String      name(24, "Foo");
    LocationSet item(42, name, mSetMembers);

    string expected("LocationSet {\n"
                    "    id: 42\n"
                    "    name: String { id: 24, text: \"Foo\" }\n"
                    "    size: 2\n"
                    "    members: [ 0, 1 ]\n"
                    "}");

    ostringstream result;
    result << setDetail(DETAIL_FULL)
           << item;
    EXPECT_EQ(expected, result.str());
}


TEST_F(LocationSetT,
       operatorLeftShift_itemAtIncreasedDetail_printsMembers)
{
    createSetMembers(2);

    String      name(24, "Foo");
    LocationSet item(42, name, mSetMembers);

    string expected("LocationSet {\n"
                    "    id: 42\n"
                    "    name: String {\n"
                    "        id: 24\n"
                    "        text: \"Foo\"\n"
                    "    }\n"
                    "    size: 2\n"
                    "    members: [\n"
                    "        Location { id: 0, name: \"Thread 0\", type: CPU thread, #events: 0, parent: 0 }\n"
                    "        Location { id: 1, name: \"Thread 1\", type: CPU thread, #events: 0, parent: 0 }\n"
                    "    ]\n"
                    "}");

    ostringstream result;
    result << setDetail(DETAIL_FULL + 1)
           << item;
    EXPECT_EQ(expected, result.str());
}


// --- Helper ---------------------------------------------------------------

void
LocationSetT::createSetMembers(const Location::IdType maxId)
{
    // Create dummy SystemNode (used as parent)
    mSystemNode =
        cxxslb::make_unique< SystemNode >(0,
                                          String::UNDEFINED,
                                          String::UNDEFINED,
                                          nullptr);

    // Create dummy LocationGroup (used as parent)
    mLocationGroup =
        cxxslb::make_unique< Process >(0,
                                       String::UNDEFINED,
                                       mSystemNode.get(),
                                       nullptr);

    // Create locations
    mSetMembers.reserve(maxId);
    for (Location::IdType id = 0; id < maxId; ++id)
    {
        mStrings.addItem(
            cxxslb::make_unique< String >(id,
                                          "Thread " + std::to_string(id)));
        mLocations.addItem(
            cxxslb::make_unique< Location >(id,
                                            *mStrings.getItemById(id),
                                            Location::TYPE_CPU_THREAD,
                                            0,
                                            mLocationGroup.get()));
        mSetMembers.push_back(mLocations.getItemById(id));
    }
}
}    // unnamed namespace
