/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2017-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "SystemTree.h"

#include <sstream>
#include <string>
#include <utility>

#include <cxxslb/memory.h>

#include <gtest/gtest.h>

#include <pearl/Location.h>
#include <pearl/LocationGroup.h>
#include <pearl/String.h>
#include <pearl/SystemNode.h>

#include "DefinitionContainer.h"
#include "Process.h"

using namespace std;
using namespace testing;
using namespace pearl;
using namespace pearl::detail;


namespace
{
// --- Helper ---------------------------------------------------------------

// Test fixture for system tree tests
class SystemTreeT
    : public Test
{
    public:
        const SystemNode&
        createSystemTree();


    protected:
        DefinitionContainer< String >        mStrings;
        DefinitionContainer< Location >      mLocations;
        DefinitionContainer< LocationGroup > mLocationGroups;
        DefinitionContainer< SystemNode >    mSystemNodes;
};


// --- print tests (system trees) -------------------------------------------

TEST_F(SystemTreeT,
       print_undefinedLocation_printsUndefined)
{
    string expected(" + (undefined)\n");

    ostringstream result;
    print(result, Location::UNDEFINED);
    EXPECT_EQ(expected, result.str());

    expected = "# + (undefined)\n";

    result.str("");
    print(result, Location::UNDEFINED, "#");
    EXPECT_EQ(expected, result.str());
}


TEST_F(SystemTreeT,
       print_systemTree_printsTree)
{
    const SystemNode& tree = createSystemTree();

    string expected(" + Machine  <sn:0>\n"
                    "    + Node 0  <sn:1>\n"
                    "    |  + Rank 0  <lg:0>\n"
                    "    |     + Master thread  <l:0>\n"
                    "    |     + GPU stream  <l:1>\n"
                    "    + Switch  <sn:2>\n"
                    "       + Node 1  <sn:3>\n"
                    "       |  + Rank 1  <lg:2>\n"
                    "       |     + Master thread  <l:3>\n"
                    "       + Switch metric  <lg:1>\n"
                    "          + Metric  <l:2>\n");

    ostringstream result;
    print(result, tree, true);
    EXPECT_EQ(expected, result.str());
}


TEST_F(SystemTreeT,
       print_systemTreeLastChild_printsTreeWithPrefix)
{
    const SystemNode& tree = createSystemTree();

    string expected(" + Machine  <sn:0>\n"
                    " |  + Node 0  <sn:1>\n"
                    " |  |  + Rank 0  <lg:0>\n"
                    " |  |     + Master thread  <l:0>\n"
                    " |  |     + GPU stream  <l:1>\n"
                    " |  + Switch  <sn:2>\n"
                    " |     + Node 1  <sn:3>\n"
                    " |     |  + Rank 1  <lg:2>\n"
                    " |     |     + Master thread  <l:3>\n"
                    " |     + Switch metric  <lg:1>\n"
                    " |        + Metric  <l:2>\n");

    ostringstream result;
    print(result, tree, false);
    EXPECT_EQ(expected, result.str());
}


// --- Helper ---------------------------------------------------------------

// Creates a system tree covering all possible situations w.r.t. the possible
// node types:
//
//    + Machine  <sn:0>
//       + Node 0  <sn:1>
//       |  + Rank 0  <lg:0>            # LG w/ different types of L
//       |     + Master thread  <l:0>
//       |     + GPU stream  <l:1>
//       + Switch  <sn:2>               # STN w/ child STN and LG
//          + Node 1  <sn:3>
//          |  + Rank 1  <lg:2>
//          |     + Master thread  <l:3>
//          + Switch metric  <lg:1>
//             + Metric  <l:2>
//
const SystemNode&
SystemTreeT::createSystemTree()
{
    // Create String definitions
    auto strMachine      = cxxslb::make_unique< String >(0, "Machine");
    auto strNode0        = cxxslb::make_unique< String >(1, "Node 0");
    auto strNode1        = cxxslb::make_unique< String >(2, "Node 1");
    auto strSwitch       = cxxslb::make_unique< String >(3, "Switch");
    auto strSwitchMetric = cxxslb::make_unique< String >(4, "Switch metric");
    auto strRank0        = cxxslb::make_unique< String >(5, "Rank 0");
    auto strRank1        = cxxslb::make_unique< String >(6, "Rank 1");
    auto strThread       = cxxslb::make_unique< String >(7, "Master thread");
    auto strStream       = cxxslb::make_unique< String >(8, "GPU stream");
    auto strMetric       = cxxslb::make_unique< String >(9, "Metric");

    // Create SystemNode definitions
    auto snMachine =
        cxxslb::make_unique< SystemNode >(0,
                                          *strMachine,
                                          String::UNDEFINED,
                                          nullptr);
    auto snNode0 =
        cxxslb::make_unique< SystemNode >(1,
                                          *strNode0,
                                          String::UNDEFINED,
                                          snMachine.get());
    auto snSwitch =
        cxxslb::make_unique< SystemNode >(2,
                                          *strSwitch,
                                          String::UNDEFINED,
                                          snMachine.get());
    auto snNode1 =
        cxxslb::make_unique< SystemNode >(3,
                                          *strNode1,
                                          String::UNDEFINED,
                                          snSwitch.get());

    // Create LocationGroup definitions
    auto lgRank0 =
        cxxslb::make_unique< Process >(0,
                                       *strRank0,
                                       snNode0.get(),
                                       nullptr);
    auto lgSwitchMetric =
        cxxslb::make_unique< LocationGroup >(1,
                                             *strSwitchMetric,
                                             LocationGroup::TYPE_UNKNOWN,
                                             snSwitch.get(),
                                             nullptr);
    auto lgRank1 =
        cxxslb::make_unique< Process >(2,
                                       *strRank1,
                                       snNode1.get(),
                                       nullptr);

    // Create Location definitions
    auto locThread0 =
        cxxslb::make_unique< Location >(0,
                                        *strThread,
                                        Location::TYPE_CPU_THREAD,
                                        0,
                                        lgRank0.get());
    auto locStream0 =
        cxxslb::make_unique< Location >(1,
                                        *strStream,
                                        Location::TYPE_ACCELERATOR_STREAM,
                                        0,
                                        lgRank0.get());
    auto locMetric =
        cxxslb::make_unique< Location >(2,
                                        *strMetric,
                                        Location::TYPE_METRIC,
                                        0,
                                        lgSwitchMetric.get());
    auto locThread1 =
        cxxslb::make_unique< Location >(3,
                                        *strThread,
                                        Location::TYPE_CPU_THREAD,
                                        0,
                                        lgRank1.get());

    mStrings.addItem(std::move(strMachine));
    mStrings.addItem(std::move(strNode0));
    mStrings.addItem(std::move(strNode1));
    mStrings.addItem(std::move(strSwitch));
    mStrings.addItem(std::move(strSwitchMetric));
    mStrings.addItem(std::move(strRank0));
    mStrings.addItem(std::move(strRank1));
    mStrings.addItem(std::move(strThread));
    mStrings.addItem(std::move(strStream));
    mStrings.addItem(std::move(strMetric));

    auto* root = snMachine.get();
    mSystemNodes.addItem(std::move(snMachine));
    mSystemNodes.addItem(std::move(snNode0));
    mSystemNodes.addItem(std::move(snSwitch));
    mSystemNodes.addItem(std::move(snNode1));

    mLocationGroups.addItem(std::move(lgRank0));
    mLocationGroups.addItem(std::move(lgSwitchMetric));
    mLocationGroups.addItem(std::move(lgRank1));

    mLocations.addItem(std::move(locThread0));
    mLocations.addItem(std::move(locStream0));
    mLocations.addItem(std::move(locMetric));
    mLocations.addItem(std::move(locThread1));

    return *root;
}
}    // unnamed namespace
