/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2023                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "SmallBlockAllocator.h"

#include <algorithm>
#include <vector>

#include <cxxslb/memory.h>

#include <gtest/gtest.h>

using namespace testing;
using namespace pearl::detail;


namespace
{
// --- Helpers --------------------------------------------------------------

// Test fixture used for small block allocator tests
class SmallBlockAllocatorT
    : public TestWithParam< std::size_t >
{
    public:
        // Number of blocks to allocate; this fills at least 8 chunks
        static constexpr std::size_t MAX_BLOCKS = 2048;


        SmallBlockAllocatorT()
            : mAllocator(cxxslb::make_unique< SmallBlockAllocator >()),
              mBlockSize(GetParam())
        {
            mBlocks.reserve(MAX_BLOCKS);
        }

        ~SmallBlockAllocatorT()
        {
            // Deallocate in LIFO order
            while (!mBlocks.empty())
            {
                mAllocator->deallocate(mBlocks.back(), mBlockSize);
                mBlocks.pop_back();
            }
        }


    protected:
        std::unique_ptr< SmallBlockAllocator > mAllocator;
        std::size_t                            mBlockSize;
        std::vector< void* >                   mBlocks;
};


// --- SmallBlockAllocator tests --------------------------------------------

TEST_P(SmallBlockAllocatorT,
       allocate_allocator_returnsNonNullptr)
{
    for (std::size_t i = 0; i < MAX_BLOCKS; ++i)
    {
        void* const ptr = mAllocator->allocate(mBlockSize);
        mBlocks.push_back(ptr);

        EXPECT_NE(ptr, nullptr);
    }

    // Implicitly tests deallocation in LIFO order via fixture destructor
}


TEST_P(SmallBlockAllocatorT,
       allocate_allocator_returnsUniquePointers)
{
    for (std::size_t i = 0; i < MAX_BLOCKS; ++i)
    {
        void* const ptr = mAllocator->allocate(mBlockSize);
        mBlocks.push_back(ptr);
    }

    // Create vector with sorted pointers & remove duplicates
    std::vector< void* > sorted(mBlocks);
    std::sort(sorted.begin(), sorted.end());
    sorted.erase(std::unique(sorted.begin(), sorted.end()), sorted.end());

    EXPECT_EQ(mBlocks.size(), sorted.size());

    // Implicitly tests deallocation in LIFO order via fixture destructor
}


// Instantiate tests with various block sizes, especially with the boundary
// values triggering different behavior
INSTANTIATE_TEST_SUITE_P(SmallBlockAllocator,
                         SmallBlockAllocatorT,
                         Values(0, 1, 31, 32,
                                SmallBlockAllocator::MAX_BLOCKSIZE,
                                SmallBlockAllocator::MAX_BLOCKSIZE + 1));
}    // unnamed namespace
